%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%        ML-CAM Translation                             %%%
%%%        This version uses de Bruijn indexing to        %%%
%%%        generate code for identifiers.                 %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

module camml.

import mldecl camdecl.


type  translate        tm -> prog -> o.
type  trans            int -> tm -> com -> o.
type  transop          tm -> oper -> o.
type  generate_path    int -> com -> o.
type  assoc            tm -> int -> o.


doc translate "(translate ML CAM) translates mini-ML expression `ML' into CAM program CAM.".


translate E (program C) :-
        trans 0 E C.



trans Depth (# N) (quote (## N)) :- !.     %%% (1) Numbers

trans Depth tt (quote ctt) :- !.           %%% (2) true
trans Depth ff (quote cff) :- !.           %%% (3) false
trans Depth nill (quote cnil) :- !.

trans Depth (cond E1 E2 E3)                %%% (5) Conditional
      (& push (& C1 (branch C2 C3))) :- !,
        trans Depth E1 C1,
        trans Depth E2 C2,
        trans Depth E3 C3.

trans Depth (&& E1 E2)                     %%% (6) Product
      (& push (& C1 (& swap (& C2 pair)))) :- !,
        trans Depth E1 C1,
        trans Depth E2 C2.

trans Depth (@ Proj E) (& push (& C CamProj)) :- %%% (7,8) Projections
	(Proj = fst, CamProj = cfst ;
	 Proj = snd, CamProj = csnd), !,
        trans Depth E C.

trans Depth  (lamb E)  (cur C) :- !,       %%% (9) Lamb
        NewDepth is Depth + 1,
        pi K\(assoc K NewDepth  =>  trans NewDepth (E K) C).

trans Depth (let M E2)                     %%% (11) Let
      (& push (& C2 (& pair C1))) :- !,
        trans Depth E2 C2,
        NewDepth is Depth + 1,
        pi K\(assoc K NewDepth  =>  trans NewDepth (M K) C1).

trans Depth (fix M) (& push (rec C)) :- !, %%% (12) Fix
        NewDepth is Depth + 1,
        pi K\(assoc K NewDepth  =>  trans NewDepth (M K) C).

                                                 %%% (10) Application
trans Depth (@ (@ Op E1) E2)               %%% binary primitives
      (& push (& C1 (& swap (& C2 (& pair (op Camop)))))) :-
        member Op (cons == (cons != (cons lt (cons gt (cons cns
			(cons plus (cons minus (cons times nil)))))))), !,
        trans Depth E1 C1,
        trans Depth E2 C2,
        transop Op Camop.

trans Depth (@ Op E1)                      %%% unary primitives
      (& C1 (op Camop)) :-
%      (push & C1 & (op Camop)) :-
        member Op (cons hd (cons tl (cons empty nil))), !,
        trans Depth E1 C1,
        transop Op Camop.

trans Depth (@ E1 E2)                      %%% General Apply
      (& push (& C1 (& swap (& C2 (& pair app))))) :- !,
        trans Depth E1 C1,
        trans Depth E2 C2.

trans Depth K C :-                         %%% (4) Identifier 
        assoc K D,
        Distance is (Depth - D) + 1,
        generate_path Distance C.


%%% translations for primitive operators %%%
transop ==    cequal.
transop !=    cnequal.
transop lt    clt.
transop gt    cgt.
transop cns   ccons.
transop plus  cplus.
transop minus cminus.
transop times ctimes.
transop hd    chd.
transop tl    ctl.
transop empty cempty.


%%% translate N+1 into (cfst^N & csnd) %%%
generate_path 1 csnd :- !.
generate_path N (& cfst C) :- !,
        N1 is N - 1,
        generate_path N1 C.

%%%   End of module camml   %%%

