% A meta-interpreter for LambdaProlog without higher-order predicates.
% This handles the most general case for implication in programs.
% It does not interpret cut or special predicates.
% Author: Frank Pfenning (fp@cs.cmu.edu)

%% This version keeps track of the proof that LambdaProlog finds for
%% the original query.
%% It uses a neat trick to generate subgoals, by generating a meta-subgoal
%% instead!
%% This allows disjunctions in goals (like the other meta-interpreters),
%%  which seems somewhat out of place.
%% Cuts are merely for efficiency.
% We could distinguish between a lam abstracting over proofs
% (from => ) and one abstracing over term: (from pi )
% similarly for pair and fst.

%% This is an attempt to implement a meta-interpreter for a language with
%% cut.  It can be inherited in some simple cases!?

module metacut.

type  psolve     o -> proof -> o.
type  phyp       o -> proof -> o.
type  pmatch_atom o -> o -> proof -> proof -> o -> o.

type  one	 proof.
type  pair	 A -> proof -> proof.
type  inleft	 proof -> proof.
type  inright	 proof -> proof.
type  lam	 (A -> proof) -> proof.	         % UNTYPED lambda!

type  fst	 proof -> proof.
type  snd	 proof -> proof.
type  app	 proof -> A -> proof.

psolve true      one          :- !.
psolve (G1 , G2) (pair M1 M2) :- !, psolve G1 M1, psolve G2 M2.
psolve (G1 ; G2) (inleft M1)  :-    psolve G1 M1.             % No cut here!
psolve (G1 ; G2) (inright M2) :- !, psolve G2 M2.
psolve (D => G)  (lam M)      :- !, pi U \ (phyp D U => psolve G (M U)).
psolve (pi G)    (lam M)      :- !, pi X \ (psolve (G X) (M X)).
psolve (sigma G) (pair T M)   :- !, psolve (G T) M.

psolve Atom	 M :- !, phyp D U, 
		         pmatch_atom D Atom U M MetaSubgoal,
			 MetaSubgoal.

pmatch_atom (D1 , D2) Atom N 	     M MetaSubgoal :-     % No cut here!
    pmatch_atom D1    Atom (fst N)   M MetaSubgoal.

pmatch_atom (D1 , D2) Atom N         M MetaSubgoal :- !,
    pmatch_atom D2    Atom (snd N)   M MetaSubgoal.

pmatch_atom (D1 ; D2) Atom N	     M MetaSubgoal :- !,
    error (writesans "Illegal disjunction in program: ", write (D1 ; D2)).

%% This can handle a top-level cut, written a specific way.
%% All the cuts in this file could be written this way, except for
%% the ones on the following line.
pmatch_atom (! => D)  Atom N	      M (! , MetaSubgoal) :- !,
    pmatch_atom D     Atom N	      M MetaSubgoal.

pmatch_atom (G => D)  Atom N	      M (psolve G N1 , MetaSubgoal) :- !,
    pmatch_atom D     Atom (app N N1) M MetaSubgoal.

pmatch_atom (pi D)    Atom N	      M MetaSubgoal :- !,
    pmatch_atom (D T) Atom (app N T)  M MetaSubgoal.

pmatch_atom (sigma D) Atom N	      M MetaSubgoal :- !,
    error (writesans "Illegal existential in program: ", write (sigma D)).

pmatch_atom Atom      Atom N          N true.
    % No cut here: if there are multiple unifiers for Atom Atom, we
    % want to make sure they are all enumerated!

