% A meta-interpreter for LambdaProlog.
% This handles the most general case for implication in programs.
% It does not interpret cut or special predicates.
% Author: Frank Pfenning (fp@cs.cmu.edu)

%% This version keeps track of the proof that LambdaProlog finds for
%% the original query.
%% It uses a neat trick to generate subgoals, by generating a meta-subgoal
%% instead!
%% This allows disjunctions in goals (like the other meta-interpreters),
%%  which seems somewhat out of place.

% We could distinguish between a lam and app over proofs
% (from => ) over terms: (from pi ).
% Similarly for pair and fst. (from ``,'' and sigma, respectively).
% Also, the lambda terms generated are purely untyped (though they have
% LambdaProlog types, of course.

module metaproof.

type  psolve     o -> proof -> o.
type  phyp       o -> proof -> o.
type  pmatch_atom o -> o -> proof -> proof -> o -> o.

type  one	 proof.
type  pair	 A -> proof -> proof.
type  inleft	 proof -> proof.
type  inright	 proof -> proof.
type  lam	 (A -> proof) -> proof.

% fst has really type ``proof -> A'', but is guaranteed to be applied
% only to pairs of proofs: ``sigma D'' is not a legal assumption!
% Destructors for inleft and inright may be omitted for the same reason:
% ``D1 ; D2'' is not a legal assumption.
type  fst	 proof -> proof.
type  snd	 proof -> proof.
type  app	 proof -> A -> proof.

doc   psolve
 "``psolve G M'' expects a goal G and attempts to find a proof M for this
goal.  It does not handle cut or special predicates, but everything else,
including higher-order predicates.".

psolve true      one          :- !.
psolve (G1 , G2) (pair M1 M2) :- !, psolve G1 M1, psolve G2 M2.
psolve (G1 ; G2) (inleft M1)  :-    psolve G1 M1.             % No cut here!
psolve (G1 ; G2) (inright M2) :- !, psolve G2 M2.
psolve (D => G)  (lam M)      :- !, pi U \ (phyp D U => psolve G (M U)).
psolve (pi G)    (lam M)      :- !, pi X \ (psolve (G X) (M X)).
psolve (sigma G) (pair T M)   :- !, psolve (G T) M.
psolve !	 M	      :- !, error (write "Cannot handle cut.").
psolve Atom	 M :- !, phyp D U, 
		         pmatch_atom D Atom U M MetaSubgoal,
			 MetaSubgoal.

doc    pmatch_atom
 "``pmatch_atom D Atom N M MetaSubgoal'' expects a program (D-formula) D,
an atomic goal Atom, and generates a MetaSubgoal that remains to be solved,
in order to solve Atom.  On backtracking, it enumerates MetaSubgoals, until
all the possibilities in D are exhausted, whereupon it fails.
  N is expected to contain the proof of D (in the current context) and on
success M will contain the proof of Atom (also in the current context).
Note that typically M and MetaSubgoal will share variables, since the proof
of Atom depends on proofs of the subgoals (embodied in the MetaSubgoal).
  The fact that LambdaProlog is higher-order is convenient here, not because
the goals have type o (which could be replaced by a new type constant prop),
but because the MetaSubgoal can be used effectively to postpone (freeze)
subgoals, until is clear that it would be worthwile trying to solve them.".

pmatch_atom (D1 , D2) Atom N 	     M MetaSubgoal :-     % No cut here!
    pmatch_atom D1    Atom (fst N)   M MetaSubgoal.

pmatch_atom (D1 , D2) Atom N         M MetaSubgoal :- !,
    pmatch_atom D2    Atom (snd N)   M MetaSubgoal.

pmatch_atom (D1 ; D2) Atom N	     M MetaSubgoal :- !,
    error (writesans "Illegal disjunction in program: ", write (D1 ; D2)).

pmatch_atom (G => D)  Atom N	      M (psolve G N1 , MetaSubgoal) :- !,
    pmatch_atom D     Atom (app N N1) M MetaSubgoal.

pmatch_atom (pi D)    Atom N	      M MetaSubgoal :- !,
    pmatch_atom (D T) Atom (app N T)  M MetaSubgoal.

pmatch_atom (sigma D) Atom N	      M MetaSubgoal :- !,
    error (writesans "Illegal existential in program: ", write (sigma D)).

pmatch_atom Atom      Atom N          N true.
    % No cut here: if there are multiple unifiers for Atom Atom, we
    % want to make sure they are all enumerated!

