%%% Module to put a D-formula (legal program or assumption in eLP) into the
%%% normal form.  This is also done by the eLP interpreter, which explains the
%%% differences between an assumption in a program and the clauses that are
%%% added to the definition of a predicate when the clause is assumed.
%%% Author: fp.

module norm_dform.

%%% The exported predicate.

type   norm_dform	o -> o -> o.
doc    norm_dform
 "``norm_dform D ND'' expects a D-formula D and returns in ND the clausal
normal form of D.  The clausal normal form Nd is defined by
   Nd ::= Qnd | (Nd , Nd)	% Normal D-formula.
  Qnd ::= Ind | pi Qnd		% Quantified Normal D-formula.
  Ind ::= Atom :- G		% Implicational Normal D-formula.
where G stands for a G-formula and Atom for an atomic formula.".

%%% Auxiliary predicates.

%%% Moving a universal quantifier inside (across conjunctions).

type	quantify_inside o -> o -> o.
doc	quantify_inside
 "``quantify_inside (pi ND) NewND'' succeeds if NewND is the clausal normal
form of (pi ND), where ND is expected to be already in clausal normal form.
It distributes the top-level universal quantifier in (pi ND) across top-level
conjunctions in NewND.".

quantify_inside (pi V \ (ND1 V , ND2 V)) (NewND1 , NewND2) :-
   quantify_inside (pi V \ (ND1 V)) NewND1,
   quantify_inside (pi V \ (ND2 V)) NewND2.

quantify_inside ND ND.

%%% Moving an assumption inside (across conjunctions and universal
%%% quantifiers).

type	conjoin_inside	o -> o -> o -> o.
doc	conjoin_inside
 "``conjoin_inside G ND NewND'' succeeds if NewND is the clausal normal form
of ND :- G, where ND is expected to be in clausal normal form already.
It moves G through conjunctions and universal quantifiers to the left-hand
side of an implication.".

conjoin_inside G (ND1 , ND2) (NewND1 , NewND2) :-
  conjoin_inside G ND1 NewND1,
  conjoin_inside G ND2 NewND2.

conjoin_inside G (pi ND) (pi NewND) :-
  pi V \ (conjoin_inside G (ND V) (NewND V)).

conjoin_inside G (Head :- true) (Head :- G). % A simple optimization.

conjoin_inside G (Head :- Body) (Head :- (G, Body)).

%%% The main predicate.

norm_dform (D1 , D2) (ND1 , ND2) :-
   norm_dform D1 ND1,
   norm_dform D2 ND2.

norm_dform (pi D) ND :-
   pi V \ (norm_dform (D V) (ND1 V)),
   quantify_inside (pi ND1) ND.

norm_dform (D :- G) ND :-
   norm_dform D ND1,
   conjoin_inside G ND1 ND.

norm_dform true ND :-
   error (writesans "``true'' is not a legal assumption.").
norm_dform (D1 ; D2) ND :-
   error (writesans "Disjunction ", write (D1 ; D2),
	  writesans " not legal as an assumption.").
norm_dform (sigma D) ND :-
   error (writesans "Existentially quantified ", write (sigma D),
	  writesans " not legal as an assumption.").

norm_dform Atom (Atom :- true).

% Some examples:
% norm_dform q ND.
% norm_dform (true => q) ND.
% norm_dform (p1 => p2 => q) ND.
% norm_dform (p1 => pi X \ (p3 X => q)) ND.
% norm_dform (p1 => pi X \ (p3 X => (q,r))) ND.
% norm_dform (p1 => pi X \ (s, (p3 X => (q,r)))) ND.
