%%% "Compilation" for the second-order polymorphic lambda calculus.
%%% Right now the main concern is to recognize the defined natural numbers
%%% and compile them into the primitive natural numbers.
%%% Author: fp.

module polycompile.
import polysig lamsig lameval.

type	compile		tm -> ta -> term -> o.
doc	compile
 "``compile U A M'' compiles the fully annotated term U of type A to the
untyped term M.".

% This first section repeats type inference, except that the "fully" typed
% term is created for future analysis by the compiler.   By fully typed
% we mean annotated to such a degree that a top-down traversal of the
% term can always carry along the correct type for each subterm.

type  ftype	tm -> tm -> ta -> o.
index ftype	1.

%% inherited from polysig.
% type  vartype	tm -> ta -> o.
% index vartype	1.

ftype thisisahack U Alpha :- !,
  error (write "Illegal free variable in input.").

ftype (pmmlam     U) (mmlam Beta V) Alpha :- !,
  Alpha = (afun Beta Gamma),
  pi X\ (vartype X Beta => ftype (U X) (V X) Gamma).

ftype (mmlam Beta U) (mmlam Beta V) Alpha :- !,
  Alpha = (afun Beta Gamma),
  pi X\ (vartype X Beta => ftype (U X) (V X) Gamma).

% Application must know the type of the argument for top-down traversal.
ftype (mmapp U0 U1) (mmapp V0 (cm V1 Beta)) Alpha :- !,
  ftype U0 V0 (afun Beta Alpha),
  ftype U1 V1 Beta.

ftype (amlam U) (amlam V) Alpha :- !,
   Alpha = (del Beta),
   pi theta\ (ftype (U theta) (V theta) (Beta theta)).

% Type application must know the type of the function for top-down traversal.
ftype (pmaapp U    ) (maapp (cm V (del Beta)) Gamma) Alpha :- !,
  ftype U V (del Beta),
  Alpha = (Beta Gamma).

% Here, too: make function type known.
ftype (maapp U Gamma) (maapp (cm V (del Beta)) Gamma) Alpha :- !,
  ftype U V (del Beta),
  Alpha = (Beta Gamma).

ftype (cm U Beta) V Alpha :- !,
  ftype U V Beta,
  Alpha = Beta.

ftype X X Alpha :- vartype X Beta, !,
   Alpha = Beta.

ftype X U Alpha :- !,
    error (writesans "Unknown constant ", write X, nl).

%%% There are two compile predicates: one for a context where a term
%%% is used as a function, and one where it is used as an argument.
%%% These are called fun_compile and arg_compile, respectively.
%%% var_compile is used to associated tm and term-variables during the
%%% descent.

type fun_compile tm -> ta -> term -> o.
type arg_compile tm -> ta -> term -> o.
type var_compile tm -> term -> o.

index fun_compile 1.
index arg_compile 1.
index var_compile 1.

fun_compile U thisisatypehack M :-
   error (writesans "Illegal free type variable.").

fun_compile U (del A\ (afun A (afun (afun A A) A))) (app *iternat* M) :- !,
   arg_compile U (del A\ (afun A (afun (afun A A) A))) M.

fun_compile (mmlam Beta U) (afun Beta Gamma) (lam M) :- !,
   pi x:tm \ (pi y:term \ (var_compile x y => arg_compile (U x) Gamma (M y))).

fun_compile (mmapp U0 (cm U1 Beta)) Gamma (app M0 M1) :- !,
   arg_compile U1 Beta M1,
   fun_compile U0 (afun Beta Gamma) M0.

fun_compile (amlam U) (del Gamma) M :- !,
   pi Theta \ (arg_compile (U Theta) (Gamma Theta) M).

fun_compile (maapp (cm U (del Gamma)) Beta) (Gamma Beta) M :- !,
   fun_compile U (del Gamma) M.

fun_compile X Alpha Y :- var_compile X Y, !.

fun_compile X Alpha V :- !,
    error (writesans "Unknown constant or constructor or ill-typed phrase",
	   write X, nl). 

%%% Now compilation in an argument position.

arg_compile U thisisatypehack M :-
   error (writesans "Illegal free type variable.").

arg_compile (mmlam Beta U) (afun Beta Gamma) (lam M) :- !,
   pi x:tm \ (pi y:term \ (var_compile x y => arg_compile (U x) Gamma (M y)).

arg_compile (mmapp U0 (cm U1 Beta)) Gamma (app M0 M1) :- !,
   arg_compile U1 Beta M1,
   fun_compile U0 (afun Beta Gamma) M0.

arg_compile (amlam U) (del A\ (afun A (afun (afun A A) A))) M :- !,
   pi Theta \ 
    (arg_compile (U Theta) (afun Theta (afun (afun Theta Theta) Theta)) M1,
     do_red M1 *zero* M2,
     do_red M2 *succ* M
     ).

arg_compile (amlam U) (del Gamma) M :- !,
   pi Theta \ (arg_compile (U Theta) (Gamma Theta) M).

arg_compile (maapp (cm U (del Gamma)) Beta) (Gamma Beta) M :- !,
   fun_compile U (del Gamma) M.

arg_compile X Alpha Y :- var_compile X Y, !.

arg_compile X Alpha M :- !,
    error (writesans "Unknown constant or constructor or ill-typed phrase ",
	   write X). 

%%% Now some interface predicates.

compile U Alpha M :-
   %% First: infer fully typed term.
   ((ftype U V Alpha, !);
    writesans "Type inference failed.  Trying diagnostics...",
    (failure_diagnostic U Alpha;
     error (writesans "Diagnostic succeeded where inference failed!."))),
   %% Then compile fully typed term.
   arg_compile V Alpha M.
