%% This module is implicitly imported by all other modules.
%% It contains all the basic primitives, special predicates, plus
%% a number of defined utilities.
%% use ``?- help.'' and ``?- describe Item.'' for on-line documentation,
%% and ``?- document "filename".'' to generate up-to-date off-line
%% documentation from this file.

module elp.

kind	o		type.	% Propositions (a salute to Church)

%% Primitive types.

kind	string	type.		% strings.  (e.g. "Some text")
kind	int     type.		% integers. (e.g. 3)

%% First, the logical constants:
%% Documentation for these comes later.

type	true		o.
type	','		o -> o -> o.
type	';'		o -> o -> o.
type	'=>'		o -> o -> o.
%% ``p :- q'' is an alternative concrete syntax for the implication
%% ``q => p'', but ':-' will not be used on unparsing or recognized
%% as an operator.

type	pi		(A -> o) -> o.
type	sigma		(A -> o) -> o.

%% Documentation predicates that are NOT special.

kind	elp_topic       type.
type	topic		elp_topic -> string -> o.
index	topic		1.

type	special		A -> elp_topic -> o.
type	const		A -> elp_topic -> o.
type	defined		A -> elp_topic -> o.
index	special		2.
index	const		2.
index	defined		2.

type	doc		A -> string -> o.
index	doc		1.

type    elp_help        elp_topic.
type    elp_log_prim    elp_topic.
type    elp_interpreter elp_topic.
type    elp_input       elp_topic.
type    elp_output      elp_topic.
type    elp_top_level   elp_topic.
type    elp_modules     elp_topic.
type	elp_parsing	elp_topic.
type    elp_switching   elp_topic.
type    elp_switches    elp_topic.
type    elp_traces      elp_topic.
type    elp_warnings    elp_topic.
type    elp_arith       elp_topic.
type	elp_lists	elp_topic.

topic	elp_help	"Help".

%%% We'll leave the internal constants 'topic', 'special', and 'const'
%%% unlisted by ``help''.
% const	topic		elp_help.
doc	topic
 "``topic T String'' succeeds if T is a topic of special predicates and
constants known to elp.".

% const	special	   	elp_help.
doc	special
 "``special Pred Topic'' succeeds if Pred is a special predicate of topic
Topic.".

% const	const		elp_help.
doc	const
 "``const C Topic'' succeeds if C is a constant that is defined in the elp
module under topic Topic.".

const	doc	        elp_help.
doc	doc
 "``doc C String'' succeeds if C is a constant that has been given documentation
string String.".

%% The logical primitives.
%% These aren't really specials, but they are `special' in the sense that
%% their meaning cannot be changed.
topic	elp_log_prim	"Logical Primitives".

special	true		elp_log_prim.
doc	true
 "Always succeeds.".

special	','		elp_log_prim.
doc	','
 "Conjunction.  ``G1, G2'' succeeds if G1 succeeds and then G2 succeeds.".

special	';'		elp_log_prim.
doc	';'
 "Disjunction. ``G1; G2'' succeeds if G1 succeeds, or if G1 fails, if G2 then
succeeds.".

special	'=>'		elp_log_prim.
doc	'=>'
 "Implication.  ``D => G'' succeeds if, under the assumption D, G succeeds.
Note that after G succeeds, D is not assumed for subsequent goals.  However,
if they fail, and the interpreter backtracks to obtain more solutions for
G, D will again be assumed.".

special	pi		elp_log_prim.
doc	pi
 "Universal quantification.  ``pi X \ G'' succeeds if G succeeds after
substituting a new parameter for X in G.  This new parameter may not
occur in the substitution term for any variable that is (implicitly or
explicitly) quantified `outside of' X.  Thus, ``pi X \ (X = Y)'' will
fail, since Y may not be instantiated with the parameter instantiating X.
However, ``pi X \ (X => X)'' will succeed, as will
``pi X \ (sigma Y \ (X = Y))''.".

special	sigma		elp_log_prim.
doc	sigma
 "Existential quantification.  ``sigma X \ G'' succeeds if G succeeds after
a new free variable is substituted for X in G.".

%% Then the special predicates:
topic elp_interpreter   "Interpreter".

type	fail		o.
special fail		elp_interpreter.
doc	fail
 "``fail'' always simply fails.
It could be defined by  ``fail :- pi C \ C''.".

type	!		o.
special !		elp_interpreter.
doc	!
 "``!'' commits to the current clause and all the choices made since the
clause was selected.  Note that this will also commit to the particular choice
of unifier that was made when unifying the head of the clause with a goal.
This situation does not arise when unique most general unifiers exist.  A
common idiom to avoid this commitment is to move unification to occur after
the commitment has been made.".

type	once		o -> o.
special once		elp_interpreter.
doc	once
 "``once G'' solves G and then commits to all choices made while solving
G.  It could (almost) be defined by
 once G :- pi a \ ((a :- G, !) => a).
except that a cut in G will commit to the current clause selection.
Thus, with the defined ``once'', ``?- once (!, fail) ; true.'' would
succeed, while the special ``once'' fails.
  The primary reason for making ``once'' a special predicate is that it will
eagerly remove the choice points from the stack, thus providing the only
effective way of controlling stack growth in eLP.  Another use is to achieve
closing of streams, when it is guaranteed that they will no longer be read
from or written to.  For example:
  read_first_item File Item :-
    once (reading File Instream \ (seeing Instream (reread X \ (Item = X;
      error (writesans ^"First item, ^", write X,
	     writesans ^", contains undeclared constants.^"))))).
".

type	'='		A -> A -> o.
special	'='		elp_interpreter.
doc	'='
 "``M = N'' simply unifies M and N.  '=' could be defined by ``M = M.''".

type	error		o -> o.
special	error		elp_interpreter.
doc	error
 "``error G'' signals an error by writing ^"eLP Error: ^" and then solves G.
G is intended to write out an error message.  For example,
  ``error (writesans ^"Cannot take square root of ^", write N,
	   writesans ^", a negative number.^")''".

type	error_catch	o -> o -> o.
special error_catch	elp_interpreter.
doc	error_catch
 "``error_catch G EG'' solves G.  If an error occurs (either signalled by
eLP or with the ``error'' special predicate) while solving G, EG will be
invoked after the error message has been printed.".

type	solve		o -> o.
solve G :- G.
defined solve		elp_interpreter.
doc	solve
 "``solve G'' simply solves goal G.
Writing ``solve G'' instead of ``G'' in the body of a clause has the effect
that any commit in G, will not commit to the current clause.  Thus
  ``(!, fail); true''	      will fail, and
  ``solve (!, fail); true''   will succeed.".

type	not	        o -> o.
not G :- solve G, !, fail.	% ``solve G'' instead of G is critical!
not G.
defined not		elp_interpreter.
doc	not
 "``not G'' succeeds if G fails.".

%% Special predicates for io.
topic   elp_input       "Input".
topic	elp_output	"Output".

kind	instream        type.

type	standard_input	instream.
const	standard_input	elp_input.
doc	standard_input
 "The standard input stream, usually the terminal.
  This is NOT affected by ``seeing'' and can be used to reliably read
input from the terminal.".

type	reading		string -> (instream -> o) -> o.
special reading		elp_input.
doc	reading
 "``reading Infile (Instream \ G)'' attempts to open an inputfile Infile
in the current directory (that is, ^"./^").  If successful, an input stream
is created, and the second argument is applied to the new input stream.
The stream is only available while solving G, but is not closed until
G itself fails.  An error is signalled if the file does not exist.".

type	seeing		instream -> o -> o.
special seeing		elp_input.
doc	seeing
 "``seeing Instream G'' sets current input stream to Instream while
solving G.  Calls to read or reread while solving G will take input from
Instream.".

kind	outstream	type.

type	standard_output	outstream.
const	standard_output	elp_output.
doc	standard_output
 "The standard output stream, usually the terminal.
  This is NOT affected by ``telling'' and can be used to reliably redirect
output to the terminal.".

type	writing		string -> (outstream -> o) -> o.
special	writing		elp_output.
doc	writing
  "``writing Outfile (Outstream \ G)'' attempts to open an outputfile Outfile
in the current directory  (that is, ^"./^").  If successful, an output stream
is created, and the second argument is applied to the new output stream.
The stream is only available while solving G, but is not closed until
G itself fails.".

type	telling		outstream -> o -> o.
special	telling		elp_output.
doc	telling
 "``telling Instream G'' sets current output stream to Outstream while
solving G.  Calls to write while solving G will write to Outstream.".

type	prompting	string -> o -> o.
special prompting	elp_output.
doc	prompting
 "``prompting String G'' sets the prompt for reading from the terminal to
String while solving G.".

type	reread		(A -> o) -> o.
special	reread		elp_input.
doc	reread
 "``reread X \ G'' reads a term M from the current input stream and then
solves the goal ``(X \ G) M''.  The type of M is unified with the type of X,
and an error is signalled if this type-unification fails.  When G fails, the
next item from the current input stream is read and passed as argument to
(X \ G).  reread fails when the end of the input stream is reached.
See also ``read''.".

type	write		A -> o.
special	write		elp_output.
doc	write
 "``write X'' writes X to the current output stream.  Various switches control
how the output appears, but it should always be possible to ``read'' something
that was written.".

type    writetype	A -> o.
special writetype	elp_output.
doc	writetype
 "``writetype X'' writes the type of X to the current output stream.".

type	writesans	string -> o.
special	writesans	elp_output.
doc	writesans
 "``writesans String'' writes the string String to the current output stream
without surrounding double-quotes.  It can therefore not be read again.".

type	writeered	A -> o.
special	writeered	elp_output.
doc	writeered
 "``writeered X'' writes an eta-reduced version of X to the current output
stream.".

type	tab	        int -> o.
special	tab		elp_output.
doc	tab
 "``tab N'' tabs to position N in the current output stream.  If this position
cannot be calculated, or if the cursor is beyond this position, it has no
effect.".

type	fl		o.
special fl		elp_output.
doc	fl
 "``fl'' write a newline character to the current output stream, unless eLP
thinks the cursor is alread at the beginning of a line (``fl'' for
fresh-line).  This is not always accurate, but prevents interference for
example of tracing with normal output.".

type	nl		o.
nl :- writesans "
".
defined nl		elp_output.
doc	nl
 "Writes a newline to the current output stream.".

type	read		(A -> o) -> o.
read G :- reread X \ (!, G X).
read G :- error (writesans "Trying to read from an empty stream."). 
defined read	        elp_input.
doc    read
 "``read (X \ G)'' reads a term M from the current input stream and then
solves the goal ``(X \ G) M''.  The type of M is unified with the type of X,
and an error is signalled if this type-unification fails.  When G fails,
the read also fails.  If the end of the input stream is reached, read
signals an error.  See also ``reread''.".

%% The environment special predicates.
topic   elp_top_level   "Top Level".

type    top	 	o.		     % The top-level.
special	top		elp_top_level.
doc	top
 "Starts a new top level from which only ``pop'' can exit.
This is mainly useful in conjunction with higher-order predicates
and programs that augment the interpreter's context, such as
``switching'', ``=>'', ``==>'', ``pi'', etc.".

type    pop   		o.		     % Exit from a top-level.
special	pop		elp_top_level.
doc	pop
 "Pops to the previous top level, or signals an error if there are none.".

type	bye		o.		     % Exit from lp to lisp.
special	bye		elp_top_level.
doc	bye
 "Exits from eLP.".

%%% Module related predicates.
topic   elp_modules     "Modules".

type	searching	string -> o -> o.    % search directory when loading.
special	searching	elp_modules.
doc	searching
 "``searching Directory G'' adds Directory to the list of directories to
search for files containing modules.  The most recently added directory
is looked in first.".

type	search		string -> o.
search Dir :- searching Dir top.
defined search	elp_modules.
doc     search
 "``search Directory'' starts a new top level on which Directory is searched
for files containing modules.".

type    searches	o.
special searches	elp_modules.
doc	searches
 "Writes a message giving information about which directories are currently
searched for files containing modules.".

type	searching_none	o -> o.		     % only as arg to default!
searching_none G :- error (writesans
 "``searching_none'' is non-logical and destructive.
It can only be used as argument to ``default''.").
const   searching_none	elp_modules.
doc     searching_none
 "``default searching_none'' will destructively remove all directories but the
current one (that is, ^"./^") from the search path.  A goal constructed from
``searching_none'' will result in an error.".

type	load		string -> o.
special	load		elp_modules.
doc	load
 "``load Module'' tries to find a file ^"Module.mod^" in the directories
assumed with ``searching'' (which always contains the current directory).
Gives an error message if no such file can be found.  If the given module
already exists, its clauses are completely erased before the new clauses are
installed.".

type    '==>'	 	string -> o -> o.
special '==>'		elp_modules.
doc	'==>'
 "``Module ==> G'' makes all the clauses in Module visible while trying to
solve G.  It also affects parsing (on read), since constants in M are now
recognized.  M is loaded, if it is not already known, which may lead to a
whole avalanche of loads, in case M imports modules that have not been loaded
either.  Modules can be loaded and made visible by default using a line such
as ``default ('==>' ^"lists^").'' in the initialization file.".

type    use	 	string -> o.
use Module :- Module ==> top.
defined use 		elp_modules.
doc     use
 "``use Module'' starts a new top level in which M's clauses and type
declarations are visible.  See ``'==>'''.".

type    uses	 	o.
special uses		elp_modules.
doc	uses
 "Writes a message with information about which modules are currently visible.".

type	using_none	o -> o.		     % only as arg to default!
using_none G :- error (writesans "``using_none'' is non-logical and destructive.
It can only be used as argument to ``default''.").
const   using_none	elp_modules.
doc	using_none
 "``default using_none'' destructively makes all modules except elp (the
system module) invisible.  A goal constructed from ``using_none'' will
result in an error.".

%%% Grammar and parsing related predicates.
topic	elp_parsing	"Parsing".

type	grammar 	string -> o -> o.    % only as arg to default!
grammar Dir G :- error (writesans "``grammar ^"Dir^"'' is non-logical and destructive.
It can only be used as argument to ``default''.").
const   grammar 	elp_parsing.
doc	grammar
 "``default (grammar ^"Dir^")'' destructively causes all further parsing and
unparsing to be done according to the grammar slterm-gr.txt in directory
^"Dir^".  A goal constructed from ``grammar'' will result in an
error.".

type	parses		o.
special parses		elp_parsing.
doc	parses
 "Writes a message with information about which grammar is currently used
for parsing and unparsing of modules and user input.".

type	standard_grammar  o -> o.  % only as arg to default!
standard_grammar G :-
 error (writesans "``standard_grammar'' is non-logical and destructive.
It can only be used as argument to ``default''.").
const   standard_grammar  elp_parsing.
doc	standard_grammar
 "``default standard_grammar'' destructively causes all further parsing and
unparsing to be done according to the standard grammar for eLP.  A goal
constructed from ``standard_grammar'' will result in an error.".

type	listing_pred	A -> o.
special	listing_pred	elp_interpreter.
doc	listing_pred
 "``listing_pred Pred'' lists all program clauses defining the extension of
Pred in the order in which they would be searched.  Currently invisible
clauses are shown in [ ... ].  Signals an error if applied to a special
predicate.".

type	take		string -> o.
take Goalfile :- 
     reading Goalfile Instream \
      (fl, writesans "Taking goals from file ", write Goalfile, writesans ".",
       seeing Instream (reread G \ (G, fail))).
take Goalfile.

defined take		elp_interpreter.
doc	take
 "``take Goalfile'' will solve all goals in Goalfile.  All solutions to each
goal are generated, but no solutions are printed.  ``(!, fail)'' can be
used to abort taking goals from Goalfile and cause the ``take'' to fail.".

type    default		(o -> o) -> o.	     % Setting a global default.
special default		elp_interpreter.
doc	default
 "``default'' destructively establishes a default for an environment
parameter on the current top-level.  It is applicable only to
  ('==>' Module), (searching Directory), (switching Sw Onoff),
  (prompting String), (grammar Directory), or (depth_bounding Int)
and such constants as
 using_none, searching_none, or standard_grammar.
It is intended primarily for the user's init file.".

%%% Switches.
topic elp_switching     "Switching".

kind	onoff	type.		% A switch value.

type	on	onoff.
const	on	elp_switching.
doc	on
 "The on-value for switches.".

type	off	onoff.
const	off	elp_switching.
doc	off
 "The off-value for switches.".

kind 	sw	type.		% A switch

topic elp_switches      "Switches".

type	printtypes	sw.
const	printtypes	elp_switches.
doc	printtypes
 "If on, the types of variables will be shown during printing.".

type	projfirst	sw.
const	projfirst	elp_switches.
doc	projfirst
 "If on, projections will be tried first during the depth-first search for
a unifier.".

type	normalize	sw.
const	normalize	elp_switches.
doc	normalize
 "If off, terms are not normalized before they are printed.".

type	bounded_unify	sw.
const	bounded_unify	elp_switches.
doc	bounded_unify
 "If on, unification will proceed only to the depth specified by depth_bound.".

type	depth_bounding	int -> o -> o.
special depth_bounding	elp_switches.
doc	depth_bounding
 "``depth_bounding Int G'' sets the depth bound for the unifier to Int while
solving G.  This will only have an effect, if the switch ``bounded_unify'' is
on.".

type	printing_depth	int -> o -> o.
special printing_depth	elp_switches.
doc	printing_depth
 "``printing_depth Int G'' sets the depth bound for printing to Int while
solving G.  A negative value means to print everything.".

type	timing		sw.
const	timing		elp_switches.
doc	timing
 "If on, timing statistics are printed after solving each query.".

type	timing_totals	sw.
const	timing_totals	elp_switches.
doc	timing_totals
 "If on, cumulative timing statistics are printed after solving each query.".

type    work_ahead      sw.
const	work_ahead	elp_switches.
doc	work_ahead
 "If on, the interpreter will work ahead after giving the first solution,
looking for others.  External side-effects (such as opening a file, or writing
output) will suspend the work-ahead, until the user has requested more
solutions.  If present_eager is also on, successive solutions will also be
printed.".

type    present_eager      sw.
const	present_eager	elp_switches.
doc	present_eager
 "If on, and if work_ahead is also on, successive solutions will continue to
be printed even without the user typing ``;''.  However, the user can still type
return to stop the search.".

type	show_undeclared_constants	sw.
const 	show_undeclared_constants	elp_switches.
doc	show_undeclared_constants
 "If off, the interpreter will not show the list of undeclared constants
with their inferred types after a module has been read.".

type	switching       sw -> onoff -> o -> o.
special	switching	elp_switching.
doc	switching
 "``switching Sw Onoff G'' sets switch Sw to Onoff while solving G.".

type	switch		sw -> onoff -> o.    % Global switch setter.
switch Sw Onoff :- switching Sw Onoff top.
defined switch       elp_switching.
doc	switch
 "``switch Sw Onoff'' enters a new top-level in which switch Sw is Onoff.".

type    switches        o.
special switches	elp_switching.
doc	switches
 "Displays information about the current value of switches.".

%%% Tracing options.
kind	trace_opt	type.
topic	elp_traces      "Trace Options".

type	goal		trace_opt.
const	goal		elp_traces.
doc	goal
 "If ``(tracing goal)'' is on, the interpreter will show the current goal
each time a new goal is attempted.".

type	clause		trace_opt.
const	clause		elp_traces.
doc	clause
 "If ``(tracing clause)'' is on, attempts to unify the head of a clause
with the current goal will be traced.".

type	all_goals	trace_opt.
const	all_goals	elp_traces.
doc	all_goals
 "If ``(tracing all_goals)'' is on, all goals (and not just atomic ones)
will be traced.".

type	assumptions	trace_opt.
const	assumptions	elp_traces.
doc	assumptions
 "If ``(tracing assumptions)'' is on, assumptions (from modules and
from implicational goals) and their visibility status will be traced.".

type	io		trace_opt.
const	io		elp_traces.
doc	io
 "If ``(tracing io)'' is on, opening of files for reading or writing will
be traced.".

type	match		trace_opt.
const	match		elp_traces.
doc	match
 "If ``(tracing match)'' is on, messages are printed when MATCH is 
entered and when it selects a substitution term.  MATCH is the only
place where unification can branch.  Tabbing of MATCH is separate
from tabbing in the tracing of goals or clauses.".

type	tracing		trace_opt -> sw.
const	tracing		elp_switches.
doc	tracing
 "``(tracing Opt)'' is a switch that controls whether Opt will be traced.".

type	trace		trace_opt -> o.
trace Opt :- switch (tracing Opt) on.
defined trace		elp_switching.
doc	trace
 "``trace Opt'' starts a new top-level, in which tracing of option Opt is on.".

%%% Warning options.
kind    warn_opt	type.		% Warning options.
topic	elp_warnings    "Warnings".

type	typevar		warn_opt.
const	typevar		elp_warnings.
doc	typevar
 "If ``(warning typevar)'' is off, the unifier will not issue warning messages
when the result type of an argument to MATCH is a type variable.  In such
a situation, the unifier is incomplete --- hence the warning.  This often
signifies that the types in a program or query were underspecified.".

type    type_unify      warn_opt.
const	type_unify	elp_warnings.
doc	type_unify
 "If ``(warning type_unify)'' is off, the unifier will not issue warning
messages when type unification fails.  Although failures in type unification
do not result in unification-incompleteness, they often arise if the type of a
predicate in an imported module has changed, but the calls in the importing
module have not been updated.  Hence, this warning may be useful for catching
such mistakes.  On the other hand, this warning is the one most often given
from actually correct programs and goals.".

type	too_deep	warn_opt.
const	too_deep	elp_warnings.
doc	too_deep
 "If ``(warning too_deep)'' is off, the unifier issues no warning messages
when unification fails because the depth bound is exceeded.".

type	no_clauses	warn_opt.
const	no_clauses	elp_warnings.
doc	no_clauses
 "If ``(warning no_clauses)'' is off, the interpreter issues no warning
messages in case it encounters a predicate at the head of the goal for
which it has no clauses.".

type	no_visible_clauses	warn_opt.
const	no_visible_clauses	elp_warnings.
doc	no_visible_clauses
 "If ``(warning no_visible_clauses)'' is off, the interpreter issues no warning
messages in case it encounters a predicate at the head of the goal for
which all clauses are currently `invisible'.".

type	multiple_decl	warn_opt.
const	multiple_decl	elp_warnings.
doc	multiple_decl
 "If ``(warning multiple_decl)'' is off, the interpreter issues no warning
messages in case it encounters a multiple declaration of a constant in
a file, a declaration after a use of a constants, or a shadowing of a
constant already declared and accessible in an imported module, at the time
a module is read.".

type    warning		warn_opt -> sw.
const   warning		elp_switches.
doc	warning
 "``(warning Wopt)'' is a switch that controls whether the warning Wopt will
be given.".

%%% Help predicates.

type	help		o.
defined help		elp_help.
doc	help
 "Lists all the special and defined predicates and declared constants that
are known to eLP.  They are organized into different topics.  For more
help on a given item X, use ``describe X''.".

type    help_topics     o.
help :- writesans
 "This is eLP, the Ergo project's implementation of lambda Prolog.
  The following help listing contains three lines for each topic.  The first
line contains special predicates, the second defined predicates, the third
declared constants or constructors.",
 "help" ==> help_topics, nl,
 writesans
  "Use ``describe X'' or ``listing X'' for more information on any item X.".

type	describe	A -> o.	
defined describe	elp_help.
doc	describe
 "``describe X'' writes information about X.  X should be a special or
defined predicate, or a declared constant (or constructor).".

% ``describe'' triggers type unification warnings that are suppressed locally
% by switching them off before calling ``describe_item''.

type	describe_item 	A -> o.
describe X :- "help" ==> describe_item X.

type	listing		A -> o.
defined	listing		elp_help.
doc	listing
 "``listing X'' writes information about X, including any clauses that may
define its meaning, in case it is a predicate.  Currently invisible clauses
are shown in [ ... ].".

type	listing_item	A -> o.
listing X :- "help" ==> listing_item X.

type	document	string -> o.
defined document	elp_help.
doc	document
 "``document Outputfile'' writes all available on-line documentation to the
file Outputfile for off-line reference.".

type	document_in_file  string -> o.
document Outputfile :- "help" ==> document_in_file Outputfile.

%%% Then some commonly used constants and "evaluation".

topic	elp_arith	"Arithmetic".

type	'+'	int -> int -> int.
type	'-'	int -> int -> int.
type	'*'	int -> int -> int.
type	'div'	int -> int -> int.
type	'mod'	int -> int -> int.

const	'+'	elp_arith.
doc	'+'
 "``N + M'' is not treated specially by the unifier, that is, 1 + 1 will not
unify with 2, but it can be evaluated using the ``is'' special predicate.
Note that the + operator must be surrounded by whitespace.".

const	'-'	elp_arith.
doc	'-'
 "``N - M'' is not treated specially by the unifier, that is, 1 - 1 will not
unify with 0, but it can be evaluated using the ``is'' special predicate.
Note that the - operator must be surrounded by whitespace.".

const	'*'	elp_arith.
doc	'*'
 "``N * M'' is not treated specially by the unifier, that is, 1 * 1 will not
unify with 1, but it can be evaluated using the ``is'' special predicate.
Note that the * operator must be surrounded by whitespace.".

const	'div'	elp_arith.
doc	'div'
 "``N div M'' is not treated specially by the unifier, that is, 3 div 2 will
not unify with 1, but it can be evaluated using the ``is'' special predicate
(interpreted as integer division, rounding down).".

const	'mod'	elp_arith.
doc	'mod'
 "``N mod M'' is not treated specially by the unifier, that is, 3 mod 2 will
not unify with 1, but it can be evaluated using the ``is'' special predicate.
Note that the mod operator must be surrounded by whitespace.".

type	'is'	int -> int -> o.
special 'is'	elp_arith.
doc	'is'
 "``X is Y'' succeeds by when X is (unifies with) the term encoding the value
of the expression represented by Y (which must not have free variables).
The interpreter currently only allows the evaluation of integer expressions
with operations +, -, *, div, and mod.".

type	'=:='	int -> int -> o.
special '=:='	elp_arith.
doc	'=:='
 "``X =^^:= Y'' succeeds when both X and Y are integers, and X is equal
to Y.  ':' is a lexical operator in the standard grammar and must
therefore be quoted with '^^' unless already inside quotes.".

type	'=\='	int -> int -> o.
special '=\='	elp_arith.
doc	'=\='
 "``X =^^\= Y'' succeeds when both X and Y are integers, and X is not equal
to Y.  `\' is a lexical operator in the standard grammar and must
therefore be quoted with '^^' unless already inside quotes.".

type	'<'	int -> int -> o.
special '<'	elp_arith.
doc	'<'
 "``X < Y'' succeeds when both X and Y are integers, and X is less than Y.".

type	'>'	int -> int -> o.
special '>'	elp_arith.
doc	'>'
 "``X > Y'' succeeds when both X and Y are integers, and X is greater than Y.".

type	'=<'	int -> int -> o.
special '=<'	elp_arith.
doc	'=<'
 "``X =< Y'' succeeds when both X and Y are integers, and X is less than or
equal to Y.".

type	'>='	int -> int -> o.
special '>='	elp_arith.
doc	'>='
 "``X >= Y'' succeeds when both X and Y are integers, and X is greater than
or equal to Y.".


%% Constructors for lists.  No built in utilities right now.

topic	elp_lists	"Lists".

kind	list	type -> type.

type	nil	list A.
type	'::'	A -> list A -> list A.

const	nil	elp_lists.
doc	nil
 "``nil'' stands (by convention) for the empty list.  It is not interpreted
specially, but used by the lists library module.".

const	'::'	elp_lists.
doc	'::'
 "``X :: L'' stands (by convention) for a list with head X and tail L.  '::'
is not treated specially by the interpreter, but used by the lists library
module.  It associates to the right.  For example, ``1::2::3::4::nil'' is
the list of the integers from 1 through 4.".
