;;; -*- Mode: Lisp; Package: IDENT -*-
;;; 
;;; This ADT allows objects to be associated with unique identifiers.  The
;;; identifiers are constructed using a given name, which will be used as a
;;; prefix.
;;;
;;; Author: Conal Elliott.  Last Modified Mon Sep 11 06:08:46 1989
;;;
;;; Sccs Id @(#)ident.lisp	1.20 9/15/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :IDENT) (use-package :ergolisp)

(export '(Ident-string
	  shadowing-ident
	  ident-init new-ident-avoiding new-ident
	  ident-retrieve ident-retrieve-check))

(export '(string-to-ident ident-to-string))
(export '(shadow-ident set-ident unset-ident))

(deftype Ident-string ()
  "A string that is associated with an object via the ident facility."
  `String)

(defvar *ident-table* (make-hash-table :test #'EQ)
  "Hash table mapping an interned ident string to a saved objects.")

(defvar *next-num-table* (make-hash-table :test #'EQ)
  "Hash table mapping from an interned prefix string to the next
available suffix number.")

(defun string-to-ident (str)
  "Coerces a name string STR into an ident."
  (intern str :IDENT))

(defun ident-to-string (ident)
  "Returns the name of an ident as a string."
  (symbol-name ident))

(defun shadow-ident (str object)
  "Make a shadowing Ident-string given STR and OBJECT.  Also stores the
association.  The OBJECT can be retrieved with the function ident-retrieve.
The shadowed object is returned as a second value, or NIL, if nothing
was shadowed."
  (declare-ftype shadow-ident (String T) Ident-string T)
  (let* ((ident (string-to-ident str))
	 (shadowed-object (gethash ident *ident-table*)))
    (setf (gethash ident *ident-table*) object)
    (values ident shadowed-object)))

(defun set-ident (ident object)
  "Destructively change the identifier association for IDENT to OBJECT."
  (setf (gethash ident *ident-table*) object))

(defun unset-ident (ident)
  (set-ident ident nil))

(defmacro shadowing-ident (&rest body)
  "Shadow the ident records while evaluating body.  Used, e.g., when loading
a module."
  `(let ((*ident-table* (make-hash-table :test #'EQ))
	 (*next-num-table* (make-hash-table :test #'EQ)))
     ,@body))

(defun ident-init ()
  "Clear out all ident associations.  This should be done with care since it
renders currently ident'ed variables unnameable."
  (clrhash *ident-table*)
  (clrhash *next-num-table*))

(defun coerce-first-char-case (string name-case)
  (concatenate 'string
    (make-string 1 :initial-element
		 (ecase name-case
		   (:uppercase (char-upcase (schar string 0)))
		   (:lowercase (char-downcase (schar string 0)))))
    (subseq string 1)))

(defun next-num (str)
  "Get the next free suffix number for STR or nil if this string has not been
used before.  Remember that it is being used."
  (declare-ftype next-num (String) (Optional Integer))
  (let* ((interned (intern str :IDENT))
	 (next (gethash interned *next-num-table*)))
    (setf (gethash interned *next-num-table*)
	  (if next (1+ next) 1))
    next))

(defun new-ident (str object name-case)
  "Like new-ident-avoiding, but doesn't take an avoid predicate.  This function
only exists for backward compatibility.  If you are using it, it might be a
bug."
  (declare-ftype new-ident (String T) Ident-string)
  (new-ident-avoiding str object 
		      #'(lambda (x) (declare (ignore x)) nil)
		      name-case))

(defun new-ident-avoiding (str object avoid-pred name-case)
  "Make a new Ident-string given STR and OBJECT, avoiding names that satisfy
AVOID-PRED.  Also stores the association.  The OBJECT can be retrieved with the
function ident-retrieve."
  (declare-ftype new-ident-avoiding 
    (String T (Function-of (String) Boolean)) Ident-String)
  (when name-case
    (setq str (coerce-first-char-case str name-case)))
  (let* ((suffix-num (next-num str))
	 (ident-string (if suffix-num
			   (format nil "~a~d" str suffix-num)
			   str)))
    ;; Check that the name passes ok-pred.  Also,
    ;; it is still possible that the ident-string chosen is already in use,
    ;; e.g., if the prefix "A1" is in use and we are adding a suffix of 1 to
    ;; the name "A".
    (if (or (funcall avoid-pred ident-string)
	    (gethash (intern ident-string :IDENT) *ident-table*))
	;; already have done coercion - using NIL as case
	(new-ident-avoiding str object avoid-pred nil)
	(progn
	  (setf (gethash (intern ident-string :IDENT) *ident-table*) object)
	  ident-string))))

(defun ident-retrieve (ident-string)
  "Retrieve an object associated with IDENT-STRING, or return nil if not there."
  (declare-ftype ident-retrieve (Ident-string) (Optional T))
  (gethash (intern ident-string :IDENT) *ident-table*))


(defun ident-retrieve-check (ident-string type)
  "Like ident-retrieve, but make sure that the object associated with
IDENT-STRING has the given Lisp TYPE.  Gives an error if found but of the
wrong type.  Perhaps it should just return nil instead.  Hmmm."
  (declare-ftype ident-retrieve-check (Ident-string T) (Optional T))
  (let ((object (ident-retrieve ident-string)))
    (if object
	 (if (typep object type)
	     object
	     (error "The object, ~s, associated with ident-string ~s is not of type ~s"
		    object ident-string type))
	 nil)))
