;;; -*- Mode: Lisp; Package: ELP -*-
;;;
;;; Evaluation of expressions, for the IS special predicate.
;;;
;;; Author: Conal Elliott.  Last Modified Tue Sep 12 15:37:45 1989
;;; Author: Conal Elliott.  Last Modified Mon Sep 11 06:08:44 1989
;;;
;;; Sccs Id @(#)lp-evaluate.lisp	1.22 9/12/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :ELP) (use-package :ergolisp)
(export '(*exp-name-alist*))

(defsconstr "

    DATATYPE Exp =
        Ebasic of Number		% Basic value
      |	Ecall of Symbol, Function, Exp*	% Function call
      | Eflex of Term			% Flexible-headed Exp term
")
;;; Note: In Ebasic and Ecall, we store the originally given name, as well as
;;; the corresponding lisp symbol.


;;; Alist to convert eLP names into Lisp names.
(defparameter *exp-name-alist*
  '((|+| . +)
    (|-| . -)
    (|*| . *)
    (|div| . floor)
    (|mod| . mod))
  "Alist mapping elp constant names to lisp symbols.")

(defun copy-exp (exp type-args term-args depth)
  "Copy EXP, filling in TYPE-ARGS and TERM-ARGS at given DEPTH."
  (labels ((copy-rec (exp)
	     (dcase exp
	       (Ebasic
		exp)
	       ((Ecall name function exp-args)
		(re-Ecall exp name function exp-args 
			  name function
			  (re-mapcar #'copy-rec exp-args)))
	       ((Eflex term)
		(Eflex (tlang-call copy-term
				   term type-args term-args depth))))))
    (copy-rec exp)))


(defreptype rexp Exp
  "Representing an lterm as an expression."
  :repfun #'(lambda (term context)
	      (declare (ignore context))
	      (tlang-call term-to-exp term))
  :unrepfun #'(lambda (exp vtypes tvnames vnames depth)
		(tlang-call exp-to-sterm exp vtypes tvnames vnames depth))
  :unrepfun #'(lambda (exp vtypes tvnames vnames depth)
		(tlang-call exp-to-sterm exp vtypes tvnames vnames depth))
  :copyfun #'copy-exp)

(defspecial |is| (rterm rexp) nil
  ( (list unif-term exp) con _ sc)
  "``X is Y'' succeeds by when X is (unifies with) the term encoding the value
of the expression represented by Y (which must not have free variables)."
  (tlang-call unify unif-term (tlang-call value-term (evaluate exp))
	      con sc))

(defun evaluate (top-exp)
  "Evaluate TOP-EXP, for use with the IS special predicate."
  (declare-ftype evaluate (Exp) T)
  (labels ((ev-rec (exp)
	     (dcase exp
	       ((Ebasic object)
		object)
	       ((Ecall _ function exp-args)
		(apply function
		       (mapcar #'ev-rec exp-args)))
	       ((Eflex term)
		(let ((retry-exp (tlang-call term-to-exp term)))
		  (if (is-Eflex retry-exp)
		      (lp-error "Cannot evaluate ~s.  Contains flexibly ~
headed term ~s."
				(tlang-call exp-to-sterm top-exp () () () 0)
				(tlang-call exp-to-sterm top-exp () () () 0)
				term))
		      (ev-rec retry-exp))))))
    (ev-rec top-exp)))


;;; Integer comparisons as special predicates a la C-Prolog.

(defmacro defbinop (elp-name lisp-fun-name)
  "Defines a special predicate ELP-NAME that succeeds if the binary predicate
LISP-FUN-NAME holds of the arguments."
  `(defspecial ,elp-name (rterm rterm) nil
     ( (list left right) con _ sc )
     "A binary special predicate comparing integers."
     (let ((num1 (term-number-value left))
	   (num2 (term-number-value right)))
       (declare (type integer num1 num2))
       (if (,lisp-fun-name num1 num2)
	   (sc-invoke sc con)
	   (values)))))

(defbinop |<| <)
(defbinop |>| >)
(defbinop |=<| <=)
(defbinop |>=| >=)
(defbinop |=:=| =)
(defbinop |=\\=| /=)
