;;; -*- Mode: Lisp; Package: ELP -*-
;;;
;;; Process a smodule, generating a new module
;;;
;;; Author: Conal Elliott.  Last Modified Mon Sep 11 06:08:40 1989
;;;
;;; Sccs Id @(#)lp-from-smodule.lisp	1.34 9/15/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :ELP) (use-package :ergolisp)

(export '(*warn-multiple-decl*))

(export '(*show-undeclared-constants*))

(defvar *warn-multiple-decl* t
  "Controls whether to warn when is constant is declared more than once
in a module.")

(defvar *show-undeclared-constants* t
  "Controls whether to show the types inferred for constants.")

(ddefun from-smodule ( (Smodule name imports entries) )
  "Process SMODULE, using term language specification *current-tlang*,
generating a new module.  If it already exists, the information in the
module is cleared out first."
  (declare-ftype from-smodule (Smodule) Module)
  ;; If we are reloading, get the old module and call its delete-fun to
  ;; get rid of old program clauses.  Otherwise, make a new module with the
  ;; given name and with initialized signatures.
  ;; Shadowing-ident should now only affect printing (say, of error or
  ;; warning messages), but not the parsing.  Perhaps eliminate.
  ;; (ident:shadowing-ident
   ;; Make a temporary module structure.  In case there is an error,
   ;; this will prevent problems.  Also, if there are circular dependencies
   ;; (heaven beware!), other references still point to the old module.
   (dlet ( ((:as module (Module _ _ _ kindsig typesig _))
	    (new-temp-module name imports)) )
     (let ((sig-modules (accessible-modules module))
	   (sdformlist ())
	   (undeclared-type-constants-list ()))
       (format *elp-trace-output* "~&Inferring types on module ~A." name)
       (dolist (entry entries)
	 (dcase entry
	   ((Mkinding type-name skind)
	    (dlet (( (values previous-decl previous-module)
		     (modules-sig-get type-name sig-modules #'module-kindsig) ))
	      (when previous-decl
		(if (eq previous-module module)
		    (lp-warn *warn-multiple-decl*
			     "Multiple kind declaration or declaration after use for ``~A'' in module ~A.
 Previously ~S, now ~S." type-name (module-name module) previous-decl skind)
		    (lp-warn *warn-multiple-decl*
			     "Multiple kind declaration of ``~A''.
 Previously ~S in module ~A, now ~S in module ~A."
			     type-name previous-decl (module-name previous-module)
			     skind (module-name module))))
	      (setf (sig-get type-name kindsig)
		    (tlang-call convert-skind skind))))
	   ((Mtyping const-name stype)
	    ;; Warn if redefinition.
	    (dlet (( (values previous-decl previous-module)
		     (modules-sig-get const-name sig-modules #'module-typesig) ))
	      (when previous-decl
		(if (eq previous-module module)
		    (lp-warn *warn-multiple-decl*
			     "Multiple type declaration for ``~A'' in module ~A.
 Previously ~S, now ~S." const-name (module-name module) previous-decl stype)
		    (lp-warn *warn-multiple-decl*
			     "Multiple type declaration of ``~A''.
 Previously ~S in module ~A, now ~S in module ~A."
			     const-name previous-decl (module-name previous-module)
			     stype (module-name module))))
	      (dlet ( ((values tyfam undeclared-tcs)
		       (tlang-call convert-stype stype sig-modules)) )
		;; Now enter the undeclared type constants.
		;; That's OK since a type constant occurrence uniquely
		;; determines it's kind.
		(dolist (udtc undeclared-tcs)
		  (dlet ( ((cons tconst-name kind) udtc) )
		    (setf (sig-get tconst-name kindsig) kind)))
		(push undeclared-tcs undeclared-type-constants-list)
		(setf (sig-get const-name typesig) tyfam))))
	   ((Mindexing const-name n)
	    (index-pred const-name n))
	   ((Msdform sdform)
	    (push sdform sdformlist))))
       ;; Now we convert the list of sdforms all at once.
       (dlet (( (values adformlist
			undeclared-constants undeclared-type-constants)
		(tlang-call convert-sdformlist (nreverse sdformlist)
			    sig-modules) ))
	 (push undeclared-type-constants undeclared-type-constants-list)
	 (setq undeclared-type-constants
	       (nreverse (apply #'append undeclared-type-constants-list)))
	 (when (and undeclared-type-constants *show-undeclared-constants*)
	   (format *elp-trace-output*
		   "~&Kinds inferred for undeclared types and type constructors:"))
	 (dolist (udtc undeclared-type-constants)
	   (dlet ( ((cons tconst-name kind) udtc) )
	     (setf (sig-get tconst-name kindsig) kind)
	     (when *show-undeclared-constants*
	       (format *elp-trace-output* "~&kind ~8T~A ~24T~A."
		       tconst-name kind))))
	 (when (and undeclared-constants *show-undeclared-constants*)
	   (format *elp-trace-output*
		   "~&Types inferred for undeclared constants:"))
	 (dolist (udc undeclared-constants)
	   ;; These are guaranteed to be new.
	   (dlet ( ((cons const-name const-tyfam) udc) )
	     (setf (sig-get const-name typesig) const-tyfam)
	     (when *show-undeclared-constants*
	       (format *elp-trace-output* "~&type ~8T~A ~24T~A."
		       const-name const-tyfam))))
	 ;; All possible errors should have happened up to now.
	 ;; If things go wrong in here, the state of the world will be
	 ;; inconsistent.
	 (let ((no-error-occurred nil))
	   (dlet ((existing-module (named-module name)))
	     (unwind-protect
		 (progn
		   (if existing-module
		       (progn
			 (funcall (module-delete-clause-fun existing-module))
			 (setf (module-kindsig existing-module) kindsig)
			 (setf (module-typesig existing-module) typesig)
			 (setf (module-imports existing-module)
			       (module-imports module)))
		       (setq existing-module (install-module module)))
		   (let ((pentries ()))
		     (dolist (adform adformlist)
		       (dlet (( (list tvnames vnames vtypes dform) 
				adform ))
			 (setq pentries
			       (nunion pentries
				       (install-module-dform dform tvnames vnames vtypes
							     existing-module)))))
		     (setf (module-delete-clause-fun existing-module)
			   #'(lambda ()
			       (dolist (pentry pentries)
				 (delete-module-clauses existing-module pentry))))
		     (setq no-error-occurred t)
		     existing-module))
	       (when (not no-error-occurred)
		 (lp-warn t "Unexpected error occurred while installing clauses in a module.
The system may no be an inconsistent internal state.
``listing'' may be of help in determining if this is the case."))
	       existing-module))))))
   ;;)
   )

