;;; -*- Mode: Lisp; Package: ELP -*-
;;;
;;; Definition of the term language interface, plus everything else that a
;;; term language needs to know from eLP (I think).
;;;
;;; Author: Conal Elliott.  Last Modified Sun Sep 17 20:58:28 1989
;;;
;;; Sccs Id @(#)lp-interface.lisp	1.52 9/17/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :ELP) (use-package :ergolisp)

(export '(Kind Typ Term Predicate Evar Tevar Constraint))

(export '(*current-tlang* tlang-call))

(export '(Trail *trail* trailing))
(export '(Scont sc-invoke sc-make))

(export '(lp-error lp-trace lp-warn))

(export '(*elp-standard-input* *elp-standard-output*
			       *elp-trace-output* *elp-terminal-io*))

(export '(*working-ahead*))

(export '(*max-trace-indent*))

(export 'elp-load)

(defvar *elp-standard-input* *standard-input*
  "The stream used for ELP input.")

(defvar *elp-standard-output* *standard-output*
  "The stream used for ELP standard output.")

(defvar *elp-trace-output* *trace-output*
  "The stream used for ELP trace output.")

(defvar *elp-terminal-io* *terminal-io*
  "The stream used for terminal io.")

(defvar *working-ahead* nil
  "A variable that keeps track of whether the system is working ahead after
presenting a solutions.")

(defparameter *max-trace-indent* 40
  "If tracing indents modulo this number to avoid extremely deep indenting.")

;;; Types.  I think that these should also be explicitly used somehow.

(deftype Skind () "Term language syntactic Kind" T)
(deftype Styp () "Term language syntactic Type" T)
(deftype Sterm () "Term language syntactic Term" T)

(deftype Kind () "Term language Kind" T)
(deftype Typ () "Term language Type" T)
(deftype Term () "Term language Term" T)
(deftype Predicate () "Term language Predicate" T)

(deftype Ctyp () "Term language Type in context" T)
(deftype Cterm () "Term language Term in context" T)

(deftype Evar () "Term language existential variable" T)
(deftype Tevar () "Term language type existential variable" T)

(deftype Constraint () "Term language logic programming constraint" T)

(defsconstr "

  DATATYPE Tlang of
    %% Term language interface.
    %%  First the types.
    Kind : T ,				% Kinds
    Typ : T ,				% Types
    Term : T ,				% Terms
    Predicate : T ,			% Predicates
    Tevar : T ,				% Type existential variables
    Evar : T ,				% Term existential variables
    Constraint : T ,			% Constraints
    Ctyp : T ,
    Cterm : T ,
    %%  Then the functions.
    %% Convert a syntactic kind to a real one.
    convert-skind : Skind -> Kind ,
    %% Convert a syntactic type to a real one.
    convert-stype : Styp # Module -> Typ ,
    %% Convert a syntactic term clause into a Dform.
    convert-sdform : Sterm # Module -> Dform # Symbol* # Symbol* # Typ*,
    %% Convert a list of term clauses into a list of dforms.
    convert-sdformlist : Sterm* # Module -> T # T,
    %% really something like:
    %% Sterm* # Module -> (Symbol* # Symbol* # Typ* # Dform)* # (Symbol # Typ)*,
    %% Convert a syntactic term query into a Gform.
    convert-squery : Sterm -> Gform # Tevar* # Evar*,
    %% Find the type of a term.
    typ-of : Term -> Typ ,
    %% Make a new tevar
    new-tevar : Symbol -> Tevar ,
    %% Make a new evar
    new-evar : Symbol # Typ -> Evar ,
    %% Make a new uvar
    new-uvar : Symbol # Typ -> Evar ,
    %% Make a list of typed variables to be printed.
    mk-ctyp : Symbol* # Symbol* # Typ* # Typ -> Ctyp ,
    %% Make a cterm out of type and term variable names, term variable types,
    %% and a term.
    mk-cterm : Symbol* # Symbol* # Typ* # Term -> Cterm ,
    %% Copy a type, given the new tevars to use for its loose tbvars.
    copy-typ : Typ # Tevar* -> Typ ,
    %% Copy a term, given the new tevars and evars to use for its loose
    %% bvars and tbvars, and given the depth.
    copy-term : Term # Tevar* # Evar* # Nat -> Term ,
    %% Unbind an existential variable.
    unbind-var : Evar or Tevar -> Unit ,
    %% Convert a term into a G formula.
    term-to-gform : Term # Typ* -> Gform ,
    %% Convert a term into a D formula.
    term-to-dform : Term # Typ* -> Dform ,
    %% Convert a gform with its term variable name context to a sterm.
    gform-to-sterm : Gform # Symbol* # Symbol* # Nat -> Sterm ,
    %% Convert a dform with its term variable name context to a sterm.
    dform-to-sterm : Dform # Symbol* # Symbol* # Nat -> Sterm ,
    %% Unify two terms, using the given constraint and success continuation.
    unify : Term # Term # Constraint # Scont -> Unit,
    %% Unify the arguments of a clause with the arguments of rigid atomic
    %% goal (respectively), given the new tevars and evars, and constraint
    %% and success continuation.  Hopefully copying will be postponed until
    %% after rigid-rigid comparisons.
    clause-head-unify : Term* # Term* # Tevar* # Evar* # Constraint # Scont ->
			  Unit,
    %% The empty constraint
    empty-constraint : Constraint ,
    %% Read a term of the given type.  (Do type unification if appropriate).
    read-term : Typ -> Term ,
    %% Present a solution.
    present-solution : Evar* # Constraint -> Unit,
    %% Get the `value' of a constant term as a Lisp object.
    term-value : Term -> T,
    term-string-value : Term -> String,
    %% Convert an internal term to syntactic form
    to-sterm : Term # Symbol* # Symbol* # Nat -> Sterm,
    %% parse a period-terminated syntactic term from a file or stream.
    parse-sterm : Stream -> Sterm,
    %% parse a syntactic module from a file or stream.
    parse-smodule : Stream -> Smodule,
    %% unparse a period-terminated syntactic term to a file or stream
    unparse-sterm : Sterm # Stream -> Unit,
    %% for loading a term language
    load-sterm-lang : String -> Unit,
    %% like term-to-gform, but taking an abstraction to a glam
    absts-to-gform : Nat # Term # Typ* -> Gform,
    %% Convert a Term into an Exp (for later evaluation)
    term-to-exp : Term -> Exp,
    %% Convert an Exp to a syntactic term.
    exp-to-sterm : Exp # Ltype* # Symbol* # Symbol* # Nat -> Sterm,
    %% Convert a Lisp value into a term    
    value-term : T -> Term,
    %% Eta-reduce for improved readability
    head-eta-reduce : Term -> Term,
    %% Converting a term into a pred
    term-to-pred : Term -> Pred,
    %% Anti-unifying types
    typ-anti-unify : Typ* -> Typ,
    %% For indexing
    term-nth-arg-head-value : Nat # Term -> T
")

(Defvar *current-tlang* 'current-tlang-unbound
  "The current Tlang.")

(defmacro tlang-value (field)
  "Get the value of a tlang field"
  `(,(intern (concatenate 'string "TLANG-" (symbol-name field)) :ELP)
    *current-tlang*))

(defmacro tlang-call (fun &rest args)
  "Call a tlang function."
  `(funcall (tlang-value ,fun) ,@args))



;;;;	Trail stack.  Needed by unifiers.

(deftype Trail ()
  `(List-of (or Tevar Evar)))

(defvar *trail* ()
  "The Lambda-Prolog trail stack.  Of type Trail")

(defmacro trailing (&rest forms)
  "Evaluate FORMS.  If the last one returns (which signals failure), the trail 
is unwound unwinding tevars and evars.  Also unwinds on throws."
  (let ((trail-was-var (gensym)))
    `(let ((,trail-was-var *trail*))
       (unwind-protect
	   (progn ,@forms)
	 (unwind-trail ,trail-was-var)))))

(defun unwind-trail (trail-was)
  "Unwind the trail back to TRAIL-WAS, uninstantiating variables as we go."
  (declare-ftype unwind-trail (Trail) )
  ;; There ought to be macros "dountil" and/or "dowhile" for this.
  (do ()			
      ((eq *trail* trail-was) (values))
    (tlang-call unbind-var (pop *trail*))))
       

;;;;	Scont:  Success continuations.

(deftype Scont ()
  "The type of an LP success continuation."
  ;;  Simply a function that accepts a Constraint.
  `(Function-of (Constraint)))

(proclaim '(inline sc-invoke sc-make))

(defun sc-invoke (scont con)
  "Invoke a success continuation SCONT on a constraint CON."
  (declare-ftype sc-invoke (Scont Constraint) )
  (funcall scont con))

(defun sc-make (fun)
  "Make a scont out of FUN, a function taking a constraint."
  (declare-ftype sc-make ((Function-of (Constraint))) Scont)
  fun)


;;;;	Error reporting and tracing

(defun query-success ()
  (setq *working-ahead* nil)
  (throw 'query-success t))

(defun lp-check-for-more ()
  "If we are working ahead, wait for user to say whether wants more solutions."
  (when *working-ahead*
    (if (more-answers-p)
	(setq *working-ahead* nil)
	(query-success))))

(defun lp-error (format-string &rest args)
  "Report an error to the Lambda-Prolog user, given FORMAT-STRING and rest 
ARGS.  Throws a formatted string to the catch label 'lp-error."
  (lp-check-for-more)
  (format *elp-terminal-io* "~&ELP error: ~?~%" format-string args)
  (throw 'lp-error (values)))

(defmacro catch-lp-error (&rest body)
  "Catch an elp error in BODY."
  `(catch 'lp-error ,@body))

(defun lp-warn-format (format-string &rest args)
  "Print a warning message.  Should not be called from programs."
  (lp-check-for-more)
  (format *elp-terminal-io* "~&ELP warning: ")
  (apply #'format *elp-terminal-io* format-string args)
  (values))

(defmacro lp-warn (flag format-string &rest args)
  "Generates a warning to the Lambda-Prolog user, given FORMAT-STRING and
rest ARGS, if flag is true."
  `(when ,flag
     (lp-warn-format ,format-string ,@args)))

#|
(defvar *trace-stop-string* nil
  "If this variable is non-nil, the ELP tracer will enter the debugger when it
prints a string that is string-equal to this one.  This string is checked
even when tracing is turned off.  This variable is set by the special
predicates break and unbreak.")
|#

(defun lp-trace-format (indent-width format-string &rest args)
  "Print the actual tracing message.  Should not be called from programs.
Use the lp-trace macro instead."
  (lp-check-for-more)			; Problem!!!
  (format *elp-trace-output* "~&~vt" (mod indent-width *max-trace-indent*))
  (apply #'format *elp-trace-output* format-string args)
  (values))

(defmacro lp-trace (flag indent-width format-string &rest args)
  "Do Lambda-Prolog tracing if FLAG is set, indenting by INDENT-WIDTH, using
given FORMAT-STRING and rest ARGS.  A macro to avoid evaluating arguments in
case FLAG is false."
  `(when ,flag
     (lp-trace-format ,indent-width ,format-string ,@args)))

(defun lp-trace-unwind-format (indent-width format-string &rest args)
  (let ((more-requested  (if *working-ahead* (more-answers-p) t)))
    (setq *working-ahead* nil)
    (format *elp-trace-output* "~&~vT" (mod indent-width *max-trace-indent*))
    (apply #'format *elp-trace-output* format-string args)
    (if more-requested (values) (query-success))))

(defmacro lp-trace-unwind (flag indent-width format-string &rest args)
  "As lp-trace, except at this traces unwinding, which includes
a throw back to the top level.  Thus, we must show the trace,
even if no further solutions were requested."
  `(when ,flag
     (lp-trace-unwind-format ,indent-width ,format-string ,@args)))

(defun elp-load (file)
  "Load a Lambda-Prolog module from FILE."
  (not (catch-lp-error
	(load-module-from-file (file-namestring (pathname-name file)) file))))

(defun box-module-load (&key file &allow-other-keys)
  "Loader function used by the box system."
  (elp-load file))
