;;; -*- Mode: Lisp; Package: ELP -*-
;;;
;;; Lambda-prolog Module System
;;;
;;; Author: Conal Elliott.  Last Modified Mon Sep 11 06:08:39 1989
;;;
;;; Sccs Id @(#)lp-module.lisp	1.32 9/15/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :ELP) (use-package :ergolisp)

(export '(Signature new-sig sig-get))

(export '(Signature new-sig sig-get))

(export '(Module *all-modules*
		 new-module named-module named-module-or-load
		 modules-sig-get
		 currently-accessible-modules accessible-modules))

(defconstant *elp-module-name* "elp"
  "The name of the basic ELP module, containing such things as the types of
the logical constants.")


;;;	Vismarker.  I don't know where this really belongs:

(defsconstr "

    DATATYPE Vismarker of
      %% Visibility marker object
      visiblep : Ref Boolean
")


;;; If Tv works out, put it in ergo-misc.lisp

(deftype Tv (var)
  "A type variable with the given name.  For documentation only.  For instance,
we could say \"(Function-of ((Tv A) (List-of (Tv A))) (List-of (Tv A)))\"
for the type of cons. "
  (declare (ignore var))
  `T)


;;;;	Signature

(deftype Signature (type)
  "Lterm kind signatures."
  `(Hash-table-of Symbol ,type #'EQ))


(defun new-sig (&optional (size 20))
  "Construct a new signature."
  (declare-ftype new-sig () (Signature T))
  (make-hash-table :test #'EQ :size size))


(defun sig-get (name sig)
  "Get the object associated with NAME in SIG.  This function is setfable,
in the same style as gethash."
  (declare-ftype sig-get (Symbol (Signature (Tv A))) (Optional (Tv A)))
  (values (gethash name sig)))

(defun sig-set (name sig obj)
  "setf method for sig-get"
  (setf (gethash name sig) obj))

(defsetf sig-get sig-set)


;;;;	Module

(defsconstr "

    DATATYPE Module of
      name : String ,
      vismarker : Vismarker ,
      imports : Module* ,
      kindsig : Signature T ,
      typesig : Signature T ,
      delete-clause-fun : Unit -> Unit	% Function to call on reloading

")

(ddefun print-module ( (Module name (Vismarker visiblep) imports _ _ _)
		       stream _ )
  (format stream "#<Module ~a ~a ~a>"
	  name visiblep
	  (mapcar #'module-name imports)))
	  

(defvar *all-modules* ()
  "List of all known lp modules.")

#|  Decide what to do with this.

(defvar *current-module* 'current-module-unbound
  "The current module, as used, e.g., for augmenting a current signature.")
|#

(defvar *module-table* (make-hash-table :test #'EQ)
  ;; : (Hash-table-of Symbol Module #'EQ)
  "Hash table mapping interned names to modules.")

(defun module-init ()
  "Reinitialize the lp module system."
  (clrhash *module-table*)
  (setq *all-modules* ())
  )

;;; Names are interned into the ELP package.  Perhaps there should be a
;;; separate package for this purpose.

(defun named-module (name)
  "Get the module having the given NAME (string), or return nil if there is not
one.  Setfable."
  (declare-ftype named-module (String) (Optional Module))
  (values (gethash (intern name :ELP) *module-table*)))

(defun set-named-module (name module)
  "Setf method for named-module"
  (setf (gethash (intern name :ELP) *module-table*) module))

(defsetf named-module set-named-module)

(defun new-module (name import-names)
  "Like new-temp-module, but, as a side-effect, it adds the new module to
*all-modules* and makes it accessible via the function 'named-module'."
  (when (named-module name)
    (lp-error "A module named ~a already exists." name))
  (let ((module (new-temp-module name import-names)))
    (push module *all-modules*)
    (setf (named-module name) module)
    module))

(defun install-module (module)
  (push module *all-modules*)
  (setf (named-module (module-name module)) module)
  module)

(defun new-temp-module (name import-names)
  "Construct a new module having the given NAME and IMPORT-NAMES.  Returns the
new module, with the visibility marker off (except when NAME is
*elp-module-name*).  Signals an error if there already is a module of the given
name, or if any module in IMPORT-NAMES cannot be found. 
The kindsig, typesig, and delete-clause-fun are initialized to
be empty and can be augmented using setf."
  (declare-ftype new-module (String (List-of String)) Module)
  (let* ((is-sys-mod (string= name *elp-module-name*))
	 (new-module (Module name
			    (Vismarker is-sys-mod)
			    (mapcar #'named-module-or-load import-names)
			    (new-sig)
			    (new-sig)
			    #'(lambda ()) )))
    (if is-sys-mod
	(let ((old-sys-mod (named-module *elp-module-name*)))
	  (when old-sys-mod
	    ;; If the system module already exists, get rid of it.
	    (funcall (module-delete-clause-fun old-sys-mod))
	    (setq *all-modules* (delete old-sys-mod *all-modules*)))
	  (install-module new-module))
	new-module)))

(defun named-module-or-bust (name)
  "Like named-module, but signals an error if the module is not present."
  (declare-ftype named-module-or-bust (String) Module)
  (or (named-module name)
      (lp-error "Module ~s not found" name)))

(defun named-module-or-init (name)
  "Like named-module, but creates and empty module if not present.
Is used to allow circularities, such as with the system module."
  (declare-ftype named-module-or-init (String) Module)
  (or (named-module name)
      (new-temp-module name ())))

(defun named-module-or-load (name)
  "Like named-module, but signals an error if the module is not present."
  (declare-ftype named-module-or-load (String) Module)
  (or (named-module name)
      (load-module name)))

(defun load-module (module-name)
  "Loads are reloads a module.  Signals an error if the module cannot be
found in the current search path for modules.
Searching the path for a module may be very time-consuming!"
  (let ((module-file (find-module-file module-name)))
    (load-module-from-file module-name module-file)))

(defun load-module-from-file (module-name module-file)
  "Loads a module (supposedly) MODULE-NAME from file MODULE-FILE and issues
a warning if there seems to be an inconsistency."
  (let ((smodule (with-open-file (instream module-file
					   :direction :input)
		   (format *elp-trace-output* "~&Loading module ~A from file ~S."
			   module-name module-file)
		   (tlang-call parse-smodule instream))))
    (when (not (string= module-name (Smodule-name smodule)))
      (lp-warn t "Module name ``~A'' declared at beginning of module does not
agree with filename ~S.  Using the information inside the file."
	       (Smodule-name smodule) module-name))
    (from-smodule smodule)))

;;; Looking up and setting signature elements.

(defun modules-sig-get (name modules sig-fun)
  "Get the signature entry associated with NAME in MODULES, using SIG-FUN to
extract signatures from modules.  Does not follow import chains.  Return nil if
not found SIG-FUN should be either #'module-kindsig or #'module-typesig.  First
checks the module named by *elp-module-name*, which needn't, therefore, be in
MODULES."
  (declare-ftype module-sig-get
    (Symbol (List-of Module)
	    (Function-of Module (Signature (Tv A))))
    (Optional (Tv A)) (Optional Module))
  (labels ((get-rec (modules-left)
	     (dcase modules-left
	       ((cons first rest)
		(let ((entry (sig-get name (funcall sig-fun first))))
		  (if entry (values entry first) (get-rec rest))))
	       (nil (values nil nil)))))
    (get-rec (cons (named-module-or-init *elp-module-name*)
		   modules))))

(defun module-visible-p (module)
  "Returns T if the module is visible, NIL otherwise."
  (vismarker-visiblep (module-vismarker module)))

;;; This should no longer be used.
#|
(defun currently-visible-modules ()
  "Return the list of all currently visible modules."
  (declare-ftype currently-visible-modules () (List-of Module))
  (remove-if-not (fdlambda ( (Module _ (Vismarker visiblep) _ _ _ _) )
		     visiblep)
		 *all-modules*))
|#

(defun currently-accessible-modules ()
  "Return the list of currently accessible modules.  These are the modules
that are either visible, or imported by an accessible module.  This is used to
look up constants on parsing a query or when executing the special ``reread''."
  (declare-ftype currently-accessible-modules () (List-of Module))
  (let ((*accessibles* '()))
    (declare (special *accessibles*))
    (dolist (m *all-modules*)
      (when (module-visible-p m)
	(add-module-and-imports m)))
    *accessibles*))

(defun accessible-modules (module)
  "Returns a list of modules accessible from module.  Used for parsing
modules."
  (let ((*accessibles* '()))
    (declare (special *accessibles*))
    (add-module-and-imports module)
    *accessibles*))

(defun add-module-and-imports (module)
  (declare (special *accessibles*))
  (when (not (member module *accessibles* :test #'eq))
    (push module *accessibles*)
    (dolist (m (module-imports module))
      (add-module-and-imports m))))
