;;; -*- Mode: Lisp; Package: ELP -*-
;;;
;;; Representation and manipulation of the current Lambda-Prolog program.
;;;
;;; Author: Conal Elliott.
;;;
;;; Sccs Id @(#)lp-program.lisp	1.1 9/12/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :ELP) (use-package :ergolisp)


(export '(*unparse-clauses-p* program-init *show-context*))

(defparameter *unparse-clauses-p* t
  "Whether clauses are unparsed on printing.  Default t.")

(defparameter *show-context* t
  "Whether clauses are unparsed with their local context.")

;;;;	Clauses

(defsconstr "

    datatype Clause of
      %% Representation of program clause.  Note that neither the predicate
      %% name nor the visibility marker is here.
      tvnames : Symbol* ,		% Type variable names.
      vnames : Symbol* ,		% Term variable names.
      vtypes : Typ* ,			% Term variable types.
      head : Term ,			% Clause head term (in HNF).
      body : Gform ,
      idxval : T			% $$$ lower-case sym, value, NIL,T.
")

(ddefun print-clause ( (:as clause (Clause tvnames vnames vtypes 
					   head body _))
		       stream depth )
  "Redefinition of Clause printer."
  ;; Print term variables and their types, then the head, then the body.
  ;; Maybe we should print type variable names as well, but I don't think so.
  (if *unparse-clauses-p*
      ;; This commented code has a bug, since the names are actually just
      ;; hints.
      #| (format stream
	      "#<Clause (~{~a : ~a~^, ~}) Index ~a  ~a :- ~a >"
	      (mapcan #'(lambda (vname vtype)
			  (list vname (tlang-call mk-ctyp
						  tvnames vnames vtypes vtype)))
		      vnames vtypes)
	      idxval
	      (tlang-call mk-cterm tvnames vnames vtypes head)
	      (tlang-call gform-to-sterm body vtypes tvnames vnames 0)) |#
      (let ((cldform (Dimplies body (Datom nil head))))
	(if (not *show-context*)
	    (format stream "~A."
		    (tlang-call dform-to-sterm cldform
				vtypes tvnames vnames 0))
	    (format stream "~A"
		    (tlang-call dform-to-sterm
				(abstract-dform-over cldform
				 vnames vtypes)
				() tvnames () 0))))
      (print-clause-orig clause stream depth)))

(defun abstract-dform-over (dform vnames vtypes)
  (if (null vnames)
      dform
      (abstract-dform-over
       (Dall (Dlam (first vnames) (first vtypes) dform))
       (rest vnames) (rest vtypes))))

;;;;	Assum, Cgroup, and Program:  The current program.

(defsconstr "

    DATATYPE Assum of
      %% Clause assumed in an implication, with its visibility marker.
      vismarker : Vismarker,
      clause : Clause

    DATATYPE Cgroup of
      %% A group of clauses from same module.
      module : Module,
      clauses : Clause*

    DATATYPE Pentry of
      %% The entry in a program for a given predicate.
      assums : Assum*,			% From implications.
      cgroups : Cgroup*,		% From modules.
      index : Nat?			% $$$ experimental indexing.
")


(deftype Program ()
  "The type of a Lambda-Prolog program, i.e., a hash table, mapping a 
predicate to a Pentry."
  `(Hash-table-of Predicate Pentry #'EQ))

(defvar *program* (make-hash-table :test #'EQ)
  "The current LP program.  Has type Program.")

(defun program-init ()
  "Initialize the LP program."
  (module-init)
  (clrhash *program*)
  (values))

;; (program-init)

(defun program-lookup (pred)
  "Look up predicate PRED in the program.  Return a pentry, which is created
and installed if there was not one there already."
  (declare-ftype program-lookup (Predicate) Pentry)
  (or (gethash pred *program*)
      (setf (gethash pred *program*) (Pentry () () nil))))

(defun compute-idxval (index hnf-term)
  (if index
      (tlang-call term-nth-arg-head-value index hnf-term)
      nil))  

(defun set-compute-index (clause index hnf-term)
  (setf (clause-idxval clause)
	(compute-idxval index hnf-term)))

(defun index-pred (pred index)
  "Establishes the (pseudo)-indexing optimization for PREDICATE at argument
number INDEX.  Subtract one for efficiency."
  (setq index (if (or (not index) (< index 1))
		  nil
		  (1- index)))
  (dlet* ( ((:as pentry (Pentry assums cgroups prev-index))
	    (program-lookup pred)) )
    (when (not (eql index prev-index))
      ;; If the next set of forms is interrupted, we are in trouble.
      ;; Therefore, set NIL before starting.
      (setf (pentry-index pentry) nil)
      (mapc (fdlambda ( (Assum _ (:as clause (Clause _ _ _ head _ _))) )
	      (set-compute-index clause index head))
	    assums)
      (mapc (fdlambda ( (Cgroup _ clauses) )
	      (mapc (fdlambda ( (:as clause (Clause _ _ _ head _ _)) )
		      (set-compute-index clause index head))
		    clauses))
	    cgroups)
      (setf (pentry-index pentry) index))))


(defun add-module-clause (pentry clause module)
  "Add a module to the program, given PENTRY, CLAUSE and MODULE.  The clause is
installed after previously entered clauses for the same module.  A new cgroup
 associated with this module) will go before other cgroups."
  (declare-ftype add-module-clause (Pentry Clause Module) )
  ;; Get the cgroup for this module.   Put it BEFORE the other cgroups.
  (let ((cgroup (or (find module (pentry-cgroups pentry)
			  :key #'cgroup-module
			  :test #'EQ)
		    (let ((new-cgroup (Cgroup module ())))
		      (setf (pentry-cgroups pentry)
			    (cons new-cgroup (pentry-cgroups pentry)))
		      new-cgroup))))
    ;; Add the clause to THE END.
    (setf (cgroup-clauses cgroup)
	  (nconc (cgroup-clauses cgroup)
		 (list clause))))
  (values))

(defun add-implies-clause (pentry clause vismarker ctag)
  "Add a implication clause program, given PENTRY, CLAUSE and VISMARKER."
  (declare (special *trace-assumptions*))
  (declare-ftype add-implies-clause (Pentry Clause Module) )
  (lp-trace *trace-assumptions* ctag
	    "Assuming ~S." clause)
  (push (Assum vismarker clause)
	(pentry-assums pentry))
  (values))


(defun install-dform-clauses (dform tvnames vnames vtypes
				    install-fun error-fun)
  "Install the clauses generated by DFORM, with implicit universal 
quantification given by TVNAMES, VNAMES, and VTYPES, using the given function
INSTALL-FUN to install each constructed clause.  This function must take a 
pentry and clause.  See the interfaces install-module-dform and
install-implies-dform.  ERROR-FUN is called on the offending clause in case
it has a flexible head.
   The correctness of this code relies on DFORM being `normalized', that is,
DFORM must be a sequence of conjunction of universally quantified implications
of Gform's to Atoms.  This normalization happens in term-to-dform."
  (declare-ftype install-dform-clauses
    (Dform (List-of Symbol) (List-of Symbol) (List-of Ltype)
	   (Function-of (Predicate Clause) )
	   (Function-of ((List-of Symbol) (List-of Symbol) Term) )) )
  ;; The strategy is to maintain an "environment" of universal type and
  ;; term variables, and an assumed gform (built by conjunction).
  ;; Start the assumption as nil (not a legal gform).  We could start out as
  ;; (Gtrue) instead, and simplify this function, but it would be needlessly
  ;; inefficient at interpretation time.
  (labels ((install-and (dform vnames vtypes)
	    (dcase dform
	       ((Dand left right)
		(install-and right vnames vtypes)
		(install-and left vnames vtypes))
	       ((Dflex term)
		(let ((norm-dform (tlang-call term-to-dform term vtypes)))
		  (dcase norm-dform
		    ((Dflex term)
		     (funcall error-fun tvnames vnames vtypes term))
		    (_ (install-and norm-dform vnames vtypes)))))
	       (_ (install-pi dform vnames vtypes))))
	   (install-pi (dform vnames vtypes)
	    (dcase dform
	       ((Dall (Dlam name type body))
		(install-pi body (cons name vnames) (cons type vtypes)))
	       (_ (install-implies dform vnames vtypes))))
	   (install-implies (dform vnames vtypes)
	    (dcase dform
	       ((Dimplies gassum dconcl)
		(install-atom dconcl vnames vtypes gassum))
	       (_
		(install-atom dform vnames vtypes (Gtrue)))))
	   (install-atom (dform vnames vtypes clause-body)
	    (dcase dform
	      ((Datom pred term)
	       (dlet (( (:as pentry (Pentry _ _ index))
			(program-lookup pred) ))
		 (funcall install-fun
			  pentry
			  (Clause tvnames vnames vtypes
				  term clause-body
				  (compute-idxval index term)))))
	      (_ (error "Dform passed to INSTALL-DFORM-CLAUSES not properly normalized.")))))
    (install-and dform vnames vtypes))
  (values))

(defun install-module-dform (dform tvnames vnames vtypes module)
  "Install the clauses generated by DFORM, with implicit universal
quantification given by TVNAMES, VNAMES, and VTYPES, in the given
MODULE.  Returns the set (list without duplicates) of pentries for
which clauses were added."
  (declare-ftype install-module-dform
    (Dform (List-of Symbol) (List-of Symbol) (List-of Ltype) Module)
    (List-of Pentry))
  ;; Maintain a set "pentries" of affected pentries.
  (let ((pentries ()))
    (install-dform-clauses
     dform tvnames vnames vtypes
     #'(lambda (pentry clause)
	 (pushnew pentry pentries)
	 (add-module-clause pentry clause module))
     #'(lambda (tvnames vnames vtypes flex-term)
	 (lp-warn t "Ignoring clause ~S with flexible head."
		  (tlang-call mk-cterm tvnames vnames vtypes flex-term))
	 (values)))
    pentries))

(defun install-implies-dform (dform ctag)
  "Install the clauses generated by DFORM.  Returns two values:
 (1) the new vismarker associated with this group of clauses (which is
initially on), and (2) a function to delete the installed clauses.  WARNING:
installation and deletion must be properly nested."
  (declare-ftype install-implies-dform (Dform) Vismarker (Function-of () ))
  ;; Maintain an alist "old-pred-assums" from pentries to the assums they
  ;; had before we started.
  (let ((new-vismarker (Vismarker t))
	(old-pentry-assums ()))
    (install-dform-clauses
     dform () () ()
     (fdlambda ( (:as pentry (Pentry assums _ _)) clause )
       ;; Unless we already have the old assums for this, get it.
       (unless (assoc pentry old-pentry-assums :test #'EQ)
	 (push (cons pentry assums) old-pentry-assums))
       (add-implies-clause pentry clause new-vismarker ctag))
     #'(lambda (tvnames vnames vtypes flex-term) ; must be () () () flex-term.
	 (declare (ignore tvnames vnames vtypes))
	 (lp-error "Dformula ~S with flexible head cannot be assumed."
		   flex-term)))
    (values new-vismarker
	    #'(lambda ()
		;; (format t "Erasing assumptions.")
		(mapc (fdlambda ( (cons pentry assums) )
			(setf (pentry-assums pentry) assums))
		      old-pentry-assums)
		(values)))))
		  



(defun delete-module-clauses (module pentry)
  "Delete the clauses associated with MODULE from the given PENTRY, but
keep the empty cgroup so that the cgroup order will be maintined."
  (Declare-ftype delete-clauses (Module Pentry) )
  (let ((cgroup (find module (pentry-cgroups pentry)
		      :key #'cgroup-module
		      :test #'eq)))
    (assert cgroup ()
	    "DELETE-MODULE-CLAUSES: Something is wrong.  There is no cgroup for ~s in ~s"
	    module pentry)
    (setf (cgroup-clauses cgroup) ())
    (values)))
    
