;;; -*- Mode: Lisp; Package: ELP -*-
;;;
;;; The LP interpreter.  Based on Mats Carlsson's "success continuations"
;;; control strategy: for failure, return no values, and for success, invoke
;;; the current success continuation.  Ultimate success
;;; is handled by a throw from the initial success continuation.
;;;
;;; Author: Conal Elliott.  Last Modified Mon Sep 11 06:08:36 1989
;;;
;;; Sccs Id @(#)lp-solve.lisp	1.46 9/15/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :ELP) (use-package :ergolisp)

(export '(*warn-no-clauses* *warn-no-visible-clauses*))

(defvar *warn-no-clauses* t
  "If nil, the interpreter will issue no warning messages if a predicate
at the head of a goal has no clauses.")

(defvar *warn-no-visible-clauses* t
  "If nil, the interpreter will issue no warning messages if a predicate
at the head of a goal has no visible clauses.")

;;;; Special predicate data structure.  See lp-special.lisp

(defsconstr "
    DATATYPE Special of
       name : Symbol ,
       goalfun : 'a # Con # Ctag # Scont -> Unit,
       help : String ,
       arg-reptypes : Reptype* ,
       side-effect-p : Boolean
")

;;;;	Ctag: Cut tags.

(deftype Ctag ()
  "A cut tag, i.e., a tag used for catching and throwing cut failures."
  ;; Use integers, starting at zero.  Warning: distinctly constructed
  ;; integers are not guaranteed by the Common Lisp standard to be EQ.
  `Integer)

(defvar *prev-ctag* -1
  "The highest Ctag number used so far.")

(defun new-ctag ()
  "Make a new ctag."
  (declare-ftype new-ctag () Ctag)
  (incf *prev-ctag*))

(defun throw-cut (ctag)
  "Throw the given CTAG."
  (declare-ftype throw-cut (Ctag) )
  (throw ctag
	 (values)))

(defun ctag-init ()
  "Initialize the Ctag generator."
  (setq *prev-ctag* -1))

(defmacro catch-cut (ctag-var &rest body)
  "Make a new ctag called CTAG-VAR to use in BODY.  A throw-cut to that ctag
will cause this form to return (no values), thus signalling failure."
  ;; The failure throw may bypass waiting trail fixers.
  `(let ((,ctag-var (new-ctag)))
      (catch ,ctag-var ,@body)))

(defvar *trace-goal* nil
  "Whether we are tracing goals.")
(defvar *trace-clause* nil
  "Whether we are tracing clauses.")
(defvar *trace-io* nil
  "Whether we are tracing input and output.")

(defvar *trace-all-goals* nil
  "Whether all intermediate goals are shown.")
(defvar *trace-assumptions* nil
  "Whether dynamically made assumptions are shown.")

;;;;	Solve: The main interpreter.

(defun solve (goal con ctag sc)
  "Solve the given gform GOAL, extending the given constraint CON, with the
current cut tag CTAG and success continuation SC.  This function will either
 (a) return no value, in case of simple failure, (b) throw the given CTAG in
case of cut failure, or (c) throw 'SUCCESS in case of complete success."
  (declare-ftype solve (Gform Constraint Ctag Scont) )
  ;; tracing is here because otherwise either do-solve would lose some
  ;; tail calls, or ctag should have to be passed to resolve-program.
  (if (or (and *trace-goal* (is-Gatom goal)) *trace-all-goals*)
      (progn
	(lp-trace t ctag "Solving ~a" goal)
	(do-solve goal con ctag sc)
	(lp-trace t ctag "Goal ~a failed" goal))
      (do-solve goal con ctag sc)))

(defun do-solve (goal con ctag sc)
  "Helper function for solve.  Does everything but handling goal tracing."
  (when (and *working-ahead* (user-top-input-p))
    (if (more-answers-p)
	(setq *working-ahead* nil)
	(query-success)))
  (labels ((solve-quant (qarg uvarp)
	   ;; Solve the quantifier arg, substituting in a new uvar or a new evar
	   ;; depending on whether uvarp is true or false.  It should
	   ;; be a Glam or a translatable Gflex.
	   (dcase qarg
	     ((Glam name type body)
	      (solve (copy-gform body
				 ()
				 (list (if uvarp
					   (tlang-call new-uvar name type)
					   (tlang-call new-evar name type)))
				 0)
		     con ctag sc))
	     ((Gflex term)
	      (let ((retry-goal (tlang-call term-to-gform term ())))
		(when (is-Gflex retry-goal)
		  (lp-error "SOLVE: Flexible term ~s" term))
		(solve-quant retry-goal uvarp))))))
    (dcase goal
      (Gtrue
       (sc-invoke sc con))
      ((Gand left right)
       ;; Work on the LEFT, saving the RIGHT for later, i.e., adding it to
       ;; the success continuation.
       (solve left con ctag
	      (sc-make #'(lambda (new-con)
			   (solve right new-con ctag sc)))))
      ((Gor left right)
       ;; Try the LEFT.  If it returns, it will have failed, so try the RIGHT.
       (solve left con ctag sc)
       (solve right con ctag sc))
      ((Gall qarg)
       (solve-quant qarg t))
      ((Gexists qarg)
       (solve-quant qarg nil))
      ((Gimplies dassum gconcl)
       (dlet ( ((values vismarker erase-fun)
		(install-implies-dform dassum ctag)) )
	 (if *trace-assumptions*
	     (let ((nametag (gensym)))
	       (lp-trace *trace-assumptions* ctag
			 "Assumption group ~A visible [" nametag)
	       (unwind-protect
		   (solve gconcl con ctag
			  (trace-vismarkers-off-on-sc sc (list vismarker)
						      ctag nametag))
		 (lp-trace-unwind *trace-assumptions* ctag
				  "] Erasing assumption group ~A" nametag)
		 (funcall erase-fun)))
	     (unwind-protect
		 (solve gconcl con ctag
			(vismarkers-off-on-sc sc (list vismarker)))
	       (funcall erase-fun)))))
      ((Gspecial (Special _ goalfun _ _ side-effect-p) arg)
       (when side-effect-p (lp-check-for-more))
       (funcall goalfun arg con ctag sc))
      ((Gatom pred term)
       (resolve-program pred term con sc))
      ((Gflex term)
       (let ((retry-goal (tlang-call term-to-gform term ())))
	 (when (is-Gflex retry-goal)
	   (lp-error "SOLVE: Flexible term ~s" term))
	 (solve retry-goal con ctag sc))))))


(defun vismarkers-off-on-sc (sc vismarkers)
  "Given a success continuation SC and VISMARKERS, build a success 
continuation that (1) turns off the VISMARKERS, (2) invokes SC, and (3) turns
the VISMARKERS back on.  The third step happens even if the second step does not
return due to success or to cut failure.  This function is useful for
interpreting implication and module implication goals."
  (declare-ftype vismarkers-off-on-sc (Scont (List-of Vismarker)) Scont)
  #'(lambda (new-con)
      (dolist (vismarker vismarkers)
	(setf (vismarker-visiblep vismarker) nil))
      ;; If all predicates are properly scoped, nothing in the success
      ;; continuation should be able to throw to something that is caught
      ;; but does not leave that predicate in which this is called.  Thus
      ;; the unwind-protect here should be unnecessary.
      (prog1				; unwind-protect
	  (sc-invoke sc new-con)
	(dolist (vismarker vismarkers)
	  (setf (vismarker-visiblep vismarker) t)))))

(defun trace-vismarkers-off-on-sc (sc vismarkers ctag nametag)
  "Given a success continuation SC and VISMARKERS, build a success 
continuation that (1) turns off the VISMARKERS, (2) invokes SC, and (3) turns
the VISMARKERS back on.  The third step happens even if the second step does not
return due to success or to cut failure.  This function is useful for
interpreting implication and module implication goals."
  (declare-ftype vismarkers-off-on-sc (Scont (List-of Vismarker)) Scont)
  #'(lambda (new-con)
      (lp-trace *trace-assumptions* ctag
		"] Assumption group ~A now invisible" nametag)
      (dolist (vismarker vismarkers)
	(setf (vismarker-visiblep vismarker) nil))
      ;; If all predicates are properly scoped, nothing in the success
      ;; continuation should be able to throw to something that is caught
      ;; but does not leave that predicate in which this is called.  Thus
      ;; the unwind-protect here should be unnecessary.
      (unwind-protect
	  (sc-invoke sc new-con)
	(dolist (vismarker vismarkers)
	  (setf (vismarker-visiblep vismarker) t))
	(lp-trace-unwind *trace-assumptions* ctag
			 "Assumption group ~A now visible again [" nametag))))

(defun resolve-program (pred goal-term con sc)
  "Try to resolve with the relevant program clauses the given PRED and its
GOAL-TERM, carrying along the constraint CON, and scont SC."
  (declare-ftype resolve-program (Predicate Term Constraint Scont) )
  (dlet ( ((Pentry assums cgroups index) (program-lookup pred))
	  (found-one nil) )
    ;; First: if no assumptions and clause groups, warn if so advised.
    (when (and *warn-no-clauses* (null assums) (null cgroups))
      (lp-warn *warn-no-clauses*
	       "There are no clauses for predicate ~A." pred))
    ;; First set up a new ctag catcher
    (let ((goal-idxval (compute-idxval index goal-term)))
      (catch-cut new-ctag
	;; Try all of the visible assumed clauses and then all of the visible
	;; clause groups.  Recall that return means failure.
	(mapc (fdlambda ( (Assum (Vismarker visiblep) clause) )
		(when visiblep
		  (setq found-one t)
		  (when (or (not index)
			    (not goal-idxval)
			    (not (clause-idxval clause))
			    (eq goal-idxval (clause-idxval clause)))
		    (resolve-clause clause goal-term
				    () con new-ctag sc))))
	      assums)
	(mapc (fdlambda ( (Cgroup (Module _ (Vismarker visiblep)
					  imports _ _ _)
				  clauses) )
		(when visiblep
		  ;; Handle visibility of modules imported by the one
		  ;; in which these clauses are found.
		  (setq found-one t)
		  (let ((import-vismarkers (filter-modules-vismarkers
					    imports)))
		    (if import-vismarkers
			(progn 
			  (dolist (vismarker import-vismarkers)
			    (setf (vismarker-visiblep vismarker) t))
			  (unwind-protect
			      (dolist (clause clauses)
				(when (or (not index)
					  (not goal-idxval)
					  (not (clause-idxval clause))
					  (eq goal-idxval
					      (clause-idxval clause)))
				  (resolve-clause clause goal-term
						  import-vismarkers
						  con new-ctag sc)))
			    (dolist (vismarker import-vismarkers)
			      (setf (vismarker-visiblep vismarker) nil))))
			;; optimization: do not change context if no
			;; vismarkers have to be switched.
			(dolist (clause clauses)
			  (when (or (not index)
				    (not goal-idxval)
				    (not (clause-idxval clause))
				    (eq goal-idxval
					(clause-idxval clause)))
			    (resolve-clause clause goal-term
					    import-vismarkers
					    con new-ctag sc)))))))
	      cgroups)))
    (when (and *warn-no-visible-clauses* (not found-one))
      (lp-warn *warn-no-visible-clauses*
	       "None of the clauses for predicate ~A are visible." pred)))
  (values))

;; Not called if index value turns out not to match.
(ddefun resolve-clause ((Clause tvnames vnames vtypes clause-term body _)
			goal-term
			import-vismarkers
			con new-ctag sc)
  "Do a resolution, using given CLAUSE, with a goal having the
same predicate the given GOAL-TERM, with IMPORT-VISMARKERS, and carrying along
the constraint CON, ctag NEW-CTAG, and scont SC."
  (declare-ftype resolve-clause
    (Clause Term (List-of Vismarker) Constraint Ctag Scont) )
  ;; Set up new evar and tevar lists.
  ;; This should be bundled up into a function that is part of the term language
  ;; interface, taking three args and returning two values.  It's essentially
  ;; what new-copy-of-lterm does.
  (lp-trace *trace-clause* new-ctag "Trying clause with head ~A."
	    ;; clause-term
	    (tlang-call mk-cterm tvnames vnames vtypes clause-term))
  (let* ((new-tevars (mapcar #'(lambda (name)
				 (tlang-call new-tevar name))
			     tvnames))
	 (new-evars (mapcar #'(lambda (name type)
				(tlang-call new-evar name 
					    (tlang-call copy-typ 
							type new-tevars)))
			    vnames
			    vtypes)))
    (tlang-call clause-head-unify
		clause-term goal-term
		new-tevars new-evars
		con
		(sc-make
		 #'(lambda (new-con)
		     (lp-trace *trace-clause* new-ctag
			       "Resolved.  Instantiated goal to ~a"
			       (tlang-call mk-cterm () () () goal-term))
		     (lp-trace (and *trace-goal*
				    (not *trace-clause*))
			       new-ctag
			       "Resolved with clause whose head is ~a"
			       (tlang-call mk-cterm tvnames vnames
					   vtypes clause-term))
		     (if *trace-assumptions*
			 (let ((modnames
				(mapcar #'module-name
					(markers-to-mods import-vismarkers))))
			   (when import-vismarkers
			     (lp-trace *trace-assumptions* new-ctag
				       "Assuming modules ~{ ~A~}  ["
				       modnames))
			   (unwind-protect
			       (solve (copy-gform body new-tevars new-evars 0)
				      new-con new-ctag
				      (if import-vismarkers
					  (if *trace-assumptions*
					      (trace-vismarkers-off-on-sc
					       sc import-vismarkers new-ctag
					       (format nil "~{ ~A~} " modnames))
					      (vismarkers-off-on-sc sc import-vismarkers))
					  sc))
			     (when import-vismarkers
			       (lp-trace-unwind
				*trace-assumptions* new-ctag
				"] Discharging modules ~{ ~A~} "
				modnames))))
			 (solve (copy-gform body new-tevars new-evars 0)
				      new-con new-ctag
				      (if import-vismarkers
					  (vismarkers-off-on-sc sc import-vismarkers)
					  sc))))))))

(defun markers-to-mods (vismarkers)
  (remove-if-not (fdlambda ( (Module _ vismarker _ _ _ _) )
		   (member vismarker vismarkers :test #'eq))
		 *all-modules*))

(defun filter-modules-vismarkers (modules)
  "Returns a list of the vismarkers in MODULES that are currently invisible."
  (declare-ftype filter-modules-vismarkers
    ((List-of Module)) (List-of Vismarker) )
  (dcase modules
    (() ())
    ((cons (Module _ (:as vismarker (Vismarker visiblep)) _ _ _ _)
	   rest-modules)
     (if visiblep
	 (filter-modules-vismarkers rest-modules)
	 (cons vismarker (filter-modules-vismarkers rest-modules))))))


(defun copy-gform (gform new-tevars new-evars depth)
  "Make a copy of GFORM, substituting the given NEW-TEVARS and NEW-EVARS
for the corresponding bvars and btvars, at the given DEPTH"
  (declare-ftype copy-gform
    (Gform (List-of Tevar) (List-of Term) Integer) Gform)
  (flet ((copy-sub-term (term depth)
	   ;; Copy a term, using NEW-TEVARS and NEW-EVARS.
	   (tlang-call copy-term term new-tevars new-evars depth))
	 (copy-sub-typ (typ)
	   (tlang-call copy-typ typ new-tevars)))
    (labels ((gcopy (sub-gform depth)
	       ;; Copy SUB-GFORM, putting in NEW-TEVARS and NEW-EVARS, at
	       ;; given quantifier depth.
	       (dcase sub-gform
		 (Gtrue sub-gform)
		 ((Gand left right)
		  (re-Gand sub-gform left right
			   (gcopy left depth) (gcopy right depth)))
		 ((Gor left right)
		  (re-Gor sub-gform left right
			  (gcopy left depth) (gcopy right depth)))
		 ((Gall qarg)
		  (re-Gall sub-gform qarg
			   (gcopy qarg depth)))
		 ((Gexists qarg)
		  (re-Gexists sub-gform qarg
			      (gcopy qarg depth)))
		 ((Gimplies dassum gconcl)
		  (re-Gimplies sub-gform dassum gconcl
			       (dcopy dassum depth) (gcopy gconcl depth)))
		 ((Gspecial special arg)
		  (re-Gspecial sub-gform special arg
			       special
			       (copyrep-special
				special arg new-tevars new-evars depth)))
		 ((Gatom pred term)
		  (re-Gatom sub-gform pred term
			    pred (copy-sub-term term depth)))
		 ((Glam name type body)
		  (re-Glam sub-gform name type body
			   name (copy-sub-typ type) (gcopy body (1+ depth))))
		 ((Gflex term)
		  ;; Retry the translation.
		  (let ((new-trans (tlang-call term-to-gform
					       (copy-sub-term term depth)
					       ())))
		    (dcase new-trans
		      ((Gflex new-term)
		       (re-Gflex sub-gform term new-term))
		      (_ new-trans))))))
	     (dcopy (sub-dform depth)
	       ;; Copy SUB-DFORM, putting in NEW-TEVARS and NEW-EVARS, at
	       ;; given quantifier depth.
	       (dcase sub-dform
		 ((Dand left right)
		  (re-Dand sub-dform left right
			   (dcopy left depth) (dcopy right depth)))
		 ((Dall qarg)
		  (re-Dall sub-dform qarg
			   (dcopy qarg depth)))
		 ((Dimplies gassum dconcl)
		  (re-Dimplies sub-dform gassum dconcl
			       (gcopy gassum depth) (dcopy dconcl depth)))
		 ((Datom pred term)
		  (re-Datom sub-dform pred term
			    pred (copy-sub-term term depth)))
		 ((Dlam name type body)
		  (re-Dlam sub-dform name type body
			   name (copy-sub-typ type) (dcopy body (1+ depth))))
		 ((Dflex term)
		  ;; Retry the translation.
		  (let ((new-trans (tlang-call term-to-dform
					       (copy-sub-term term depth) ())))
		    (dcase new-trans
		      ((Dflex new-term)
		       (re-Dflex sub-dform term new-term))
		      (_ new-trans)))))))
      (gcopy gform depth))))

