;;; -*- Mode: Lisp; Package: ELP -*-
;;;
;;; Special predicates.
;;;
;;; Definition of the "Special" datatype, specifying interpretation of special
;;; predicates.  Also the collection of standard specials.
;;;
;;; Author: Conal Elliott.  Last Modified Mon Sep 18 16:53:15 1989
;;;
;;; Sccs Id @(#)lp-special.lisp	1.2 9/19/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :ELP) (use-package :ergolisp)

(export '(special-lookup defspecial
			 rep-special unrep-special copyrep-special))

;;;;	General stuff.


;;; Special predicates can devise their own internal representation.
;;; This is now embodied in something called a "representation type":
;;; one has to specify representation function (lterm -> sprep),
;;; unrepresentation function (sprep -> lterm) and copy function
;;; (sprep -> sprep).  The name of the latter is misleading,
;;; since free bvars have to be made into evars.

(defsconstr "
    DATATYPE Reptype of
       name : Symbol ,
       rtype : Type ,
       repfun : Term # Typ* -> 'a ,
       unrepfun : 'a # Typ* # Symbol* # Symbol* # Nat -> Sterm ,
       copyfun : 'a -> 'a ,
       help-string : String
")

(defvar *reptype-list* ()
  "Specification of representation types.")

(ddefun reptype-add ((:as reptype (Reptype name rtype
					   repfun unrepfun copyfun
					   help-string)))
  "Add REPTYPE to the collection of known representation types."
  (declare-ftype reptype-add (Reptype) )
  (let ((old-reptype-def (reptype-lookup name)))
    (if (not old-reptype-def)
	(push reptype *reptype-list*)
	(progn
	  (setf (reptype-name old-reptype-def) name)
	  (setf (reptype-rtype old-reptype-def) rtype)
	  (setf (reptype-repfun old-reptype-def) repfun)
	  (setf (reptype-unrepfun old-reptype-def) unrepfun)
	  (setf (reptype-copyfun old-reptype-def) copyfun)
	  (setf (reptype-help-string old-reptype-def) help-string))))
  (values))

(defun reptype-lookup (name)
  "Look up the Reptype associated with NAME.  Returns NIL if not found."
  (declare-ftype reptype-lookup (Symbol) (Optional Reptype))
  (find name *reptype-list* :key #'reptype-name))

(defmacro defreptype (name rtype help-string &key repfun unrepfun copyfun)
  `(reptype-add (Reptype ',name ',rtype
			 ,repfun ,unrepfun ,copyfun ',help-string)))

(defun rep-arglist (arg-reptypes arglist context)
  (mapcar #'(lambda (reptype arg) (funcall (reptype-repfun reptype)
					   arg context))
	  arg-reptypes arglist))

(defun unrep-arglist (arg-reptypes arglist vtypes tvnames vnames depth)
  (mapcar #'(lambda (reptype arg)
	      (funcall (reptype-unrepfun reptype)
		       arg vtypes tvnames vnames depth))
	  arg-reptypes arglist))

(defun copyrep-arglist (arg-reptypes arglist new-tevars new-evars depth)
  (mapcar #'(lambda (reptype arg)
	      (funcall (reptype-copyfun reptype) arg
		       new-tevars new-evars depth))
	  arg-reptypes arglist ))

(defreptype rterm Term
  "Representing an lterm as lterm."
  :repfun #'(lambda (term context)
	      (declare (ignore context))
	      term)
  :unrepfun #'(lambda (term vtypes tvnames vnames depth)
		(tlang-call to-sterm term vtypes tvnames vnames depth))
  :copyfun #'(lambda (term new-tevars new-evars depth)
	       (tlang-call copy-term term new-tevars new-evars depth)))

(defreptype rgform Gform
  "Representing an lterm as gform, since it will eventually be called as goal."
  :repfun #'(lambda (gterm context) (tlang-call term-to-gform gterm context))
  :unrepfun #'(lambda (gform vtypes tvnames vnames depth)
		(tlang-call gform-to-sterm gform vtypes tvnames vnames depth))
  :copyfun #'copy-gform)

;;; If more arguments are required, only the repfun must change.

(defreptype rgabst Gform
  "Representing an lterm as an abstraction over gforms."
  :repfun #'(lambda (gterm context)
	      (tlang-call absts-to-gform 1 gterm context))
  :unrepfun #'(lambda (gform vtypes tvnames vnames depth)
		(tlang-call gform-to-sterm gform vtypes tvnames vnames depth))
  :copyfun #'copy-gform)

;;;; See the Special data type in lp-solve.lisp 

;;; Now for defining specials.

(defvar *specials* (make-hash-table :test #'EQ)
  "Specification of the special predicates.  Has type 
Hash-table-of Symbol Special #'EQ")

(ddefun special-add ((:as special (Special name goalfun help
					   arg-reptypes
					   side-effect-p)))
  "Add SPECIAL to the collection of known specials."
  (declare-ftype special-add (Special) )
  (let ((old-special-def (special-lookup name)))
    (if (not old-special-def)
	(setf (gethash name *specials*) special)
	(progn
	  (setf (special-name old-special-def) name)
	  (setf (special-goalfun old-special-def) goalfun)
	  (setf (special-help old-special-def) help)
	  (setf (special-arg-reptypes old-special-def) arg-reptypes)
	  (setf (special-side-effect-p old-special-def) side-effect-p))))
  (values))

(defun special-lookup (name)
  "Look up the Special associated with NAME.  Returns NIL if not found."
  (declare-ftype special-lookup (Symbol) (Optional Special))
  (values (gethash name *specials*)))

(eval-when (compile load eval)
(defun reintern-lc (sym package)
  "Interns the print-name of the given symbol in package after lower-casing."
  (intern (string-downcase (symbol-name sym)) package))
)

(defmacro defspecial (name arg-reptypes side-effect-p
			   arg-pats help-string &rest body)
  "Define a special predicate called NAME with arguments of representation
types ARG-REPTYPES.
 SIDE-EFFECT-P of T means that the solving the special predicate will have
a global side-effect (such as reading or writing).
The function to execute the special predicate is synthesized from the
taking the given ARG-PATS (argument patterns), HELP-STRING,
and BODY.  ARG-PATS should consist of 4 elements: the representation of
the argument list, and patterns binding to `con', `ctag', and
`sc' (in the package :ELP), respectively, to the constraint, cut tag,
and success continuation.  Remember that the way to signal failure is to return,
and the thing to return is (values), and that the way to succeed is to say
 (sc-invoke sc con)."
  `(special-add (Special ',(reintern-lc name :ELP)
			 (fdlambda ,arg-pats
			   ,@body)
			 ',help-string
			 (mapcar #'reptype-lookup ',arg-reptypes)
			 ,side-effect-p)))

(defun rep-special (special args context)
  "Return the representation of special SPECIAL with arguments ARGS in context
CONTEXT."
  (rep-arglist (special-arg-reptypes special) args context))

(defun unrep-special (special args vtypes tvnames vnames depth)
  (unrep-arglist (special-arg-reptypes special) args vtypes tvnames vnames depth))

(defun copyrep-special (special args new-tevars new-evars depth)
  (copyrep-arglist (special-arg-reptypes special) args
		   new-tevars new-evars depth))

;;;; Macros to achieve global defaulting.

(defsconstr "
  DATATYPE Default of name : Symbol, setfun : T
")

(defvar *default-list* ()
  "List of defaults currently defined.  There should be one entry for each name.")

(defun default-lookup (default-name)
  "Looks up the a default."
  (find default-name *default-list* :key #'default-name))

(ddefun default-add ( (:as default (Default name setfun)) )
  "Add a default to the global list of defaults."
  (declare-ftype default-add (Default) )
  (let ((old-default (default-lookup name)))
    (if (not old-default)
	(push default *default-list*)
	(progn
	  (setf (default-name old-default) name)
	  (setf (default-setfun old-default) setfun))))
  (values))

(defmacro defdefault (name varlist &rest body)
  "Defines a Default with name NAME and function constructed from VARLIST and
BODY to affect the global change."
  `(default-add (Default ',(reintern-lc name :ELP)
		  #'(lambda ,varlist ,@body))))

;;; Useful conversion functions.

(defun term-simple-value (term)
  "Converts the term to a value.  If the result is not atomic, it signals an
error."
  (let ((v (tlang-call term-value term)))
    (when (consp v) (lp-error "Term ~S is not a simple value." term))
    v))

(defun term-string-value (term)
  "Converts a term to a string.  If the result is not atomic, it signals
an error."
  (tlang-call term-string-value term))

(defun term-number-value (term)
  "Converts the term to a number.  If the result is not a number, it signals an
error."
  (let ((v (tlang-call term-value term)))
    (when (not (numberp v)) (lp-error "Term ~S is not a true number." term))
    v))

;;;; Some variables that should be known as special
;;;; when special predicate handlers are defined.

;;;;	The special predicate handlers

(defspecial fail () nil
  ( () _ _ _ )
  "Simply fails."
  ;; Returns (no values), thus signalling failure
  (values))

(defspecial |!| () nil
  ( () con ctag sc )
  "The usual cut operator to commit to clause selection.
Note that subsequent failure will also remove choice points in
higher-order unification."
  ;;  Invoke the scont.  If it returns, then it has failed, so throw to the
  ;;  catch point.
  (sc-invoke sc con)
  (throw-cut ctag))

;;; once is cleaner than cut, but not generally as powerful.

(defmacro catch-once (ctag-var catch-fun &rest body)
  (let ((blockname (gensym)))
    `(let ((,ctag-var (new-ctag)))
       (block ,blockname
	 (funcall ,catch-fun 
		  (catch ,ctag-var (return-from ,blockname (progn . ,body))))))))

#|;; Instead of the following, we are redefining the unwind-trail function.
 (defvar *dont-unwind* nil
  "Set to T to prevent unwinding the trail when G succeeds in ``once G''.")

 (defun unwind-trail (trail-was)
  "Unwind the trail back to TRAIL-WAS, uninstantiating variables as we go."
  (declare-ftype unwind-trail (Trail) )
  (if *dont-unwind*
      (values)
      (do ()			
	  ((eq *trail* trail-was) (values))
	  (tlang-call unbind-var (pop *trail*)))))
|#

(defparameter *standard-unwind-function*
  (symbol-function 'unwind-trail))

(defspecial once (rgform) nil
  ( (list once-goal) con ctag sc )
  "Solves its arguments and then throws away all remaining choice points.
Note that this will still BUILD a lot of continuations!"
  ;; This is somewhat of a hack for efficient construction of the tag.
  ;; We use cut-tags.
  (catch-once new-ctag
	      ;; First form: when caught this is applied to the thrown
	      ;; value.
	      #'(lambda (new-con)
		  ;; (setq *dont-unwind* nil)
		  (setf (symbol-function 'unwind-trail)
			*standard-unwind-function*)
		  (sc-invoke sc new-con))
	      ;; Second form: Body
	      (solve once-goal con ctag
		     (sc-make #'(lambda (new-con)
				  ;; (setq *dont-unwind* t)
				  (setf (symbol-function 'unwind-trail)
					#'(lambda (trail-was)
					    (declare (ignore trail-was))
					    (values)))
				  (throw new-ctag new-con))))))

(defspecial |=| (rterm rterm) nil
  ( (list left right) con _ sc )
  "Unifies its arguments."
  ;;  Unifies LEFT and RIGHT, using CON and SC.
  (tlang-call unify left right con sc))

(defspecial error (rgform) t
  ( (list error-goal) con _ _ )
  "Solves its argument and then returns to the top-level."
  (format *elp-terminal-io* "ELP error: ")
  ;; Make sure the error goal cannot CUT out of here.
  (catch-cut new-ctag
	     (solve error-goal con new-ctag
		    (sc-make #'(lambda (new-con)
				 (declare (ignore new-con))
				 (throw 'lp-error (values)))))
	     (lp-error "~%Solving user supplied error goal failed."))
  (lp-error "~%Solving user supplied error goal tried to commit and then fail."))

(defspecial error_catch (rgform rgform) nil
  ( (list goal error-handler) con ctag sc )
  "Solves GOAL.  If an error occurs while solving GOAL, ERROR-HANDLER is
called AFTER the error message is printed.  This is not a general error
handling facility, but allows one to write a top level or interactive
program, for example.  From the Lisp debugger, one can also return to
the most recent error handler with (throw 'elp:lp-error nil), having the
exact same effect as if an error had occurred."
  (block error_catch
    (let ((throw-further nil))
      (catch 'lp-error
	(solve goal con ctag
	       (sc-make #'(lambda (new-con)
			    (setq throw-further t)
			    (prog1 (sc-invoke sc new-con)
			    ;; if this returns, catch again.
			    ;; if there is a throw, it whould go through
			    ;; here, so we don't need to manipulate
			    ;; throw-further.
			      (setq throw-further nil)))))
	;; We failed.  Just fail out of the error_catch invocation.
	(return-from error_catch (values)))
      ;; There was an error: call the error handler without protection.
      ;; UNLESS we have left the scope of the error_handler!
      (if throw-further
	  (throw 'lp-error (values))
	  (solve error-handler con ctag sc)))))

;;; Special predicates dealing with the top-level.

(defspecial top () t
  ( () con _ sc )
  "Calls another top level.  A top-level always succeeds: a pop is taken
as success."
  ;; Note that the top-level can only return: it can never "throw through"
  ;; this functions, 'bye is the only exception.  Therefore, in this case,
  ;; returning means SUCCESS.
  (top)
  (sc-invoke sc con))

(defspecial pop () t
  ( () _ _ _ )
  "Pop up to the previous top level."
  (throw 'pop nil))

(defspecial bye () t
  ( () _ _ _ )
  "Exit to Common Lisp."
  (throw 'bye nil))

;;;; Input/Ouput related special predicates.

(defvar *elp-input-prompt* " ?- "
  "The prompt used when reading input from a stream that is also an
output stream.")

(defspecial prompting (rterm rgform) nil
  ( (list string-term gform) con ctag sc )
  "``prompting S G'' sets the input prompt to S while solving G."
  (let ((old-prompt *elp-input-prompt*)
	(*elp-input-prompt* (term-string-value string-term)))
    (solve gform con ctag
	   (switch-variable-value '*elp-input-prompt*
				  *elp-input-prompt* old-prompt sc))))

(defdefault prompting (string)
  (value-string-value string)
  (setq *elp-input-prompt* string))

(defun switch-variable-value (var new-value old-value sc)
  "Returns a success continuation that will switch the value of the
Lisp variable VAR between NEW-VALUE and OLD-VALUE."
  (declare-ftype switch-switch (Symbol T T Scont) Scont)
  (sc-make
   #'(lambda (new-con)
       (set var old-value)
       ;; unwind-protect here is unnecessary, if throws are always properly
       ;; scoped.
       (prog1				; unwind-protect
	   (sc-invoke sc new-con)
	 (set var new-value)))))

(defspecial reread (rgabst) t
  ( (list gform) con ctag sc )
  "Reads a term and applies its second argument to this term.
The term must have the type expected by the second argument.
Free variables in the input will be allowed to depend on all uvars in
the current environment.
This will re-succeed until an end-of-stream is encountered,
at which point it fails."
  (when (is-Gflex gform)
    (setq gform (tlang-call term-to-gform (Gflex-first gform) ())))
  (dcase gform
    ((Glam name type body)
     ;; This "trailing" may seem strange, but is correct, because the
     ;; call to read-term does type unification that must be trailed.
     (catch 'end-of-file
       (do ((i 0 (1+ i)))
	   (nil)
	 ;;(declare (ignore i))		; ignore i for now.
	 ;; Pause in case we are working head.  Otherwise the prompt
	 ;; will show up ON FAILURE!
	 (lp-check-for-more)
	 (when (output-stream-p *elp-standard-input*)
	   (format *elp-standard-input* *elp-input-prompt*))
	 ;; trailing is for type unification done by read-term
	 (trailing
	  (let ((read-term (tlang-call read-term type)))
	    ;; Order is important: the time stamp of this evar must
	    ;; be higher than the time stamps of all Uvars in the term
	    ;; that is read!
	    (let ((new-evar (tlang-call new-evar name type)))
	      (tlang-call unify new-evar read-term con
			  (sc-make #'(lambda (new-con)
				       (solve (copy-gform body () (list new-evar) 0)
					      new-con ctag sc))))))))))
    ((Gflex term)
     (lp-error "SOLVE: Flexible term ~S" term))))


(defspecial write (rterm) t
  ( (list term) con _ sc )
  "Writes a term without surrounding spaces, newlines, or quotes."
  ;; Just writes a term (without its environment).
  ;; This should not use format because it would print quotes.
  (format *elp-standard-output* "~A" term)
  (sc-invoke sc con))

(defspecial writesans (rterm) t
  ( (list term) con _ sc )
  "Write a string without its quotes."
  ;; write a string without quotes.  Compatible with LP2.7.
  (format *elp-standard-output* "~A" (term-string-value term))
  (sc-invoke sc con))

(defspecial writeered (rterm) t
  ( (list term) con _ sc )
  "Writes a term in head-eta-reduced form."
  (format *elp-standard-output* "~A" (tlang-call head-eta-reduce term))
  (sc-invoke sc con))

(defspecial tab (rterm) t
  ( (list term) con _ sc )
  "Tab to the given position."
  ;; write a string without quotes.  Compatible with LP2.7.
  (format *elp-standard-output* "~VT" (term-number-value term))
  (sc-invoke sc con))

(defspecial fl () t
  ( () con _ sc )
  "Print a new-line, if not already at the beginning of a line."
  (format *elp-standard-output* "~&")
  (sc-invoke sc con))

(defspecial writetype (rterm) t
  ( (list term) con _ sc )
  "Given a term, write its type."
  (format *elp-standard-output* "~A" (tlang-call typ-of term ()))
  (sc-invoke sc con))

;;; reading and writing now open files and return streams.
;;; Thus the second argument is an abstraction!

(defvar *stream-alist* `((elp::|standard_input| . ,*elp-standard-input*)
			 (elp::|standard_output| . ,*elp-standard-output*))
  "An association list between uvars and streams.")

(defun solve-with-open-stream (uvar-name stream gform con ctag sc)
  "Solve GFORM with open stream STREAM.  Close on failure unless working
ahead."
  ;; the code in here is stolen from solve-quant in function
  ;; solve and slightly modified.
  (unwind-protect
      (progn
	(when (is-Gflex gform)
	  (setq gform (tlang-call term-to-gform (Gflex-first gform) ())))
	(dcase gform
	  ((Glam _ type body)
	   ;; ignore the name: it is generated from the filename.
	   (let ((new-uvar (tlang-call new-uvar
				       uvar-name
				       type)))
	     (let ((*stream-alist* (acons new-uvar stream *stream-alist*)))
	       (solve (copy-gform body () (list new-uvar) 0)
		      con ctag sc))))
	  ((Gflex term)
	   (lp-error "SOLVE: Flexible term ~S" term))))
    ;; lp-check-for-more is not necessary, since the no matter if
    ;; the user requests more solutions or not, the stream will be closed.
    ;; (lp-check-for-more)
    (close stream)))

(defun instream-var-name (filename-string)
  (intern (format nil "instream_~A" filename-string) :elp))

(defspecial reading (rterm rgabst) t
  ( (list string-term gform) con ctag sc )
  "Create an input stream from the file named by STRING, which should
be visible while solving GFORM."
  (let* ((filename-string (term-string-value string-term))
	 (full-filename (merge-pathnames filename-string
					 *default-pathname-defaults*)))
    (let ((instream (open full-filename :direction :input
			  :if-does-not-exist nil)))
      (when (not instream)
	(lp-error "File ~S does not exist." (namestring full-filename)))
      (lp-trace *trace-io* ctag "Opened file ~S for reading."
		(namestring full-filename))
      ;; The created stream should really be invisible while solving
      ;; goals in the "sc" aftern GFORM succeeds.  However, that
      ;; should not have to be reflected here --- the uvar itself
      ;; should be invisible!?  Thus, I am just "pushing" the
      ;; association between uvar and stream on an association list.
      (solve-with-open-stream (instream-var-name filename-string) instream
			      gform con ctag sc))))

(defspecial seeing (rterm rgform) nil
  ( (list stream-term gform) con ctag sc )
  "Take input stream from the stream named by STREAM while solving GFORM."
  (let* ((normal-stream-term (tlang-call term-value stream-term))
	 (instream (cdr (assoc normal-stream-term *stream-alist*))))
    (when (not instream)
      (lp-error "Stream ~S is not a known input stream." normal-stream-term))
    (let ((old-standard-input *elp-standard-input*)
	  (*elp-standard-input* instream))
      (solve gform con ctag
	     (elp-standard-input-switch instream old-standard-input sc)))))

(defun elp-standard-input-switch (instream old-instream sc)
  "Given currently open file INSTREAM and previously open file OLD-INSTREAM
and a success continuation SC, it returns a new success continuation that
switches the roles of instream and old-instream when before calling SC
and upon backtracking."
  (declare-ftype elp-standard-input-switch (Stream Stream Scont) Scont)
  (sc-make 
   #'(lambda (new-con)
      (setq *elp-standard-input* old-instream)
      ;; Unwind-protect should be unnecessary if throw/catch is always
      ;; properly scoped.
      (prog1				; unwind-protect
	  (sc-invoke sc new-con)
	(setq *elp-standard-input* instream)))))

(defun outstream-var-name (filename-string)
  (intern (format nil "outstream_~A" filename-string) :elp))

(defspecial writing (rterm rgabst) t
  ( (list string-term gform) con ctag sc )
  "Create an output stream to a file named by STRING while, visible while
solving GFORM."
  (let* ((filename-string (term-string-value string-term))
	 (full-filename (merge-pathnames filename-string
					 *default-pathname-defaults*)))
    (let ((outstream (open full-filename :direction :output
			   :if-exists :supersede
			   :if-does-not-exist :create)))
      ;; with-open-file does not work here, since an abnormal exit
      ;; (eg throw) will delete the file, but shouldn't.  
      ;; Should test what happens if the DIRECTORY doesn't exist.
      (lp-trace *trace-io* ctag "Opened file ~S for writing."
		(namestring full-filename))
      (solve-with-open-stream (outstream-var-name filename-string) outstream
			      gform con ctag sc))))


(defspecial telling (rterm rgform) nil
  ( (list stream-term gform) con ctag sc )
  "Write output to stream named by STREAM-TERM, while solving GFORM."
  ;; Please see SEEING for comments on this.
  (let* ((normal-stream-term (tlang-call term-value stream-term))
	 (outstream (cdr (assoc normal-stream-term *stream-alist*))))
    (when (not outstream)
      (lp-error "Stream ~S is not a known output stream." normal-stream-term))
    (let ((old-standard-output *elp-standard-output*)
	  (*elp-standard-output* outstream))
      (solve gform con ctag
	     (elp-standard-output-switch outstream old-standard-output sc)))))

(defun elp-standard-output-switch (outstream old-outstream sc)
  "Given currently open file OPENFILE and previously open file OLD-OPENFILE
and a success continuation SC, it returns a new success continuation that
switches the roles of openfile and old-outstream when before calling SC
and upon backtracking."
  (declare-ftype elp-standard-output-switch (Stream Stream Scont) Scont)
  (sc-make 
   #'(lambda (new-con)
       (setq *elp-standard-output* old-outstream)
       ;; if throws are properly scoped, don't need unwind-protect.
       (prog1				; unwind-protect
	   (sc-invoke sc new-con)
	 (setq *elp-standard-output* outstream)))))

;;;; Module related predicates.

(defspecial load (rterm) t
  ( (list module-term) con _ sc )
  "Loads or reloads a module."
  (load-module (term-string-value module-term))
  (sc-invoke sc con))

(defspecial |==>| (rterm rgform) nil
  ( (list module-term gform) con ctag sc )
  "``M ==> G'' assumes modules M while solving G."
  (dlet (( (Module name vismarker _ _ _ _)
	   (named-module-or-load
	    (term-string-value module-term) )))
    (if (vismarker-visiblep vismarker)
	(solve gform con ctag sc)
	(progn (lp-trace *trace-assumptions* ctag
			 "Assuming module ~A ["
			 name)
	       (setf (vismarker-visiblep vismarker) t)
	       (unwind-protect
		   (solve gform con ctag
			  (if *trace-assumptions*
			      (trace-vismarkers-off-on-sc
			       sc (list vismarker) ctag name)
			      (vismarkers-off-on-sc sc (list vismarker))))
		 (setf (vismarker-visiblep vismarker) nil)
		 (lp-trace-unwind *trace-assumptions* ctag
				  "] Module ~A invisible"
				  name))))))

(defdefault |==>| (module-spec)
  (let ((module-name (value-string-value module-spec)))
    (dlet (( (Module _ vismarker _ _ _ _)
	     (named-module-or-load module-name) ))
      (setf (vismarker-visiblep vismarker) t))))

(defdefault using_none ()
  (let ((elp-sys-module (named-module-or-bust *elp-module-name*)))
    (mapc #'(lambda (m)
	      (unless (eq m elp-sys-module)
		(setf (vismarker-visiblep (module-vismarker m)) nil)))
	  *all-modules*)))

(defspecial uses () t
  ( () con _ sc )
  "Shows the list of currently visible modules."
  (format *elp-standard-output*
	  "~&Currently visible modules are: ~{ ~A~}." ; ~A since name is string
	  (mapcar #'module-name
		  (remove-if-not #'module-visible-p *all-modules*)))
  (sc-invoke sc con))

(ddefun write-assum-clause ( visiblep clause )
  "Write out an assumed clause, in square brackets in case it is not
currently visible.  The index value is ignored."
  (declare-ftype write-assum-clause (Bool Clause) )
  (let ((*show-context* t))
    (declare (special *show-context*))
    (format *elp-standard-output*
	    "~&~3T~:[[ ~;~]~A~:[ ]~;~]"
	    visiblep clause visiblep)))

(defun write-module-clauses (mod visiblep clauses)
  "Write out the clauses defining PRED in module name MOD, in square brackets
in case the module is not currently visible."
  (format *elp-standard-output*
	  "~&~2T~:[[ ~;~]Module ~A:~:[] ~;~]"
	  visiblep mod visiblep)
  (let ((*show-context* nil))
    (declare (special *show-context*))
    (dolist (clause clauses)
      (format *elp-standard-output* "~&~3T~:[[ ~;~]~A~:[ ]~;~]"
	      visiblep clause visiblep))))

(defun write-clauses (pred)
  "Write out the program clauses for the given PRED."
  (declare-ftype write-clauses
    (Predicate) )
  (dlet ( ((Pentry assums cgroups index) (program-lookup pred)) )
    (when index
      (format *elp-standard-output*
	      "~&Predicate ~A is indexed on argument number ~D."
	      pred (1+ index)))
    (if (and (null assums) (null cgroups))
	(format *elp-standard-output*
		"~&There are no clauses for predicate ~A." pred)
	(progn
	  (if (null assums)
	      (format *elp-standard-output*
		      "~&No assumptions resulting from implicational goals.")
	      (progn
		(format *elp-standard-output*
			"~&Assumptions resulting from implicational goals (no implicit quantifiers):")
		(mapc (fdlambda ( (Assum (Vismarker visiblep) clause) )
			(write-assum-clause visiblep clause))
		      assums)))
	  (if (null cgroups)
	      (format *elp-standard-output*
		      "~&No clauses defined in modules.")
	      (progn
		(format *elp-standard-output*
			"~&Clauses defined in modules (universal quantifiers implicit):")
		(mapc (fdlambda ( (Cgroup (Module mod (Vismarker visiblep) _ _ _ _)
					  clauses) )
			(write-module-clauses mod visiblep clauses))
		      cgroups)))))))

(defspecial listing_pred (rterm) t
  ( (list term) con _ sc )
  "Show the clauses for the given predicate."
  (write-clauses (tlang-call term-to-pred term))
  (sc-invoke sc con))

;;; Similar functions are in $tools/sccs.lisp, now obsolete. -fp.

(defun elp-directory-only (pathname)
  "Takes a pathname (for example, a string) which is supposed to be
a directory and returns a string with that directory.
Signals an lp-error if the directory does not exist."
  (when (pathname-name pathname)
    (setq pathname (elp-directorify (namestring pathname))))
  (let ((truename (#-cmu probe-file #+cmu lisp::sub-probe-file pathname)))
    (cond ((not truename)
	   (lp-error "Directory ~S does not exist." pathname))
	  (t (directory-namestring truename)))))

(defun elp-directorify (pathname)
  "Takes a string which is supposed to be only a directory and tries
to coerce it into one.  This often means appending a /."
  (declare-ftype elp-directorify (string) string)
  (if (equal (char pathname (1- (length pathname)))
	     #\/)
      pathname
      (format nil "~A/" pathname)))

(defvar *elp-search-path* (list "./")
  "A variable with a list of pathname: the paths to search.
These could be strings.")

(defspecial searching (rterm rgform) nil
  ( (list string-term gform) con ctag sc )
  "``searching \"directory\" G'' searches \"directory\" when loading
a module while solving G."
  (let* ((string (term-string-value string-term))
	 (directory (elp-directory-only string)))
    (lp-trace *trace-io* ctag "Now also searching ~S when loading a module."
	      directory)
    (let ((*elp-search-path* (cons directory *elp-search-path*)))
      (solve gform con ctag
	     (sc-make #'(lambda (new-con)
			  (pop *elp-search-path*)
			  ;; With properly scoped throws, need only prog1.
			  (prog1	; unwind-protect
			      (sc-invoke sc new-con)
			    (push directory *elp-search-path*))))))))

(defdefault searching (string)
  (push (elp-directory-only (value-string-value string)) *elp-search-path*))

(defdefault searching_none ()
  (setq *elp-search-path* (list "./")))

(defun find-module-file (module-name)
  "Try to find the file name corresponding to MODULE-NAME."
  ;; This should certainly try harder.
  (let ((filename-name (concatenate 'string module-name ".mod")))
    (dolist (dir *elp-search-path*)
      (let ((existent-file (probe-file (merge-pathnames filename-name dir))))
	(when existent-file
	  (return-from find-module-file
	    (namestring (truename existent-file))))))
    (lp-error "The module ~A was not found in current search path: ~{~% ~S~}."
	      module-name *elp-search-path*)))

(defspecial searches () t
  ( () con _ sc )
  "Shows the list of directories currently searched when attempting to locate
the file defining a module."
  (format *elp-standard-output*
	  "~&Currently searching in directories ~{~% ~S~}."
	  *elp-search-path*)
  (sc-invoke sc con))

;;; Defaulting of grammar.  Only as default, not scoped.

(defvar *current-grammar* :standard
  "A variable with the directory of the current grammar, or :standard.
This might be inaccurate if the grammar is loaded without calling
load-slterm-lang.")

(defdefault grammar (string)
  (let ((gr-dir (elp-directory-only (value-string-value string))))
    (tlang-call load-sterm-lang gr-dir)
    (setq *current-grammar* gr-dir)))

(defdefault standard_grammar ()
  (tlang-call load-sterm-lang :standard))

(defspecial parses () t
  ( () con _ sc )
  "Shows the directory in which the currently used grammar resides."
  (if (eq *current-grammar* :standard)
      (format *elp-standard-output*
	      "~&Currently using standard grammar.")
      (format *elp-standard-output*
	      "~&Currently using grammar in directory ~S."
	      *current-grammar*))
  (sc-invoke sc con))

;;; Switches (these can currently not be queried, only set!).

;;; A switch name may be a complex structure: it is compared with
;;; EQUAL.  This is used for tracing options.

(defsconstr "
  DATATYPE Switch of name : T, varname : Symbol , help : String
")

(defvar *switch-list* ()
  "List of switches currently defined.  There should be only one entry for
each name.")

(defun switch-lookup (switch-name)
  "Looks up the internal name for a switch."
  (find switch-name *switch-list* :key #'switch-name
	:test #'equal))

(ddefun switch-add ((:as switch (Switch switch-name varname help)))
  "Add SWITCH to the list of defined switches."
  (declare-ftype switch-add (Switch) )
  (let ((old-switch-def (switch-lookup switch-name)))
    (if (not old-switch-def)
	(push switch *switch-list*)
	(progn
	  (setf (switch-varname old-switch-def) varname)
	  (setf (switch-help old-switch-def) help))))
  (values))

(eval-when (compile load eval)
(defun map-tree (fun tree)
  "Apply FUN to every atom in TREE.  Tree must end in non-NIL symbols."
  (labels ((map-tree-rec (tree)
			 (etypecase tree
			   (null (error "Tree ~A should not contain leaves ()."
					tree))
			   (symbol (funcall fun tree))
			   (cons (cons (map-tree-rec (car tree))
				       (map-tree-rec (cdr tree)))))))
    (map-tree-rec tree)))
)

(defmacro defswitch (name varname help)
  "Define a switch called NAME tied to the Lisp variable VARNAME and
help-string HELP.  This switch must also be added as a type declaration
to elp.mod!"
  `(switch-add (Switch ',(map-tree #'(lambda (sym) (reintern-lc sym :ELP))
				   name)
		       ',varname
		       ',help)))

;;; The function for representation of the SWITCHING special predicate.
;;; This is not very efficient right now.

(defun switch-term-lookup-or-error (switch-term)
  (switch-lookup-or-error (tlang-call term-value switch-term)))

(defun switch-lookup-or-error (switch-name)
  "Coerces the given term to a switch.  Error if undefined.
Switch-name is compared with EQUAL, since it may stem from (tracing X)."
  (let* ((switch (switch-lookup switch-name)))
    (when (not switch) (lp-error "Switch ~A is undefined." switch-name))
    switch))

(defun translate-onoff-term (onoff-term)
  "Takes an term that should be ON or OFF and translates to T or NIL.
Signals an lp error, if it is neither of those two."
  (translate-onoff (term-simple-value onoff-term)))

(defun translate-onoff (onoff)
  "Translates a symbol for on or off to T or NIL."
  (case onoff
    (|on| t)
    (|off| nil)
      (t (lp-error "Switch value ~A is neither on nor off." onoff))))

(defspecial switching (rterm rterm rgform) nil
  ( (list switch-term setting-term gform) con ctag sc )
  "``switching Sw Onoff G'' sets switch Sw to Onoff while solving G."
  (let* ((switch (switch-term-lookup-or-error switch-term))
	 (switch-var (switch-varname switch))
	 (setting (translate-onoff-term setting-term))
	 (old-setting (eval switch-var)))
    (if (eq old-setting setting)
	(solve gform con ctag sc)
	(unwind-protect
	    (progn
	      (set switch-var setting)
	      (solve gform con ctag
		     (switch-variable-value switch-var
					    setting old-setting sc)))
	  (set switch-var old-setting)))))

(defdefault switching (switch-name value)
  (let* ((switch (switch-lookup-or-error switch-name))
	 (switch-var (switch-varname switch))
	 (setting (translate-onoff value)))
    (set switch-var setting)))

;;;; The definition of the switches.

(defvar *timings-after-each-solution* nil "Used by special predicate timing.")
(defvar *timings-totals* nil "Used by special predicate timing_totals.")

(defswitch timing elp::*timings-after-each-solution*
  "If ON, time each solution.")

(defswitch timing_totals *timings-totals*
  "If ON, give total runtime.")

;;;; Tracing and Debugging

(defswitch (tracing . goal) elp::*trace-goal*
  "If ON, trace the current goal.")

(defswitch (tracing . clause) elp::*trace-clause*
  "If ON, trace the clause selection.")

(defswitch (tracing . all_goals) elp::*trace-all-goals*
  "If ON, all goals (not just atomic goals) are shown while tracing.")

(defswitch (tracing . assumptions) elp::*trace-assumptions*
  "If ON, assumptions and their visibility is traced.")

(defswitch (tracing . io) elp::*trace-io*
  "If ON, trace the opening and closing of files.")

(defswitch (warning . no_clauses) elp::*warn-no-clauses*
  "If OFF, no warning messages will be given if a predicate at the head
of the goal has no clauses.")

(defswitch (warning . no_visible_clauses) elp::*warn-no-visible-clauses*
  "If OFF, no warning messages will be given if a predicate at the head
of the goal has no clauses.")

(defswitch (warning . multiple_decl) elp::*warn-multiple-decl*
  "If OFF, no warnings about multiple declarations will be given.")

(defswitch show_undeclared_constants elp::*show-undeclared-constants*
  "If OFF, no listing of undeclared constants in a module will be given.")

(defswitch work_ahead elp::*work-ahead*
  "If ON, the interpreter will work ahead, looking for more solutions after
giving the first one, but not presenting them unless present_eager is on.")

(defswitch present_eager elp::*present-eager*
  "If ON, and if work_ahead is also on, successive solutions will continue to
be printed even without the user typing ``;''.  However, the user can still type
return to stop the search.")

;;; When we add more, add them to elp:save.

(defspecial switches () t
  ( () con _ sc )
  "Shows the current switch values."
  (format *elp-standard-output* "~&Current switch values are:")
  (dolist (switch (reverse *switch-list*))
    (dlet (( (Switch name varname _) switch ))
      (format *elp-standard-output* "~%  ~20<~A~> is ~A."
	      (if (consp name) (list (car name) (cdr name)) name)
	      (if (eval varname) "on" "off"))))
  (sc-invoke sc con))

;;;; Global initialization functions.
;;;; These allow global initialization of various switches and values.

(defun analyze-default-arg (default-arg)
  "Returns the function and argument portion of the result of calling
term-value on the argument to ``default''."
  ;; the cadr in the body of this function comes from the inefficient
  ;; representation of "values".  Should use dotted pairs instead.
  (do ((darg default-arg (car darg))
       (arglist () (cons (cdr darg) arglist)))
      ((not (consp darg))
       (values darg arglist))))

(defun value-string-value (value)
  "Given something which is alread a value, coerce it into a string."
  (typecase value
    (string value)
    (symbol (symbol-name value))
    ((or cons number) (lp-error "~A is not a true string."))
    (t ;; Now it better be a Uvar.
     (tlang-call term-string-value value))))

(defspecial default (rterm) t
  ( (list term) con _ sc )
  "``default E''  is the destructive counterpart of ``assume E''.
It is intended mostly for initialization files.  ``default E''
can only be applied to a limited number of special predicates.
Currently E must not contain an abstraction."
  (let* ((arg-value (tlang-call term-value term)))
    (dlet (( (values default-name default-args)
	     (analyze-default-arg arg-value) ))
      (let ((default (default-lookup default-name)))
	(when (not default)
	  (lp-error "~A is not a defaultable entity." arg-value))
	(dlet (( (Default _ setfun) default ))
	  (apply setfun default-args)))))
  (sc-invoke sc con))


;;;; Some maintenance functions.

(defun special-init ()
  "Clears out all information about special predicates and switches."
  (warn "SPECIAL-INIT:  You should not have to call this.  Please tell Frank.")
  (clrhash elp::*specials*)
  (setq elp::*switch-list* ())
  (setq elp::*default-list* ()))
