;;; -*- Mode: Lisp; Package: LTERM -*-
;;;
;;; The ADTs of LTerms, Ltypes, etc.
;;;
;;; Author: Conal Elliott.  Last Modified Mon Sep 11 06:09:26 1989
;;;
;;; Sccs Id @(#)lterm-adts.lisp	1.30 9/15/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :LTERM) (use-package :ergolisp)

;;;;	Lterm

(eval-when (compile load eval)
(defsconstr "

    DATATYPE Lterm  =
        Const of cobj : Lcobj, type : Ltype
      |	Bvar of index : Nat
      |	Evar of name : Symbol,
		type : Ltype,
		stamp : Stamp,
		ident : Ref ident:Ident-string?,
		value-slot : Ref Lterm?
      |	Uvar of name : Symbol,
		type : Ltype,
		stamp : Stamp,
		ident : Ref ident:Ident-string?
      | Appl of rator : Lterm , rand : Lterm
      |	Abst of vname : Symbol, type : Ltype, body : Lterm
")
)

(defun chase-evars (lterm)
  "Follow any evar indirections."
  (declare-ftype chase-evars (Lterm) Lterm)
  (dcase lterm
    ((Evar _ _ _ _ value-slot)
     (if value-slot
	 (chase-evars value-slot)
	 lterm))
    (_ lterm)))


;;;;	Ltype

(export '(new-tevar chase-tevars))


(eval-when (compile load eval)
(defsconstr "

    DATATYPE Ltype  =
        Tcon of  name : Symbol, kind : Lkind
      | Tbvar of index : Nat
      |	Tevar of name : Symbol,
		 ident : Ref ident:Ident-string?,
		 value-slot : Ref Ltype?
      | Tfun of domain : Ltype, range : Ltype
      |	Tappl of rator : Ltype, rand : Ltype
")
)

(defun new-tevar (name)
  "Make a new tevar ltype, with the given NAME."
  (declare-ftype new-tevar (Symbol) Tevar)
  (Tevar name nil nil))


(defun chase-tevars (ltype)
  "Follow any tevar indirections."
  (declare-ftype chase-tevars (Ltype) Ltype)
  (dcase ltype
    ((Tevar _ _ value-slot)
     (if value-slot
	 (chase-tevars value-slot)
	 ltype))
    (_ ltype)))


;;;;	Ltyfam

(eval-when (compile load eval)
(defsconstr "
    DATATYPE Ltyfam of tvnames : Symbol*, type : Ltype
")
)

;;;;	Lkind

;;; These should do some caching.  At least, (Ktype) should always return the
;;; same (eq) thing.

(eval-when (compile load eval)
(defsconstr "

    DATATYPE Lkind  =
        Ktype
      |	Kfun of range : Lkind

")
)

;;;; Instead, we use a compile-time constant!

(eval-when (compile load eval)
(defconstant *ktype* (Ktype)
  "The constant standing for the kind `Type'.  This is used for efficiency,
to prevent many many (Ktype)'s to be built.")
)

;;;;	Lcobj

;;; Constant objects.  Supports strings, numbers, symbols, and named constants.
;;; The implementation must make instance of "the same" lcobj be EQ.

#| The ADT is specified, but not implemented, by

    DATATYPE Lcobj =
	String-lcobj of string : String
      | Number-lcobj of number : Integer
      | Named-lcobj of name : Symbol

|#

(export '(String-lcobj string-lcobj-string is-string-lcobj *string-ltype*))
(export '(Number-lcobj number-lcobj-number is-number-lcobj *number-ltype*))
(export '(Named-lcobj named-lcobj-name is-named-lcobj))


;;; The plan for implementing this type:  Uses strings, numbers, and symbols
;;; directly.  To distinguish a symbol-cobj from a named-cobj, put all of the
;;; symbol-cobj's into a special package.  (SYMBOL-COBJ OBSOLETE)
;;;
;;; We must guarantee that same lcobjs are eq.  We do this
;;; by hashing the numbers and strings into their own hash tables.  We could
;;; intern the strings instead of directly hashing them.

(eval-when (compile load eval)

(deftype Lcobj ()
  "Object with which to build a constant.  Same lcobjs guaranteed to be EQ.
Can be string, number, symbol as symbolic lcobj, or symbol as a named lcobj."
  `(or String Number Symbol Symbol))

(proclaim '(inline lcobj-value))

(defun lcobj-value (lcobj)
  "Return the `value' associated with LCOBJ."
  (declare-ftype lcobj-value (Lcobj) (or String Number Symbol Symbol))
  lcobj)


(unless (find-package :LCOBJ-STRING-PACKAGE)
  (make-package :LCOBJ-STRING-PACKAGE :use ()))
(defvar *lcobj-number-table* (make-hash-table :test #'eql))
(unless (find-package :LCOBJ-SYMBOL-PACKAGE)
  (make-package :LCOBJ-SYMBOL-PACKAGE :use ()))

(deftype String-lcobj () `String)
(declare-constructor String-lcobj (string-lcobj-string) is-string-lcobj)
(defun String-lcobj (string)
  "Convert STRING into a string-lcobj.  Guarantees eq-ness for equal strings."
  (declare-ftype String-lcobj (String) String-lcobj)
  ;; Intern it in the string package and then get the symbol name.
  (symbol-name (intern string :LCOBJ-STRING-PACKAGE)))
(defun string-lcobj-string (string-lcobj)
  (declare-ftype string-lcobj-string (String-lcobj) String)
  string-lcobj)
(defun is-string-lcobj (lcobj)
  (stringp lcobj))

(deftype Number-lcobj () `Number)
(declare-constructor Number-lcobj (number-lcobj-number) is-number-lcobj)
(defun Number-lcobj (number)
  "Convert NUMBER into a number-lcobj.  Guarantees eq-ness for = numbers."
  (declare-ftype Number-lcobj (Number) Number-lcobj)
  (let ((old-num  (gethash number *lcobj-number-table*)))
    (or old-num
	(progn
	  (setf (gethash number *lcobj-number-table*) number)
	  number))))
(defun number-lcobj-number (number-lcobj)
  (declare-ftype number-lcobj-number (Number-lcobj) Number)
  number-lcobj)
(defun is-number-lcobj (lcobj)
  (numberp lcobj))

(deftype Named-lcobj () `Symbol)
(declare-constructor Named-lcobj (named-lcobj-name) is-named-lcobj)
(defun Named-lcobj (name)
  "Convert NAME (a symbol) into a named-lcobj.  Guarantees eq-ness for eq
names."
  (declare-ftype Named-lcobj (Symbol) Named-lcobj)
  ;; Just return the name.
  name)
(defun named-lcobj-name (named-lcobj)
  (declare-ftype named-lcobj-name (Named-lcobj) Symbol)
  named-lcobj)
(defun is-named-lcobj (lcobj)
  (and (symbolp lcobj)
       (not (eq (symbol-package lcobj) (find-package :LCOBJ-SYMBOL-PACKAGE)))))


(defparameter *string-ltype* (Tcon 'elp::|string| *ktype*))
(defparameter *number-ltype* (Tcon 'elp::|int| *ktype*))

)
;;;;  Other functions


(ddefun ltyfam-instance ( (Ltyfam names ltype) )
  "Make a new copy of the ltyfam, replacing the tbvars with tevars."
  (declare-ftype ltyfam-instance (Ltyfam) Ltype)
  (if (null names)
      ltype
      (copy-ltype ltype (mapcar #'new-tevar names))))



(defun lkind-of (ltype)
  "Compute the lkind of LTYPE.  Assume well-formed."
  (declare-ftype lkind-of (Ltype) Lkind)
  (dcase (chase-tevars ltype)
    ((Tcon _ kind)
     kind)
    (Tbvar *ktype*)
    (Tevar *ktype*)
    (Tfun *ktype*)
    ((Tappl rator _)
     (dlet ( ((Kfun range) (lkind-of rator)) )
       range))))

(defun ltype-of (lterm context)
  "Compute the type of LTERM in the given CONTEXT, which is a list of ltypes
ofr the loose bvars.  (The ith type corresponds to bvar i.  Assumes that
LTERM is well-typed in CONTEXT.  Chases tevars in result.  If there are loose
bvars that exceed the context, throw nil to :LOOSE-BVAR."
  (declare-ftype ltype-of (Lterm (List-of Ltype)) Ltype)
  (chase-tevars
   (dcase lterm
     ((Const _ type)
      type)
     ((Bvar i)
      (let ((type (nth i context)))
	(unless type
	  (throw :LOOSE-BVAR nil))
	type))
     ((Evar _ type _ _ _)
      type)
     ((Uvar _ type _ _)
      type)
     ((Appl rator _)
      ;; Must be able to fail for ill-typed terms.
      (dcase (ltype-of rator context)
	((Tfun _ range) range)
	(_ (throw :LOOSE-BVAR nil))))
     ((Abst _ type body)
      (Tfun type (ltype-of body (cons type context)))))))


;;;;	Stamp.

(deftype Stamp ()
  "For recording dependencies between evars and uvars."
  `Integer)

(defvar *current-stamp* 'current-stamp-unbound
  "The current stamp.  Used to record variable dependencies.  Incremented
right before creating a new uvar.")

(defun stamp-init ()
  "Initialize the current stamp, i.e., *current-stamp*."
  (setq *current-stamp* 0))

(stamp-init)

;;; This should be inline expanded later.
(defun stamp-dominates (stamp-a stamp-b)
  "Whether STAMP-A dominates (i.e., came before) STAMP-B.  An evar e cannot be
instantiated to any term containing a uvar or evar whose stamp is dominated
by e's stamp.  The intuition behind the name `dominates' is in the tree
representation of formulas, in which some unifiers are above (dominate) others."
  (declare-ftype stamp-dominates (Stamp Stamp) Boolean)
  (< stamp-a stamp-b))


;;; Making new evars and uvars, putting the right stamps in, and, for uvars,
;;; incrementing the current time stamp.

(defun new-evar (name type)
  "Make a new Evar lterm, given NAME and TYPE, using the current stamp."
  (declare-ftype new-evar (Symbol Ltype) Evar)
  (Evar name type *current-stamp* nil nil))

(defun new-uvar (name type)
  "Make a new Uvar lterm, given NAME and TYPE, incrementing and then using the
current stamp."
  (declare-ftype new-Uvar (Symbol Ltype) Uvar)
  (incf *current-stamp*)
  (Uvar name type *current-stamp* nil))


;;; Uninstantiating a tevar or evar.  Used by the interpreter for trail
;;; management

(defun unbind-var (var)
  "Reset the given tevar or evar VAR to its uninstantiated state."
  (declare-ftype unbind-var ((or Tevar Evar)) )
  (etypecase var
    (Evar (setf (evar-value-slot var) nil))
    (Tevar (setf (tevar-value-slot var) nil)))
  (values))


;;; Copy a term, replacing its loose tbvars and bvars with tevars and evars.

(defun copy-lterm (lterm tevars evars depth)
  "Make a new copy of LTERM, replacing its loose tbvars and bvars with TEVARS
and EVARS, starting at given DEPTH."
  (declare-ftype copy-lterm
    (Lterm (List-of Tevar) (List-of Evar) Nat) Lterm)
  (if (and (null evars) (null tevars))
      lterm
      (labels ((copy (sub-lterm depth)
		 (let ((sub-lterm (chase-evars sub-lterm)))
		   (dcase sub-lterm
		     ((Const cobj type)
		      (re-Const-type sub-lterm cobj type
				     (copy-ltype type tevars)))
		     ((Bvar i)
		      (if (< i depth)
			  sub-lterm
			  (nth (- i depth) evars)))
		     (Evar sub-lterm)
		     (Uvar sub-lterm)
		     ((Appl rator rand)
		      (re-Appl sub-lterm rator rand
			       (copy rator depth)
			       (copy rand depth)))
		     ((Abst vname type body)
		      (re-Abst sub-lterm vname type body
			       vname
			       (copy-ltype type tevars)
			       (copy body (1+ depth))))))))
	(copy lterm depth))))


(defun copy-ltype (ltype tevars)
  "Make a new copy of LTYPE, replacing its loose tbvars with TEVARS."
  (declare-ftype copy-ltype (Ltype (List-of Tevar)) Ltype)
  (if (null tevars)
      ltype
      (labels ((copy (sub-ltype)
		 (let ((sub-ltype (chase-tevars sub-ltype)))
		   (dcase sub-ltype
		     (Tcon
		      sub-ltype)
		     ((Tbvar i)
		      (nth i tevars))
		     (Tevar
		      sub-ltype)
		     ((Tfun domain range)
		      (re-Tfun sub-ltype domain range
			       (copy domain)
			       (copy range)))
		     ((Tappl rator rand)
		      (re-Tappl sub-ltype rator rand
				(copy rator)
				(copy rand)))))))
	(copy ltype))))

(defun new-copy-of-lterm (orig-lterm tvnames vnames vtypes)
  "Make a \"new\" copy of ORIG-LTERM, replacing tbvars and bvars with tevars
and evars, according to TVNAMES, VNAMES, and VTYPES.  Returns three values: (a)
the new term, (b) the new tevars (sequence), and (c) the new evars (sequence)."
  (declare-ftype new-copy-of-lterm
    (Lterm (List-of Symbol) (List-of Symbol) (List-of Ltype))
    Lterm (List-of Tevar) (List-of Evar))
  (let* ((new-tevars (mapcar #'new-tevar tvnames))
	 (new-evars (mapcar #'(lambda (vname vtype)
				(new-evar vname (copy-ltype vtype new-tevars)))
			    vnames vtypes)))
    (values (copy-lterm orig-lterm new-tevars new-evars 0)
	    new-tevars new-evars)))
  
  
(defun ltype-closed-p (ltype)
  "Determine whether LTYPE is closed, i.e., has no (loose) tbvars."
  (declare-ftype ltype-closed-p (Ltype) Boolean)
  (dcase (chase-tevars ltype)
    (Tcon t)
    (Tbvar nil)
    (Tevar t)
    ((Tfun dom ran)
     (and (ltype-closed-p dom)
	  (ltype-closed-p ran)))
    ((Tappl rator rand)
     (and (ltype-closed-p rator)
	  (ltype-closed-p rand)))))


(deftype Ltype-closed ()
  "A closed ltype, i.e., containing no tbvars"
  `(and Ltype (satisfies ltype-closed-p)))



;;; Declare pseudo-constructors for "indirected" terms and types.  Their
;;; selector functions are chase-evars and chase-tevars respectively.  I'll
;;; have to try them a while before I know whether I like it.  (conal)

(declare-constructor iterm (chase-evars) is-lterm)
(declare-constructor itype (chase-tevars) is-ltype)

(ddefun iterm (_) (error "You shouldn't call iterm, just invert it."))
(ddefun itype (_) (error "You shouldn't call itype, just invert it."))


(deftype Noind ()
  "Not an indirected Lterm or Ltype, i.e., not an instantiated (t)evar.  Thus
we don't need to call chase-(t)evars.  Use this e.g., as (and Lterm Noind)."
  `(or (and (not Evar)
	    (satisfies evar-uninstantiated))
       (and (not Tevar)
	    (satisfies tevar-uninstantiated))))

(defun evar-uninstantiated (evar)
  "Whether the EVAR is still uninstantiated."
  (declare-ftype evar-uninstantiated (Evar) Boolean)
  (not (evar-value-slot evar)))

(defun tevar-uninstantiated (tevar)
  "Whether the TEVAR is still uninstantiated."
  (declare-ftype tevar-uninstantiated (Tevar) Boolean)
  (not (tevar-value-slot tevar)))


(defun re-Abst-body (abst-term vname type old-body new-body)
  "More efficient, specialized version of re-Abst."
  (declare-ftype re-Abst-body (Abst Symbol Ltype Lterm Lterm) Abst)
  (if (eq old-body new-body)
      abst-term
      (Abst vname type new-body)))

(defun re-Const-type (const-term cobj old-type new-type)
  "More efficient, specialized version of re-Const"
  (declare-ftype re-Const-type (Const Symbol Ltype Ltype) Const)
  (if (eq old-type new-type)
      const-term
      (Const cobj new-type)))
