;;; -*- Mode: Lisp; Package: LTERM -*-
;;;
;;; Various Lterm operations related to lambda conversion.
;;;
;;; Author: Conal Elliott.  Last Modified Mon Sep 11 06:09:25 1989
;;;
;;; Sccs Id @(#)lterm-convert.lisp	1.28 9/15/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :LTERM) (use-package :ergolisp)


(defun lterm-head (lterm)
  "Find the `head' of LTERM, i.e., skip over initial abstractions and down
rators."
  (declare-ftype lterm-head (Lterm) (and Lterm (not Appl) Noind))
  (labels ((down-rands (lterm)
	     (let ((lterm (chase-evars lterm)))
	       (dcase lterm
		 ((Appl rator _)
		  (down-rands rator))
		 (_ lterm)))))
    (let ((lterm (chase-evars lterm)))
      (dcase lterm
	((Abst _ _ body)
	 (lterm-head body))
	(_
	 (down-rands lterm))))))

(defun head-normal-p (lterm)
  "Is LTERM in beta head-normal form?"
  (declare-ftype head-normal-p (Lterm) Boolean)
  ;; Just see whether the head is not an abstraction.
  (not (is-Abst (lterm-head lterm))))

(deftype Lterm-hnf ()
  "The type of head normal Lterms"
  `(and Lterm
	(satisfies head-normal-p)))

(deftype Lterm-lfhnf ()
  "The type of lambda-free head normal Lterms"
  `(and Lterm-hnf
	(not Abstr)))

(defun add-domains (domains range)
  "Build a term out of nested Tfuns using DOMAINS and RANGE"
  (declare-ftype add-domains ((List-of Ltype) Ltype) Ltype)
  (reduce #'Tfun domains
	  :initial-value range
	  :from-end t))

(defun strip-domains (ltype)
  "Given an LTYPE, return two values: the list of domains (in order from
left to right), and the final (non-Tfun) result type.  An inverse to
add-domains (kind of)."
  (declare-ftype strip-domains (Ltype) (List-of Ltype) Ltype)
  (let ((ltype (chase-tevars ltype)))
    (dcase ltype
      ((Tfun domain range)
       (dlet ( ((values range-domains range-range)
		(strip-domains range)) )
	 (values (cons domain range-domains)
		 range-range)))
      (_ (values () ltype)))))


(defun add-absts (names domains body)
  "Build a term of nested Abst's using NAMES, DOMAINS, and BODY.  If NAMES
is () then some default names are used."
  (declare-ftype add-absts ((List-of Symbol) (List-of Ltype) Lterm) Lterm)
  ;; Uses (rest ()) == ().
  (dcase domains
    (() body)
    ((cons first-dom rest-doms)
     (Abst (if (null names)
	       'elp::|x|
	       (first names))
	   first-dom
	   (add-absts (rest names)
		      rest-doms body)))))

(defun strip-absts (lterm)
  "Given LTERM, strip off the Abst's, returning three values: the list of
abstracted variable names, the list of abstracted variable types, and
the body (which is not an abstraction)."
  (declare-ftype strip-absts (Lterm) (List-of Symbol) (List-of Ltype) Lterm)
  (let ((lterm (chase-evars lterm)))
    (dcase lterm
      ((Abst name type body)
       (dlet ( ((values inner-names inner-types inner-body)
		(strip-absts body)) )
	 (values (cons name inner-names)
		 (cons type inner-types)
		 inner-body)))
      (_ (values () () lterm)))))


(defun add-appls (rator rands)
  "Build a term of nested Appl's using RATOR and RANDS."
  (declare-ftype add-appls (Lterm (List-of Lterm)) Lterm)
  (reduce #'Appl rands :initial-value rator))

(defun strip-appls (lterm)
  "Given a first-order LTERM with no loose bvars, strip off the nested
applications, returning two values: a non-Appl rator and a list of rands.
An inverse to add-appls (kind of)."
  (declare-ftype strip-appls (Lterm) Head (List-of Lterm))
  ;; Implement in terms of accumulator function.
  (labels ((strip-rec (lterm rands-so-far)
	     (let ((lterm (chase-evars lterm)))
	       (dcase lterm
		 ((Appl rator rand)
		  (strip-rec rator (cons rand rands-so-far)))
		 (_ (values lterm rands-so-far))))))
    (strip-rec lterm ())))

(ddefun lterm-rands (lterm)
  "Get the rands from an LTERM."
  (declare-ftype lterm-rands (Lterm) (List-of Lterm))
  (dlet* ( ((values _ _ body) (strip-absts lterm))
	   ((values _ rands) (strip-appls body)) )
    rands))

;;; Use of these two functions should be eliminated in favor of using the head
;;; normal form terms themselves.

(defun to-fohnf (lterm)
  "Given a first-order LTERM, return head normal form as two values: a Head
rator and a list of rands."
  (declare-ftype to-fohnf (Lterm) Head (List-of Lterm))
  (dlet ( ((values _ types rator rands) (to-hnf lterm)) )
    (assert (null types) ()
      "TO-FOHNF: The term ~s is not first-order." lterm)
    (values rator rands)))

(defun to-hnf (lterm)
  "Given an LTERM, return beta head normal form as four values: a list of
abstracted variable names, a list of abstracted variable types, a Head rator
and a list of rands."
  (declare-ftype to-hnf
    (Lterm) (List-of Symbol) (List-of Ltype) (or Head Bvar) (List-of Lterm))
  ;; Just do the normalizations and strip off the absts and appls.
  (dlet ( ((values names types body)
	   (strip-absts (head-beta-norm lterm))) )
    (dlet ( ((values rator rands)
	     (strip-appls body)) )
      (values names types rator rands))))


(defun head-beta-norm (lterm)
  "Convert LTERM to head beta normal form."
  (declare-ftype head-beta-norm (Lterm) Lterm)
  ;; We don't need a context, because types play no role in beta-reduction.
  (let ((lterm (chase-evars lterm)))
    (dcase lterm
      (Bvar lterm)
      (Evar lterm)
      (Uvar lterm)
      (Const lterm)
      ((Abst name type body)
       (re-Abst-body lterm name type body (head-beta-norm body)))
      ((Appl rator rand)
       (let ((new-rator (head-beta-norm rator)))
	 (dcase new-rator
	   ((Abst name _ body)
	    ;; Radically new, the next line is.
	    (when (is-Uvar rand) (setf (Uvar-name rand) name))
	    (head-beta-norm (bvar-subst body rand)))
	   (_ (re-Appl lterm rator rand
		       new-rator rand))))))))


(defun beta-norm (lterm)
  "Convert LTERM to indirection-free beta normal form."
  (declare-ftype beta-norm (Lterm) (and Lterm Noind))
  ;; We don't need a context, because types play no role in beta-reduction.
  (let ((lterm (chase-evars lterm)))
    (dcase lterm
      (Bvar lterm)
      (Evar lterm)
      (Uvar lterm)
      (Const lterm)
      ((Abst name type body)
       (re-Abst-body lterm name type body (beta-norm body)))
      ((Appl rator rand)
       (let ((new-rator (beta-norm rator))
	     (new-rand (beta-norm rand)))
	 (dcase new-rator
	   ((Abst name _ body)
	    ;; Radically new, the next line is.
	    (when (is-Uvar new-rand) (setf (Uvar-name new-rand) name))
	    (beta-norm (bvar-subst body new-rand)))
	   (_ (re-Appl lterm rator rand new-rator new-rand))))))))



(defun bvar-subst (body rand)
  "Replace the bvar 0 in BODY by RAND, doing relocation of RAND if needed
and decrementing the other loose variables in body."
  (declare-ftype bvar-subst (Lterm Lterm) Lterm)
  ;; This probably ought to be optimized to first check rand to see if it
  ;; contains any loose bvars, and only relocate if it does.
  (labels ((subst-rec (lterm depth)
	     (let ((lterm (chase-evars lterm)))
	       (dcase lterm
		 (Const lterm)
		 ((Bvar i)
		  (cond ((< i depth)
			 lterm)
			((= i depth)
			 (relocate rand i))
			(t (Bvar (1- i)))))
		 (Evar lterm)
		 (Uvar lterm)
		 ((Appl rator rand)
		  (re-Appl lterm rator rand
			   (subst-rec rator depth)
			   (subst-rec rand depth)))
		 ((Abst vname type body)
		  (re-Abst-body lterm vname type
				body
				(subst-rec body (1+ depth))))))))
    (subst-rec body 0)))

(defun relocate (rel-rand rator-depth)
  "Relocate the loose bvars in REL-RAND by adding RATOR-DEPTH."
  (declare-ftype relocate (Lterm Nat) Lterm)
  (if (= rator-depth 0)
      rel-rand
      (labels ((reloc-rec (rel-rand rand-depth)
		 (let ((rel-rand (chase-evars rel-rand)))
		   (dcase rel-rand
		     (Const rel-rand)
		     ((Bvar j)
		      (if (< j rand-depth) ; Not loose
			  rel-rand
			  (Bvar (+ j rator-depth))))
		     (Evar rel-rand)
		     (Uvar rel-rand)
		     ((Appl rator rand)
		      (re-Appl rel-rand rator rand
			       (reloc-rec rator rand-depth)
			       (reloc-rec rand rand-depth)))
		     ((Abst vname type body)
		      (re-Abst-body rel-rand vname type
				    body
				    (reloc-rec body (1+ rand-depth))))))))
	(reloc-rec rel-rand 0))))


(defun convert-to-abst (lterm context)
  "Convert LTERM, which must be higher-order, in the given CONTEXT, to an
abstraction."
  (declare-ftype convert-to-abst (Lterm (List-of Ltype)) Lterm)
  ;; Is is already an abstraction?
  (if (is-abst lterm)
      lterm
      ;; Is its hnf an abstraction?
      (let ((lterm (head-beta-norm lterm)))
	(if (is-abst lterm)
	    lterm
	    ;; Eta-expand it.
	    (dlet ( ((Tfun domain _) (ltype-of lterm context)) )
	      (Abst 'elp::|x| domain
		    (Appl (relocate lterm 1) (Bvar 0))))))))


(defun map-rands (function lterm)
  "Map the FUNCTION over the evar-chased rands of LTERM, producing a Noind
lterm with the same lambdas and head and mapped rands."
  (declare-ftype map-rands (Lterm) (and Lterm Noind))
  (labels ((rec-absts (lterm)
	     (let ((lterm (chase-evars lterm)))
	       (dcase lterm
		 ((Abst vname type (iterm body))
		  (re-Abst-body lterm vname type
				body
				(rec-absts body)))
		 (_ (rec-appls lterm)))))
	   (rec-appls (body)		; body : Noind
	     (dcase body
	       ((Appl (iterm rator) (iterm rand))
		(re-Appl body rator rand
			 (rec-appls rator)
			 (funcall function rand)))
	       (_ body))))
    (rec-absts lterm)))


(defun head-eta-reduce (lterm)
  "Put LTERM in head-eta-reduced form."
  (dcase (chase-evars lterm)
    ((Abst _ _ body)
	 (dcase (head-eta-reduce body)
	   ((Appl rator rand)
	    (if (depends rator 0)
		lterm
		(dcase (head-eta-reduce rand)
		  ((Bvar i)
		   (if (= i 0)
		       (relocate rator -1)
		       lterm))
		  (_ lterm))))
	   (_ lterm)))
    (_ lterm)))


(tdefun depends (lterm mbvar-num)
	"Lterm # Nat -> Boolean"
  "Find out whether LTERM depends on (Mbvar MBVAR-NUM)."
  ;; Don't bother looking inside evars, since their values do not have loose
  ;; variables.
  (dcase lterm
    ((Bvar index)
     (= index mbvar-num))
    ((Appl rator rand)
     (or (depends rator mbvar-num)
	 (depends rand mbvar-num)))
    ((Abst _ _ mbody)
     (depends mbody (1+ mbvar-num)))
    ((Appl rator rand)
     (or (depends rator mbvar-num)
	 (depends rand mbvar-num)))
    (_ nil)))
