;;; -*- Mode: Lisp; Package: LTERM -*-
;;;
;;; Translate from lterms to {g,d}forms and {g,d}forms to slterms.
;;;
;;; Author: Conal Elliott.  Last Modified Fri Sep 29 22:43:09 1989
;;;
;;; Sccs Id @(#)lterm-form.lisp	1.1 9/12/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :LTERM) (use-package :ergolisp)

;;; These defparameter's should be defconstants, but the compiler
;;; complains so.

(defparameter *prop-type*
  (Tcon 'elp::|o| *ktype*)
  "The ltype of LP propositions.")

(eval-when (compile load eval)
(defconstant *true* 'elp::|true|
  "The constant standing for truth.")

(defconstant *and* 'elp::|,|
  "The constant standing for conjunction.")

(defconstant *or* 'elp::|;|
  "The constant standing for disjunction.")

(defconstant *implies* 'elp::|=>|
  "The constant standing for implication.")

(defconstant *reverse-implies-hack* 'elp::|:-|
  "The constants that hacks reverse implication for unparsing of Dforms.")

(defconstant *pi* 'elp::|pi|
  "The constant standing for universal quantification.")

(defconstant *sigma* 'elp::|sigma|
  "The constant standing for existential quantification.")
)
  
(defparameter *logical-primitives*
  '(#.*true* #.*and* #.*or* #.*implies* #.*pi* #.*sigma*)
  "A list of logical primitives.")

(defun mkc (symbol ltype)
  (Const (Named-lcobj symbol) ltype))

(defparameter *prop-binop-type* (Tfun *prop-type* *prop-type*))

(defun lterm-to-gform (lterm context)
  "Convert LTERM to a Gform, for the LP interpreter.  Assumes that
LTERM has propositional type."
  (declare-ftype lterm-to-gform (Lterm (List-of Ltype)) elp:Gform)
  (let ((lterm (head-beta-norm lterm)))
    ;; Separate into head and args
    (dlet ( ((values head args) (strip-appls lterm)) )
      (dcase head
	;; Constant heads.  either a logical construction or a (rigid)
	;; atomic formula.
	((Const (:as lcobj (named-lcobj name)) _)
	 (let ((special (elp:special-lookup name)))
	   (if special
	       (elp:Gspecial special
			     (elp:rep-special special args context))
	       (case name
		 (#.*true*
		  (elp:Gtrue))
		 (#.*and*
		  (elp:Gand (lterm-to-gform (first args) context)
			    (lterm-to-gform (second args) context)))
		 (#.*or*
		  (elp:Gor (lterm-to-gform (first args) context)
			   (lterm-to-gform (second args) context)))
		 (#.*pi*
		  (elp:Gall (lterm-abst-to-gform (first args) context)))
		 (#.*sigma*
		  (elp:Gexists (lterm-abst-to-gform (first args) context)))
		 (#.*implies*
		  (elp:Gimplies (lterm-to-dform (first args) context)
				(lterm-to-gform (second args) context)))
		 (t
		  ;; Probably this should make use of the head normalization
		  ;; work. 
		  (elp:Gatom lcobj lterm))))))
	;; Non constant heads.
        ((Bvar _)
	 (elp:Gflex lterm))
	(Evar
	 (elp:Gflex lterm))
	(Uvar
	 (elp:Gatom head lterm))))))

(defun lterm-abst-to-gform (lterm context)
  "Convert functional LTERM to a Gform.  Used by Gall, Gexists, and some
specials."
  (declare-ftype lterm-abst-to-gform (Lterm (List-of Ltype)) elp:Gform)
  #|;; old version.
  (dlet ( ((Abst name type body)
	   (convert-to-abst lterm context)) )
    (elp:Glam name type (lterm-to-gform body (cons type context))))
  |#
  (lterm-absts-to-gform 1 lterm context))

(defun lterm-absts-to-gform (n lterm context)
  "Convert functional LTERM of N arguments to a Gform.  May be used by
specials."
  (declare-ftype lterm-absts-to-gform (Nat Lterm (List-of Ltype)) elp:Gform)
  (if (= n 0)
      (lterm-to-gform lterm context)
      (dlet ( ((Abst name type body)
	       (convert-to-abst lterm context)) )
	(elp:Glam name type
		  (lterm-absts-to-gform (1- n) body (cons type context))))))

;;; Changed this function to also do the normalization into a
;;; conjunction of universally quantified implications of gforms
;;; to atoms.  This is an optimization as well as a bug fix, since
;;; relocation cannot easily be done on gforms (see function
;;; install-dform-clauses).

(defun lterm-to-dform (lterm context)
  "Convert LTERM to a Dform for the LP interpreter.  Assumes that
LTERM has propositional type."
  (declare-ftype lterm-to-dform (Lterm (List-of Ltype)) elp:Dform)
  (labels ((normdform (lterm vnames vtypes assums)
             ;; This function passes VTYPES (which is initially the context)
	     ;; and VNAMES, which is the name of the variables in the initial
	     ;; segment of the VTYPES.
             (let ((lterm (head-beta-norm lterm)))
	       ;; Separate into head and args
	       (dlet ( ((values head args) (strip-appls lterm)) )
		 (dcase head
		   ;; Constant heads.  either a logical construction or a
		   ;; (rigid) atomic formula.
		   ((Const (:as lcobj (named-lcobj name)) _)
		    (when (elp:special-lookup name)
		      (elp:lp-error "Special predicates are not legal assumptions.
 Here: ~A." lterm))
		    (case name
		      (#.*and*
		       (elp:Dand (normdform (first args) vnames vtypes assums)
				 (normdform (second args) vnames vtypes assums)))
		      (#.*pi*
		       (dlet ( ((Abst name type body)
				(convert-to-abst (first args) context)) )
			 (normdform body
				    (cons name vnames) (cons type vtypes)
				    (if (not (null vtypes))
					(mapcar #'(lambda (lterm)
						    (relocate lterm 1))
						assums)
					assums))))
		      (#.*implies*
		       (normdform (second args) vnames vtypes
				  (cons (first args) assums)))
		      (#.*true*
		       (elp:lp-error "~S is not a legal assumption." lterm))
		      (#.*or*
		       (elp:lp-error "Disjunctions are not legal assumptions.
 Here: ~A." lterm))
		      (#.*sigma*
		       (elp:lp-error "Existentially quantified formulas are not legal assumptions.
 Here: ~A." lterm))
		      (t
		       ;; Rigid atom: build the normal dform.
		       (build-normdform lcobj lterm
					vnames vtypes assums))))
		   ;; Non constant heads.  Either rigid or flexible head.
		   ((Bvar _) (build-flexdform lterm vnames vtypes assums))
		   (Evar (build-flexdform lterm vnames vtypes assums))
		   (Uvar (build-normdform head lterm
					  vnames vtypes assums))
		   )))))
    (normdform lterm () context ())))

(defun build-normdform (head lterm vnames vtypes assums)
  "Build a normal dform from lterm using quantifiers in VNAMES and VTYPES
and ASSUMS."
  (labels ((build-impl (assums clause-body)
	     (dcase assums
	       (() (elp:Dimplies clause-body (elp:Datom head lterm)))
	       ((cons assum rest)
		(build-impl rest (elp:Gand (lterm-to-gform assum vtypes)
					   clause-body)))))
	   (build-univ (vnames vtypes clause)
	     (dcase vnames
	       (() clause)
	       ((cons vname rest)
		(dlet ( ((cons vtype trest) vtypes) )
		  (build-univ rest trest
			      (elp:Dall (elp:Dlam vname vtype clause))))))))
    (build-univ vnames vtypes
		(if (null assums)
		    (elp:Datom head lterm)
		    (build-impl (rest assums)
				(lterm-to-gform (first assums) vtypes))))))

(defun build-flexdform (lterm vnames vtypes assums)
  "Build a flexible Dform from lterm, using VNAMES, VTYPES and ASSUMS.
The flexible term is pre-normalized, but the eventual predicate at the 
clause head is flexible and can thus not be fully normalized yet."
  (labels ((build-impl (assums clause-body)
		       (dcase assums
			 (() (Appl (Appl (mkc *implies* *prop-binop-type*)
					 clause-body)
				   lterm))
			 ((cons assum rest)
			  ;; Case of ``true''? -fp
			  (build-impl rest
				      (Appl (Appl (mkc *and* *prop-binop-type*)
						  assum)
					    clause-body)))))
	   (build-univ (vnames vtypes clause)
		       (dcase vnames
			 (() clause)
			 ((cons vname rest)
			  (dlet ( ((cons vtype trest) vtypes) )
			    (build-univ rest trest
					(Appl (mkc *pi* (Tfun (Tfun vtype *prop-type*)
								 *prop-type*))
					      (Abst vname vtype clause))))))))
    (elp:Dflex
     (build-univ vnames vtypes
		 (if (null assums)
		     lterm
		     (build-impl (rest assums) (first assums)))))))

;;;; Construction of terms from formula.

(defun gform-to-slterm (gform vtypes tvnames vnames depth)
  "Convert GFORM to a slterm using the VNAMES.  Used for tracing and debugging."
  (declare-ftype gform-to-slterm (elp:Gform (List-of Ltype) (List-of Symbol)
					    (List-of Symbol))
		 slterm:Slterm)
  (if (= depth *elp-print-depth*)
      (slterm:EllidedSlterm)
      (dcase gform
	((elp:Gflex lterm)
	 ;; Convert lterm to slterm.
	 (to-slterm lterm vtypes tvnames vnames depth))
	;; Otherwise, get the head and args, where the head is either an lcobj
	;; or a uvar.  Then put them together into a slterm.
	(_ (dcase gform
	     (elp:Gtrue
	      (name-args *true* ()))
	     ((elp:Gand left right)
	      (name-args *and*
			 (list (gform-to-slterm left vtypes tvnames vnames (1+ depth))
			       (gform-to-slterm right vtypes tvnames vnames (1+ depth)))))
	     ((elp:Gor left right)
	      (name-args *or*
			 (list (gform-to-slterm left vtypes tvnames vnames (1+ depth))
			       (gform-to-slterm right vtypes tvnames vnames (1+ depth)))))
	     ((elp:Gall qarg)
	      (name-args *pi*
			 (list (gform-to-slterm qarg vtypes tvnames vnames (1+ depth)))))
	     ((elp:Gexists qarg)
	      (name-args *sigma*
			 (list (gform-to-slterm qarg vtypes tvnames vnames (1+ depth)))))
	     ((elp:Gimplies dassum gconcl)
	      (name-args *implies*
			 (list (dform-to-slterm dassum vtypes tvnames vnames (1+ depth))
			       (gform-to-slterm gconcl vtypes tvnames vnames (1+ depth)))))
	     ((elp:Gspecial special args)
	      (name-args (elp::special-name special)
			 (elp:unrep-special special args vtypes tvnames vnames (1+ depth))))
	     ((elp:Gatom _ lterm)
	      (to-slterm lterm vtypes tvnames vnames depth))
	     ((elp:Glam vname type body)
	      (slterm-abst-to-form vname type vtypes tvnames vnames body depth
				   #'gform-to-slterm)))))))

(defun dform-to-slterm (dform vtypes tvnames vnames depth)
  "Convert DFORM to a slterm using the VNAMES.  Used for tracing and debugging."
  (declare-ftype dform-to-slterm (Dform (List-of Ltype)
					(List-of Symbol) (List-of Symbol))
		 slterm:Slterm)
  (if (= depth *elp-print-depth*)
      (slterm:EllidedSlterm)
      (dcase dform
	((elp:Dflex lterm)
	 (to-slterm lterm vtypes tvnames vnames depth))
	;; Otherwise, get the head and args, where the head is either an lcobj
	;; or a uvar.  Then put them together into a slterm.
	(_ (dcase dform
	     ((elp:Dand left right)
	      (name-args *and*
			 (list (dform-to-slterm left vtypes tvnames vnames (1+ depth))
			       (dform-to-slterm right vtypes tvnames vnames (1+ depth)))))
	     ((elp:Dall qarg)
	      (name-args *pi*
			 (list (dform-to-slterm qarg vtypes tvnames vnames (1+ depth)))))
	     ((elp:Dimplies gassum dconcl)
	      #| (name-args *implies*
		     (list (gform-to-slterm gassum vtypes tvnames vnames (1+ depth))
			   (dform-to-slterm dconcl vtypes tvnames vnames (1+ depth)))) |#
	      (name-args *reverse-implies-hack*
			 (list (dform-to-slterm dconcl vtypes tvnames vnames (1+ depth))
			       (gform-to-slterm gassum vtypes tvnames vnames (1+ depth)))))
	     ((elp:Datom _ lterm)
	      (to-slterm lterm vtypes tvnames vnames depth))
	     ((elp:Dlam vname type body)
	      (slterm-abst-to-form vname type vtypes tvnames vnames body depth
				   #'dform-to-slterm)))))))

(defun slterm-abst-to-form (old-vname ltype vtypes tvnames vnames form-body
				      depth convert)
  "Make a slterm abst given OLD-VNAME, TYPE, VTYPES, TVNAMES, VNAMES,
FORM-BODY, and function CONVERT to be applied recursively to the body.
Check the value of *to-slterm-with-types-p*."
  (declare-ftype slterm-abst-to-form
    (Symbol Ltype (List-of Ltype) (List-of Symbol) (List-of Symbol)
	    (Or Dform Gform)
	    (Function-of ((Or Dform Gform)) slterm:Slterm))
    slterm:Slterm)
  (let ((new-vname (choose-vname old-vname vnames)))
    (slterm:Abst (if *to-slterm-with-types-p*
		     (slterm:Vtyped new-vname
				    (to-sltype ltype tvnames depth))
		     (slterm:Vuntyped new-vname))
		 (funcall convert form-body
			  (cons ltype vtypes) tvnames (cons new-vname vnames)
			  (1+ depth)))))

(defun name-args (name args)
  "Build a slterm out the given NAME for a named constant head and the given
ARGS."
  (add-sappls (slterm:var name) args))

(defun add-sappls (rator rands)
  "Like add-apple, but with slterms."
  (declare-ftype add-sappls
    (slterm:Slterm (List-of slterm:Slterm)) slterm:Slterm)
  (reduce #'slterm:Appl rands :initial-value rator))


(defun module-to-slterm (module)
  "Convert MODULE to an lterm."
  (declare-ftype module-to-slterm (Module) slterm:Slterm)
  (slterm:Con (slterm:Cnamed (intern (elp:module-name module) :ELP))))

;;; Construction of REAL terms (not syntactic ones) from formulas.

;;; Must make lcobj's out of the constants below.  Only difficulty: type

(defun gform-to-lterm (gform)
  "Convert GFORM to an lterm."
  (declare-ftype gform-to-lterm (elp:Gform) lterm:Lterm)
  (dcase gform
    (elp:Gtrue
     (mkc *true* *prop-type*))
    ((elp:Gand left right)
     (Appl (Appl (mkc *and* *prop-binop-type*)
		 (gform-to-lterm left))
	   (gform-to-lterm right)))
    ((elp:Gor left right)
     (Appl (Appl (mkc *or* *prop-binop-type*)
		 (gform-to-lterm left))
	   (gform-to-lterm right)))
    ((elp:Gall (elp:Glam vname type body))
     (Appl (mkc *pi* (Tfun (Tfun type *prop-type*) *prop-type*))
	   (Abst vname type (gform-to-lterm body))))
    ((elp:Gexists (elp:Glam vname type body))
     (Appl (mkc *sigma* (Tfun (Tfun type *prop-type*) *prop-type*))
	   (Abst vname type (gform-to-lterm body))))
    ((elp:Gimplies dassum gconcl)
     (Appl (Appl (mkc *implies* *prop-binop-type*)
		 (dform-to-lterm dassum))
	   (gform-to-lterm gconcl)))
    ((elp:Gspecial _ _)
     (elp:lp-error "Special ~S cannot be translated to a term!" gform))
    ((elp:Gatom _ lterm)
     lterm)
    ((elp:Gflex lterm)
     lterm)
    ((elp:Glam _ _ _)
     (elp:lp-error "Glam ~S in unexpected position." gform)
     )))

(defun dform-to-lterm (dform)
  "Convert DFORM to an lterm."
  (declare-ftype dform-to-lterm (elp:Dform) lterm:Lterm)
  (dcase dform
    ((elp:Dand left right)
     (Appl (Appl (mkc *and* *prop-binop-type*)
		 (dform-to-lterm left))
	   (dform-to-lterm right)))
    ((elp:Dall (elp:Dlam vname type body))
     (Appl (mkc *pi* (Tfun (Tfun type *prop-type*) *prop-type*))
	   (Abst vname type (dform-to-lterm body))))
    ((elp:Dimplies gassum dconcl)
     (Appl (Appl (mkc *implies* *prop-binop-type*)
		 (gform-to-lterm gassum))
	   (dform-to-lterm dconcl)))
    ((elp:Datom _ lterm)
     lterm)
    ((elp:Dflex lterm)
     lterm)
    ((elp:Dlam _ _ _)
     (elp:lp-error "Dlam ~S in unexpected position." dform)
     )))
