;;; -*- Mode: Lisp; Package: LTERM -*-
;;;
;;; Translate slterms into lterms, inserting de Bruijn indices and doing type
;;; inference.
;;;
;;; Author: Conal Elliott.  Last Modified Fri Sep 29 11:08:13 1989
;;;
;;; Sccs Id @(#)lterm-from-slterm.lisp	1.2 9/29/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :LTERM) (use-package :ergolisp)

(export '(Lterm))

;;; I don't thing anything really should be exported here, since the tlang
;;; interface contains what is needed.
(export '(from-slterm from-sltermlist from-slkind ltyfam-from-sltype))

;;;; It got quite burdensome to pass around the lists of free term and type
;;;; variables and their lengths.  What I do instead is to make these
;;;; dynamic variables.

;;;;	Vcontext

(deftype Vcontext ()
  "An alist that maps names into ltypes."
  `(Alist-of Symbol Ltype))

(defparameter harmless-ltype (new-tevar 'harmless))

(defun vcontext-lookup (name vars)
  "Look up NAME in a vcontext VARS.  Returns two values.  If the name is found,
these values are the position from the beginning and the type.  Otherwise,
the first value is nil."
  ;; Even though this makes two passes over the alist, it will probably be
  ;; faster than anything one could write in Lisp.
  (declare-ftype vcontext-lookup (Symbol Vcontext) (Optional Nat) Ltype)
  (let ((pair (assoc name vars)))
    (if pair
	(values (position pair vars :test #'eq)
		(cdr pair))
	(values nil harmless-ltype))))


;;;; Do the translation.

;;;; This is new to do the type inference for undeclared constants.

(defvar *translation-modules* 'translation-modules-unbound
  "Modules to use for signature lookups while translating.")

(defsconstr "DATATYPE Translate-error of T, has : Ltype, wants : Ltype ")

(defun from-slterm (slterm ltype modules)
  "Translate SLTERM into an lterm having given LTYPE, using given MODULES for
signature lookups.  In particular, put in de Bruijn indices, and do type
inference.  Returns (1) the constructed lterm, (2) a list of the names of the
free type variables in the constructed term, (3) a list of the names of the
free variables in the term, and (3) a corresponding list of the ltypes of those
variables.  The de Bruijn indices for the free term and type variables is their
position in the second and third returned values."
  (declare-ftype from-slterm
    (slterm:Slterm Ltype (List-of elp:Module))
    Lterm (List-of Evar) (List-of Uvar) (List-of Tevar))
  (let ((*translation-modules* modules)
	(*reading-module-p* nil))
    (declare (special *translation-modules* *reading-module-p*))
    (let ((*free-evars* ())
	  (*free-uvars* ())
	  (*free-tevars* ())
	  (*free-tcons* ()))
      (declare (special *free-evars* *free-uvars* *free-tevars* *free-tcons*))
      ;; The latter two may be currently unused? -fp.
      (let ((*bound-vars* ())
	    (*bound-count* 0))
	(declare (special *bound-vars* *bound-count*))
	(let ((lterm-or-err (catch 'Translate-error
			      (from-slterm-rec slterm ltype))))
	  (dcase lterm-or-err
	    ((Translate-error thing has wants)
	     (elp:lp-error "Type inference failure while analyzing ~S.
Subterm ``~A'' has type ~S,
but is expected to have type ~S."
			   slterm thing has wants))
	    ;; Now: choose the right time-stamp for the Evars.
	    (_ (dolist (ev *free-evars*)
		 (setf (evar-stamp ev) *current-stamp*))
	       (lterm-trim-tevars lterm-or-err)
	       (values lterm-or-err
		       *free-evars*
		       *free-uvars*
		       *free-tevars*
		       *free-tcons*))))))))

(defun from-sltermlist (sltermlist modules)
  "Translate SLTERMLIST into an lterm list, using given MODULES for
signature lookups.  In particular, put in de Bruijn indices, and do type
inference.  Returns a list of LTERMS in their contexts."
  (declare-ftype from-sltermlist
    (slterm:Slterm (List-of elp:Module))
    (List-of (List-of Symbol) (List-of Symbol) (List-of Ltype) Lterm))
  (let ((*translation-modules* modules)
	(*reading-module-p* t)
	(*undeclared-constants* ())
	(*undeclared-type-constants* ()))
    (declare (special *translation-module* *reading-module-p*
		      *undeclared-constants* *undeclared-type-constants*))
    (let ((ltermlist-1 (from-sltermlist-1 sltermlist ())))
      ;; Now: stringent type inference.
      (let ((udcl-alist (fix-undeclared-constants *undeclared-constants*))
	    (udtcl-alist (fix-undeclared-type-constants *undeclared-type-constants*)))
	(let ((fixed-dformlist (quant-tevars-in-ltermlist ltermlist-1 ())))
	  (values fixed-dformlist udcl-alist udtcl-alist))))))

(defun from-sltermlist-1 (sltermlist ltermlist)
  (dcase sltermlist
    (() ltermlist)
    ((cons slterm sltermlist-rest)
     (let ((*free-vars* ())
	   (*free-count* 0)
	   (*free-tvars* ()))
       (declare (special *free-vars* *free-count* *free-tvars*))
       (let ((*bound-vars* ())
	     (*bound-count* 0))
	 (declare (special *bound-vars* *bound-count*))
	 (let ((lterm-or-err (catch 'Translate-error
			       (from-slterm-rec slterm *prop-type*))))
	   (dcase lterm-or-err
	     ((Translate-error sub-slterm has wants)
	      (elp:lp-error "Type inference failure while analyzing ~S.
Subterm ``~A''  has type ~S,
but was expected to have type ~S."
			    slterm sub-slterm has wants))
	     (_ (from-sltermlist-1 sltermlist-rest
				   (cons (list *free-vars* lterm-or-err)
					 ltermlist))))))))))

(defun fix-undeclared-type-constants (tcon-alist)
  (mapcar (fdlambda ( (cons name (Tcon _ kind)) ) (cons name kind) )
	  tcon-alist))

(defun fix-undeclared-constants (ltype-alist)
  ;; Unify types of Polytycon's in the list.
  (create-ltyfam-list (mapcar #'fix-udcl-con ltype-alist) ()))

(defun create-ltyfam-list (udcls so-far)
  (dcase udcls
    (() so-far)
    ((cons (cons name ltype) udcls-rest)
     (dlet ( ((values bvar-names qltype) (quant-tevars-in-ltype ltype)) )
       (create-ltyfam-list udcls-rest
			   (cons (cons name (Ltyfam bvar-names qltype))
				 so-far))))))

(ddefun fix-udcl-con ((cons name constraints))
  ;; unify-ltype has a side-effect, so we unify successively with the first
  ;; constraint.
  (dlet (((cons ty1 rest-constraints) constraints))
    (let ((ty (every #'(lambda (ty2) (unify-ltype ty1 ty2 nil))
		     rest-constraints)))
      (if (null ty)
	  (let ((lax-type;;(ltypelist-anti-unify constraints)
		 (new-tevar 'elp::|A|)))
	    (elp:lp-warn t "Stringent type inference for constant ~A failed.
Types ~{~8T~S~%~} could not be unified.
Assigning the super-lax type for ~A as ~S"
			 name (cons ty1 constraints) name lax-type)
	    (cons name;;(ltyfam-instance lax-type)
		  lax-type))
	  (cons name ty1)))))

(defun unify-ltype-or-bust (thing has wants)
  ;; Unify the ltypes, but signal an error on failure.  We don't
  ;; care about the unification trail.  MAYBE WE DO, TO GIVE
  ;; SENSIBLE ERROR MESSAGES.
  (unless (unify-ltype has wants nil)
    (throw 'Translate-error (Translate-error thing has wants))))

(defun from-slterm-rec (slterm ltype)
  "Do the translation, given SLTERM and the LTYPE it must have.
Returns the constructed slterm.  Side-effects *free-vars*, etc, which are
assumed to be set up.  If there is an error, throw a Translate-error structure
to 'Translate-error."
  ;; Note that, we count from the beginning for *bound-vars*, but from the
  ;; end for *free-vars*.
  (declare-ftype from-slterm-rec (slterm:Slterm Ltype) Lterm)
  (dcase slterm
    ((slterm:Con slcobj)
     (lterm-from-slcobj slcobj ltype))
    ((slterm:Var name)
     ;; First see if it is a saved evar or uvar.
     (lterm-from-named-thing name ltype))
    ((slterm:Appl slrator slrand)
     ;; The type of the rator must be "A -> ltype", and the type of the rand
     ;; must be A.
     (let ((dom-type (new-tevar 'elp::|A|)))
       (let ((lrator (from-slterm-rec slrator (Tfun dom-type ltype))))
	 (let ((lrand (from-slterm-rec slrand dom-type)))
	   (Appl lrator lrand)))))
    ((slterm:Abst vbind slbody)
     ;; sltype must be A -> B, where B is the type of the vbind.  Then the
     ;; body must have type B, in the context of vbind.
     (dlet ( ((values name domain)
	      (dcase vbind
		((slterm:Vuntyped vname)
		 (values vname (new-tevar 'elp::|A|)))
		((slterm:Vtyped vname sltype)
		 (values vname (from-sltype-rec sltype))))) )
       (let ((range (new-tevar 'elp::|A|)))
	 (progn (unify-ltype-or-bust slterm (Tfun domain range) ltype)
		(let ((*bound-vars* (acons name domain *bound-vars*))
		      (*bound-count* (1+ *bound-count*)))
		  (declare (special *bound-vars* *bound-count*))
		  (let ((lbody (from-slterm-rec slbody range)))
		    (Abst name domain lbody)))))))
    ((slterm:Typed new-slterm new-sltype)
     (prog1
	 (from-slterm-rec new-slterm ltype)
       (unify-ltype-or-bust slterm ltype (from-sltype-rec new-sltype))))
    (_ (elp:lp-error "Illegal abstract syntax constructed upon parsing.
This should not happen unless you were trying to tickle this bug."))))

(defun lterm-from-slcobj (slcobj ltype)
  "Translate given SLCOBJ to a constant lterm."
  (declare-ftype lterm-from-slcobj (slterm:Slcobj) Lterm)
  (dcase slcobj
    ((slterm:Cstring string)
     (unify-ltype-or-bust slcobj *string-ltype* ltype)
     (Const (String-lcobj string) *string-ltype*))
    ((slterm:Cnumber number)
     (unify-ltype-or-bust slcobj *number-ltype* ltype)
     (Const (Number-lcobj number) *number-ltype*))
    ((slterm:Cnamed _)
     (error "Translation should no longer generate a Cnamed, only Var's."))))

(defun case-of-name (symbol)
  (let ((first-char (schar (symbol-name symbol) 0)))
    (if (upper-case-p first-char)
	:uppercase
	:lowercase)))

(defun lterm-from-named-thing (name ltype)
  (declare (special *reading-module-p* *translation-modules*)
	   (special *bound-vars*)
	   (special *free-evars* *free-uvars*)
	   (special *undeclared-constants*))
  (dlet ( ((values bfound-depth bfound-type)
	   (vcontext-lookup name *bound-vars*)) )
    (if bfound-depth
	(progn (unify-ltype-or-bust name bfound-type ltype)
	       ;; de Bruijn indices start at 0.
	       (Bvar bfound-depth))
	;; if not a bound variable, it may be a known constant.
	(let ((ltyfam (elp:modules-sig-get
		       name
		       *translation-modules*
		       #'elp:module-typesig)))
	  ;; If there is a tyfam, use it by instantiating it.
	  (if ltyfam
	      (let ((ltyfam-inst (ltyfam-instance ltyfam)))
		(unify-ltype-or-bust name ltyfam-inst ltype)
		(Const (Named-lcobj name)
		       ltyfam-inst))
	      ;; If not a known constant, it depends on whether we
	      ;; are reading a module or not.
	      (if (not *reading-module-p*)
		  ;; If NOT reading a module, check if the variable
		  ;; is known as Uvar or Evar.
		  (let ((found-euvar
			 (ident:ident-retrieve-check (symbol-name name)
						     '(or Evar Uvar))))
		    (if found-euvar
			(dcase found-euvar
			  ((Evar _ type _ _ _)
			   (unify-ltype-or-bust name type ltype)
			   found-euvar)
			  ((Uvar _ type _ _)
			   (unify-ltype-or-bust name type ltype)
			   found-euvar))
			;; If not found, it's either an Evar or Uvar.
			(ecase (case-of-name name)
			  (:uppercase
			   (let ((nevar (Evar name ltype
					      nil ; this is the stamp: invalid.
					      nil nil)))
			     ;; Remember: the time stamp MUST be changed
			     ;; later!!!
			     (setf (evar-ident nevar)
				   (ident:new-ident
				    (symbol-name (evar-name nevar))
				    nevar nil)) ; no need to coerce case.
			     (push nevar *free-evars*)
			     nevar))
			  (:lowercase 
			   (let ((nuvar (new-uvar name ltype)))
			     ;; Establish the Uvar's name so it can be
			     ;; recognized later.
			     (setf (uvar-ident nuvar)
				   (ident:new-ident
				    (symbol-name (uvar-name nuvar))
				    nuvar nil))	; no need to coerce case.
			     (push nuvar *free-uvars*)
			     nuvar)))))
		  ;; If reading a module, it is either a free variable
		  ;; or an undeclared constant.
		  (ecase (case-of-name name)
		    (:uppercase (name-to-freevar name ltype))
		    (:lowercase 
		     ;; Just look up in the list, not as hash-table!!!
		     ;; No need to shadow ident TABLE!!!
		     (let ((polytycon (assoc name *undeclared-constants*)))
		       (if polytycon
			   ;; Undeclared constant used again.
			   (push ltype (cdr polytycon))
			   ;; New undeclared constant.
			   (push (cons name (list ltype))
				 *undeclared-constants*))
		       (Const (Named-lcobj name) ltype))))))))))

(defun name-to-freevar (name ltype)
  (declare (special *free-vars* *free-count*)
	   (special *bound-count*))
  (dlet ( ((values ffound-depth ffound-type)
	   (vcontext-lookup name *free-vars*)) )
    (Bvar (+ *bound-count*
	     (if ffound-depth
		 (progn (unify-ltype-or-bust name ffound-type ltype)
			;; Count from end.
			(- *free-count* ffound-depth 1))
		 ;; Otherwise, we need to add a new type
		 ;; variable.
		 (progn (push (cons name ltype) *free-vars*)
			(incf *free-count*)
			(1- *free-count*)))))))

(defun from-named-tcon (name kind)
  (declare (special *reading-module-p* *translation-modules*)
	   (special *undeclared-type-constants*)
	   (special *free-tcons*))
  (let ((given-kind (elp:modules-sig-get name *translation-modules*
					 #'elp:module-kindsig)))
    (if given-kind
	(progn (kind-equal-or-bust name given-kind kind)
	       (Tcon name given-kind))
	;; If not a known constant, it depends on whether we are reading
	;; a module or not.
	(if (not *reading-module-p*)
	    ;; if NOT reading a module, check if the constant is known
	    ;; from the top-level input (a Tcon in the ident table).
	    (let ((found-tcon
		   (ident:ident-retrieve-check (symbol-name name)
					       'Tcon)))
	      (if found-tcon
		  (dlet ( ((Tcon _ previous-kind) found-tcon) )
		    (kind-equal-or-bust found-tcon kind previous-kind)
		    found-tcon)
		  ;; If not found, enter it.
		  (let ((ntcon (Tcon name kind)))
		    (ident:new-ident (symbol-name name) ntcon nil)
		    (push ntcon *free-tcons*)
		    ntcon)))
	    ;; if reading a module, enter it as an undeclared type constant.
	    (let ((udcl (assoc name *undeclared-type-constants*)))
	      (if (not udcl)
		  (let ((ltype (Tcon name kind)))
		    (push (cons name ltype) *undeclared-type-constants*)
		    ltype)
		  (dlet ( ((:as previous-tcon (Tcon _ previous-kind))
			   (cdr udcl)) )
		    (kind-equal-or-bust name previous-kind kind)
		    previous-tcon)))))))

(defun from-named-tvar (name)
  ;; must be of kind Type.
  (declare (special *reading-module-p*)
	   (special *free-tevars*)
	   (special *free-tvars*))
  (if (not *reading-module-p*)
      ;; Not reading module: tevars become tevars.
      (or (ident:ident-retrieve-check (symbol-name name) 'Tevar)
	  ;; See if this name has already been seen.  If so, use it.
	  ;; If not, make a new tevar.
	  (let ((ntevar (new-tevar name)))
	    (setf (tevar-ident ntevar)
		  (ident:new-ident
		   (symbol-name (tevar-name ntevar))
		   ntevar nil))		; do not coerce case - no need.
	    (push ntevar *free-tevars*)
	    ntevar))
      ;; Reading module: tevars become tbvars later on.
      (or (cdr (assoc name *free-tvars* :test #'eq))
	  (let ((ntevar (new-tevar name)))
	    (push (cons name ntevar) *free-tvars*)
	    ntevar))))

(defun from-sltype-rec (sltype)
  "Like from-slterm-rec, but for types."
  (declare-ftype from-sltype-rec (slterm:Sltype) Ltype)
  (dcase sltype
    ((slterm:Tcon name)
     (from-named-tcon name *ktype*))
    ((slterm:Tvar name)
     ;; First see if it is a saved tevar
     (from-named-tvar name))
    ((slterm:Tfun sldomain slrange)
     (Tfun (from-sltype-rec sldomain) (from-sltype-rec slrange)))
    ((slterm:Tappl slrator slrand)
     (let ((lrator (from-sltype-constr-rec slrator (Kfun *ktype*)))
	   (lrand  (from-sltype-rec slrand)))
       (Tappl lrator lrand)))
    (_ (elp:lp-error "Illegal abstract syntax constructed upon parsing.
This should not happen unless you were trying to tickle this bug."))))

(defun from-sltype-constr-rec (sltype kind)
  "Build a type constructor (like `list') of kind KIND from SLTYPE."
  (dcase sltype
    ((slterm:Tcon name)
     (from-named-tcon name kind))
    ((slterm:Tvar _)
     (elp:lp-error "Type variables must be of kind ~S,
but ~S has kind ~S." *ktype* sltype kind))
    ((slterm:Tfun _ _)
     (elp:lp-error "Type ~S used as type constructor." sltype))
    ((slterm:Tappl slrator slrand)
     (Tappl (from-sltype-constr-rec slrator (Kfun kind))
	    (from-sltype-rec slrand)))
    (_ (elp:lp-error "Illegal abstract syntax constructed upon parsing.
This should not happen unless you were trying to tickle this bug."))))

(defun kind-equal (lk1 lk2)
  (dcase lk1
    (Ktype (dcase lk2
	     (Ktype t)
	     (_ nil)))
    ((Kfun range1)
     (dcase lk2
       ((Kfun range2) (kind-equal range1 range2))
       (_ nil)))))

(defun kind-equal-or-bust (thing has wants)
  (when (not (kind-equal has wants))
    (elp:lp-error "The kind of ~S was declared as ~S but used as ~S."
		  thing has wants)))

(defun from-slkind (slkind)
  "Convert SLKIND to an lkind."
  (declare-ftype from-slkind (slterm:Slkind) Lkind)
  (dcase slkind
    ((slterm:Ktype)
     *ktype*)
    ((slterm:Kfun slrange)
     (Kfun (from-slkind slrange)))))

(defun ltyfam-from-sltype (sltype modules)
  "Translate SLTYPE into an ltyfam using MODULES for signature lookups.  The
tbvars are gathered up and then explicitly `abstracted' over."
  (declare-ftype ltyfam-from-sltype
    (slterm::Sltype (List-of elp:Module)) Ltyfam)
  ;; First, convert using from-sltype-rec, which introduces tevars. Then,
  ;; proceed similarly to fix-tevars: Get the tevars, convert them to tbvars
  ;; and trim the result.  This could be more direct, and therefore, more
  ;; efficient, by not constructing the tevars in the first place.
  (let ((*translation-modules* modules)
	(*reading-module-p* t))
    (declare (special *translation-modules* *reading-module-p*))
    (let ((*free-tvars* ())
	  (*undeclared-type-constants* ()))
      (declare (special *free-tvars* *undeclared-type-constants*))
      (let ((ltype (from-sltype-rec sltype)))
	(dlet ( ((values tbvar-names qltype)
		 ;; This builds a redundant list of free tvars, but
		 ;; this should not be expensive.  Ignore the overhead.
		 (quant-tevars-in-ltype ltype)) )
	  (values (Ltyfam tbvar-names qltype)
		  (fix-undeclared-type-constants *undeclared-type-constants*)))))))


;;; Given an lterm, just translated from a slterm, go through and collect the
;;; remaining tevars, replacing each with an appropriate tbvar.  Also, remove
;;; tevar indirections.

(defun quant-tevars-in-ltermlist (ltermlist dformlist)
  (dcase ltermlist
    (() dformlist)
    ((cons (list freevars lterm) ltermlist-rest)
     (quant-tevars-in-ltermlist
      ltermlist-rest
      (cons (quant-tevars-in-lterm freevars lterm)
	    dformlist)))))

(defun quant-tevars-in-ltype (ltype)
  (let ((*free-tbvars* ())
	(*tbvar-count* 0))
    (declare (special *free-tbvars* *tbvar-count*))
    (let ((qltype (quant-tevars-in-ltype-rec ltype)))
      (values (tbvars-to-context *free-tbvars* ())
	      qltype))))

(defun quant-tevars-in-lterm (context lterm)
  "Assuming LTERM is Evar-free, abstract over the free type E-variables.
Destructive on CONTEXT and LTERM."
  (declare-ftype abst-over-tevars (Lterm) Lterm (List-of Symbol))
  (let ((*free-tbvars* ())
	(*tbvar-count* 0))
    (declare (special *free-tbvars* *tbvar-count*))
    (dlet ( ((values vnames vtypes)
	     (quant-tevars-in-context context () ())) )
      ;; Next line has side-effect on variable LTERM.
      (quant-tevars-in-lterm-rec lterm)
      (list (tbvars-to-context *free-tbvars* ())
	    vnames
	    vtypes
	    (lterm-to-dform lterm vtypes)))))

(defun tbvars-to-context (free-tbvars context)
  (dcase free-tbvars
    (() context)
    ((cons (cons (Tevar name _ _) _) free-tbvars-rest)
     (tbvars-to-context free-tbvars-rest (cons name context)))))

(defun quant-tevars-in-context (context vnames vtypes)
  (dcase context
    (() (values vnames vtypes))
    ((cons (cons vname vtype) context-rest)
     (quant-tevars-in-context context-rest
			      (cons vname vnames)
			      (cons (quant-tevars-in-ltype-rec vtype) vtypes)))))

(defun quant-tevars-in-lterm-rec (lterm)
  (dcase lterm
    ((Const _ ltype)
     (let ((qltype (quant-tevars-in-ltype-rec ltype)))
       (setf (const-type lterm) qltype))
     (values))
    ((Bvar _) (values))
    ;; Do the context separately.
    ((Appl rator rand)
     (quant-tevars-in-lterm-rec rator)
     (quant-tevars-in-lterm-rec rand))
    ((Abst _ ltype body)
     (let ((qltype (quant-tevars-in-ltype-rec ltype)))
       (setf (abst-type lterm) qltype))
     (quant-tevars-in-lterm-rec body))
    (_ (error "Unexpected object ~S in lterm." lterm))))

(defun quant-tevars-in-ltype-rec (ltype)
  (declare (special *free-tbvars* *tbvar-count*))
  (dcase ltype
    (Tcon ltype)
    ((Tevar _ _ value-slot)
     (if value-slot
	 (quant-tevars-in-ltype-rec value-slot)
	 (let ((is-known (assoc ltype *free-tbvars* :test #'eq)))
	   (if is-known
	       (cdr is-known)
	       (let ((free-tvar (Tbvar *tbvar-count*)))
		 (incf *tbvar-count*)
		 (push (cons ltype free-tvar) *free-tbvars*)
		 free-tvar)))))
    ((Tfun domain range)
     (re-Tfun ltype domain range
	      (quant-tevars-in-ltype-rec domain)
	      (quant-tevars-in-ltype-rec range)))
    ((Tappl rator rand)
     (re-Tappl ltype rator rand
	       (quant-tevars-in-ltype-rec rator)
	       (quant-tevars-in-ltype-rec rand)))
    (_ (error "Unexpected object ~S in ltype." ltype))))

;;; Note that lterm-trim-tevars does not need to return a value, since its
;;; argument's top-level structure cannot change.  However ltype-trim-tevars
;;; does need to return a value.

(defun lterm-trim-tevars (lterm)
  "Destructively eliminate the tevar chains in LTERM.
This operation is not necessary, but will make future operations on the term
be faster.  Note that this prevents uninstantiating the tevars."
  (declare-ftype lterm-trim-tevars (Lterm) )
  ;; This procedure is not smart enough to avoid examining the same
  ;; evar or uvar repeatedly.
  (dcase lterm
    ((Const _ type)
     ;; Note that the following mapcar makes a new list.  Hmmm ...
     (setf (const-type lterm) (ltype-trim-tevars type)))
    (Bvar)
    ((Evar _ type _ _ value-slot)
     (setf (evar-type lterm) (ltype-trim-tevars type))
     (when value-slot
       (lterm-trim-tevars value-slot)))
    ((Uvar _ type _ _)
     (setf (uvar-type lterm) (ltype-trim-tevars type)))
    ((Appl rator rand)
     (lterm-trim-tevars rator)
     (lterm-trim-tevars rand))
    ((Abst _ type body)
     (setf (abst-type lterm) (ltype-trim-tevars type))
     (lterm-trim-tevars body)))
  (values))


(defun ltype-trim-tevars (ltype)
  "Destructively eliminate thee tevar chains in LTYPE, returning the new ltype."
  (declare-ftype ltype-trim-tevars (Ltype) Ltype)
  (dcase ltype
    (Tcon
     ltype)
    (Tbvar
     ltype)
    ((Tevar _ _ value-slot)
     (if value-slot
       (ltype-trim-tevars value-slot)
       ltype))
    ((Tfun domain range)
     (setf (tfun-domain ltype) (ltype-trim-tevars domain))
     (setf (tfun-range ltype) (ltype-trim-tevars range))
     ltype)
    ((Tappl rator rand)
     (setf (tappl-rator ltype) (ltype-trim-tevars rator))
     (setf (tappl-rand ltype) (ltype-trim-tevars rand))
     ltype)))

;;;
;;; Next is type anti-unification.  This is used in order to determine
;;; the types of undeclared constants after a module is read.
;;;

(defun ltype-head (ltype)
  "Given an Ltype, find its head."
  (dcase ltype
    ((Tappl rator _)
     (ltype-head rator))
    (_ ltype)))

(defun ltype-equal (ltype-a ltype-b)
  "Checks whether two types are equal."
  (declare-ftype ltype-equal (Ltype Ltype) Boolean)
  ;; No evars allowed: don't chase indirections.
  ;; First chase type indirections.  Then we know that if either type is a
  ;; tevar, it is uninstantiated.
  (let ((ltype-a (chase-tevars ltype-a))
	(ltype-b (chase-tevars ltype-b)))
    (or (eq ltype-a ltype-b)
	(dcase ltype-a
	  ;; It might be that tcon's with the same name are guaranteed
	  ;; to be EQ.  If so, the following name eq test will always
	  ;; fail.
	  ((Tcon name-a _)
	   (dcase ltype-b
	     ((Tcon name-b _)
	      (eq name-a name-b))
	     (_ nil)))
	  ((Tbvar index-a)
	   (dcase ltype-b
	     ((Tbvar index-b)
	      (= index-a index-b))
	     (_ nil)))
	  (Tevar 
	   (dcase ltype-b
	     (Tevar (eq ltype-a ltype-b))
	     (_ nil)))
	  ((Tfun domain-a range-a)
	   (dcase ltype-b
	     ((Tfun domain-b range-b)
	      (and (ltype-equal domain-a domain-b)
		   (ltype-equal range-a range-b)))
	     (_ nil)))
	  ((Tappl rator-a rand-a)
	   (dcase ltype-b
	     ((Tappl rator-b rand-b)
	      (and (ltype-equal rator-a rator-b)
		   (ltype-equal rand-a rand-b)))
	     (_ nil)))))))

(defun ltypelist-equal (typelist1 typelist2)
  (if (and (null typelist1) (null typelist2))
      t
      (if (or (null typelist1) (null typelist2))
	  nil
	  (and (ltype-equal (first typelist1) (first typelist2))
	       (ltypelist-equal (rest typelist1) (rest typelist2))))))

(defun disagreement (ltypelist)
  "LTYPELIST is assumed to be a disagreement.  It returns a bound variable:
if the disagreement has occurred before, we choose the same variable.
Otherwise, make up a new one and record the disagreement."
  (declare (special *matches*))
  (let ((prev-match (assoc ltypelist *matches* :test #'ltypelist-equal)))
    (if prev-match
	(cdr prev-match)
	(let ((new-var (Tbvar (length *matches*))))
	  (push (cons ltypelist new-var) *matches*)
	  new-var))))

(defun ltypelist-anti-unify-rec (ltypelist)
  "Like LTYPELIST-ANTI-UNIFY, except that MATCHES carries the disagreements
encountered so far and associated fresh variables."
  ;;; Right now, return simply the "trivial" tyfam.
  ;;; (Ltyfam (list 'R) (Tbvar 0))
  (let ((fst (first ltypelist)))
    (dcase fst
      ( (Tcon name _)
	(if (every #'(lambda (lty) (and (is-Tcon lty)
					(eq name (Tcon-name lty))))
		   (rest ltypelist))
	    fst
	    (disagreement ltypelist)) )
      ( (Tbvar _)
	(error "There should be no free type variable in a list of
types to anti-unify.") )
      ( (Tevar _ _ _)
	(disagreement ltypelist) )
      ( (Tfun _ _)
	(if (every #'(lambda (lty) (is-Tfun lty)) (rest ltypelist))
	    (Tfun (ltypelist-anti-unify-rec
		   (mapcar #'Tfun-domain ltypelist))
		  (ltypelist-anti-unify-rec
		   (mapcar #'Tfun-range ltypelist)))
	    (disagreement ltypelist)) )
      ( (Tappl _ _)
	(if (every #'(lambda (lty) (is-Tappl lty)) (rest ltypelist))
	    (let* ((head-list (mapcar #'ltype-head ltypelist))
		   (fst-head-name (Tcon-name (first head-list))))
	      (if (every #'(lambda (head) (eq fst-head-name (Tcon-name head)))
			 head-list)
		  ;; This is extremely inefficient, since heads may
		  ;; be compared again, but who cares.
		  (Tappl (ltypelist-anti-unify-rec
			  (mapcar #'Tappl-rator ltypelist))
			 (ltypelist-anti-unify-rec
			  (mapcar #'Tappl-rand ltypelist)))
		  (disagreement ltypelist)))
	    (disagreement ltypelist)) )
      )))

(defun ltypelist-anti-unify (ltypelist)
  (declare-ftype ltype-anti-unify (List Ltype) Ltyfam)
  "Anti-unifies the types in LTYPELIST."
  (let ((*matches* '()))
    (declare (special *matches*))
    (let ((lty (ltypelist-anti-unify-rec
		(mapcar #'ltype-trim-tevars ltypelist))))
      (Ltyfam (make-sequence 'list (length *matches*)
			     :initial-element 'elp::|A|)
	      lty))))
