;;; -*- Mode: Lisp; Package: lterm -*-
;;;
;;; The case when two generalized variables are set equal to each other is
;;; handled specially in this file.  This code is called from lterm-unify.lisp.
;;; What is called a "gvar" here is also called a "virtual variable" or 
;;; "vvar" in other files.
;;;
;;; Author: Dale Miller.  Last Modified Mon Sep 11 06:09:13 1989
;;;
;;; Sccs Id "@(#)lterm-gvar-gvar.lisp	1.2 9/15/89
;;;
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;

(in-package :lterm) (use-package :ergolisp)

(defun do-gvar-gvar (terma heada termb headb)
  "The general gvar-gvar case is
     H x1 ... xn  =  G y1 ... ym
where H and G are evars and x1 ... xn are distinct uvars dominated
by H and y1 ... ym are distinct uvars dominated by G.  Such a dpair
can be completely and simply solved by a unique substitution.  The 
program do-gvar-gvar will compute this and bind G and/or H to the
appropriate term (two bindings may be necessary).
Case:  H is G.  Read comment for do-gvar-gvar-same.
Case:  H is different from G. Read comment for do-gvar-gvar-diff.
The list of substitution terms used in making substitutions is
returned.  This list contains either 0 (if the gvars were equal), 1,
or 2 terms."
  (declare-ftype do-gvar-gvar
    ((Virtual-Evar Evar Virtual-Evar Evar Virtual-Evar) (List-of Lterm)))
  (let ((uvarsa (lterm-rands terma))
	(uvarsb (lterm-rands termb)))
    (if (eq heada headb)
	(do-gvar-gvar-same heada uvarsa uvarsb)
	(do-gvar-gvar-diff heada uvarsa headb uvarsb))))


(ddefun do-gvar-gvar-same ((:as evar (Evar ename etype estamp _ _))
			   uvars1 uvars2)
  "Here, G and H are the same and, therefore, n and m are equal.  The
substitution term for H is then the term x1\...\xn\ (H' z1 ... zp)
where H' is a new evar (of appropriate type) and zj is the j'th variable
such that xi and yi are equal (common value being zj).  For example, the dpair
              H x y z = H u y x
has the ``most general solution'' H --> x\y\z\(H' y) (here, z1 would be y).
While this substitution is valid if the lists x1,...,xn and z1,...,zp are 
equal, it is better to simply drop this pair and generate no substitution."
  (declare-ftype do-gvar-gvar-same (Evar (List-of Uvar) (List-of Uvar))
		 (List-of Lterm))
  (if (equal uvars1 uvars2) ()
      (let ((length-of-binder (length uvars1)))
	(labels ((common-indexes-and-types (uvars1 uvars2 index)
                   (if (null uvars1) ()
		       (if (eq (car uvars1) (car uvars2))
			   (cons (cons index (uvar-type (car uvars1)))
				 (common-indexes-and-types (cdr uvars1)
							   (cdr uvars2)
							   (1+ index)))
			   (common-indexes-and-types (cdr uvars1) (cdr uvars2)
						     (1+ index)))))
		 (build-sub-term-body (indexes-and-types body-type)
	           (if (null indexes-and-types)
		       (Evar ename body-type estamp nil nil)
		       (Appl (build-sub-term-body
			      (cdr indexes-and-types)
			      (tfun (cdar indexes-and-types) body-type))
			     (Bvar (- length-of-binder
				      (caar indexes-and-types)))))))
	  (let ((subterm
		 (add-uvars-as-binder
		  (build-sub-term-body
		   (reverse (common-indexes-and-types uvars1 uvars2 1))
		   etype)
		  (reverse uvars1))))
	    (setf (evar-value-slot evar) subterm)
	    (push evar elp:*trail*)
	    (list subterm))))))

(defun do-gvar-gvar-diff (H-evar xs G-evar ys)
  "Here, H is different from G.  Assume that G's stamp dominates or equals
H's stamp. Then the following substitution almost works:
             H --> x1\...\xn\ (G y1 ... ym)
The only problem with it is that some of the ys might be free in this term
and are dominated by H, so they cannot be substituted into H.
Such argument positions of G must be pruned away.  For example, in 
          pi X\ (pi Y\ (pi Z\ (A X Y = B Y Z)))
A --> X\Y\ (B Y Z) will not work since Z is free in it.  Thus, we
need to substitute B --> U\V\ (B' U) first (this is called pruning)
to get the problem
          pi X\ (pi Y\ (pi Z\ (A X Y = B' Y)))
and then substitute A --> X\Y\ (B' Y)."
  (declare-ftype do-gvar-gvar-diff (Evar (List-of Uvar) Evars (List-of Uvar))
		 (List Lterm))
  (cond ((stamp-dominates (evar-stamp H-evar) (evar-stamp G-evar))
	 (psetq H-evar G-evar G-evar H-evar
		xs     ys     ys     xs)))
  (dlet (((evar ename etype estamp _ _) G-evar))
    (let* ((zs (remove-if #'(lambda (uvar)
			      (and (not (member uvar xs :test #'eq))
				   (stamp-dominates (evar-stamp H-evar)
						    (uvar-stamp uvar))))
			  ys)))
      (cond ((equal ys zs) ;; no pruning, so don't substitute into G-evar
	     (let ((H-subterm (abstract-out xs (add-appls G-evar ys))))
	       (setf (evar-value-slot H-evar) H-subterm)
	       (push H-evar elp:*trail*)
	       (list H-subterm)))
	    (t (let* ((new-etype (add-domains
				  (mapcar #'uvar-type zs)
				  (dlet (((values _ target)
					  (strip-domains etype))) target)))
		      (new-evar (Evar ename new-etype estamp nil nil))
		      (body (add-appls new-evar zs))
		      (G-subterm (abstract-out ys body))
		      (H-subterm (abstract-out xs body)))
		 (setf (evar-value-slot G-evar) G-subterm)
		 (push G-evar elp:*trail*)
		 (setf (evar-value-slot H-evar) H-subterm)
		 (push H-evar elp:*trail*)
		 (list G-subterm H-subterm)))))))


(defun add-uvars-as-binder (lterm uvars)
  "This function just adds the uvars in UVARS to the front of lterm.
Generally, UVARS is the reverse of the intended order, so you will probably
need to reverse this list first."
  (declare-ftype add-uvars-as-binder (Lterm (List-of Uvar)) Lterm)
  (dcase uvars
    (() lterm)
    ((cons (Uvar name ltype _ _) rest-args)
     (add-uvars-as-binder (Abst name ltype lterm) rest-args))))

(defun abstract-out (uvars lterm)
  "Abstract out the Uvars in UVARS from LTERM, replacing them with 
name-less dummies instead.  Then add the UVARS as abstractions to the
top of the resulting term."
  (declare-ftype abstract-out ((List-of Uvar) Lterm) Lterm)
  (let ((uvars (reverse uvars)))
    (labels ((abs-out (lterm depth)
		(let ((lterm (chase-evars lterm)))
		  (dcase lterm
		    ((Bvar _) lterm)
		    ((Const _ _) lterm)
		    ((Evar _ _ _ _ nil) lterm)
		    ((Uvar _ _ _ _)
		     (let ((pos (position lterm uvars :test #'eq)))
		       (if pos (Bvar (+ depth pos)) lterm)))
		    ((Abst name type body)
		     (re-Abst-body lterm name type body
				   (abs-out body (1+ depth))))
		    ((Appl rator rand)
		     (re-Appl lterm rator rand
			      (abs-out rator depth)
			      (abs-out rand  depth)))))))
      (add-uvars-as-binder (abs-out lterm 0) uvars))))



; Some examples with the same head
; ?-  pi X\ (pi Y\ (pi Z\ (H X Y = H Z Y))).
; H == X1 \ Y1 \ (H1 Y1)
;
; ?- pi U\ (pi V\ ( pi X\ (pi Y\ (pi Z\ (H X Y Z V = H Z Y U V))))).
; H == X1 \ Y1 \ Z1 \ V1 \ (H1 Y1 V1)
;
; ?- pi U\ (pi V\ ( pi X\ (pi Y\ (pi Z\ (H X Y Z V = H Z V U Y))))).
; H == X1 \ Y1 \ Z1 \ V1 \ H1
;; Some examples with different heads
; ?- pi X : a\ (pi Y : b \ (pi Z : c \ (H Y Z = G X Y))).
; H == Y \ Z \ (G1 Y)
; G == X \ Y \ (G1 Y)
; ?- pi X : a\ (sigma H\ (pi Y : b \ (pi Z : c \ (H Y Z = G X Y, write H)))).
; Solving . . .Y \ Z \ (G X Y)
; G == G
; ?- pi X : a\ (sigma H\ (pi Y : b \ (pi Z : c \ (H Y Z = G X Y Z, write H)))).
; Y \ Z \ (G X Y Z)
; G == G
; ?- pi X : a\ (pi Y : b \ (pi Z : c \ (g V (H Y Z) = g W (G X Y)))).
; Solving . . .
; V == W
; H == Y \ Z \ (G1 Y)
; W == W
; G == X \ Y \ (G1 Y)
; ?- pi X : a\ (sigma H: d\ (pi Y : b \ (H = G Y X, write H))).
; Solving . . .G1 X
; G == Y \ X1 \ (G1 X1)
; ?- pi X : a\ (sigma H: d\ (pi Y : b \ (H = G X Y, write H))).
; Solving . . .G1 X
; G == X1 \ Y \ (G1 X1)
; ?- pi X : a\ (sigma H: d\ (pi Y : b \ (H = G X Y, write H))).
; Solving . . .G1 X
; G == X1 \ Y \ (G1 X1)
