;;; -*- Mode: Lisp; Package: lterm -*-
;;;
;;; This file does some simple ``pruning'' of flexible-flexible
;;; disagreement pairs
;;;
;;; Author: Dale Miller.  Last Modified Fri Sep 15 10:57:51 1989
;;;
;;; Sccs Id "@(#)lterm-prune.lisp	1.2 9/15/89
;;;
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;

(in-package :lterm) (use-package :ergolisp)

(export '(prune-constraints))
; This symbol is needed in the ELP package because it is called by a
; special lambda Prolog goal prunecon.

(defun possible-occurrence-in (Uvar Lterm)
  "Check if UVAR has a possible occurrence in LTERM.  Beta-normalization
is not done here.  If it were, more pruning might be achieved.  If this
function returns NIL the no legal substitution of LTERM can contain UVAR."
  (declare-ftype occurs-in (Uvar Lterm) Boolean)
  (dcase (chase-evars Lterm)
    ((Const _ _) nil)
    ((Bvar _) nil)
    ((Evar _ _ estamp _ _) (not (stamp-dominates estamp (uvar-stamp uvar))))
    ((:as uv (Uvar _ _ _ _)) (eq uv Uvar))
    ((Appl rator rand) (or (possible-occurrence-in Uvar rator)
			   (possible-occurrence-in Uvar rand)))
    ((Abst _ _ body) (possible-occurrence-in Uvar body))))

(defun permanent-uvars (lterm acc)
  "LTERM must be in beta-normal form.  The result is a list of uvars
having permanent (rigid) occurrences in LTERM.  This list will not have
duplicates.  ACC accumlates uvars on the descent - initialize it to ()."
  (declare-ftype permanent-uvars (Lterm (list of Uvar)) (list-of Uvar))
  (dcase (chase-evars lterm)
    (Const acc)
    (Bvar acc)
    (Evar acc)
    ((:as uv (Uvar _ _ _ _)) (if (member uv acc :test #'eq) acc
				 (cons uv acc)))
    ((Abst _ _ body) (permanent-uvars body acc))
    ((:as term (Appl _ _))
     (dlet (((values head args) (strip-appls term)))
       (let ((rigid?
	      (dcase (chase-evars head)
		(Const t)
		(Bvar t)
		(Evar nil)
		((:as uv (Uvar _ _ _ _)) uv)
		(_ (error "Term is not normal in permanent-uvars.")))))
	 (if rigid?
	     (let ((newacc
		    (reduce #'permanent-uvars args :from-end t
			    :initial-value acc)))
	       (if (eq t rigid?) newacc
		   (if (member rigid? newacc :test #'eq) newacc
		       (cons rigid? newacc))))
	     acc))))))


(defun can-prune-p (arg lterm)
  "Determine whether or not the argument ARG of a flexible term can be
pruned given the flexible term is to me unified with LTERM."
  (declare-ftype can-prune-p (Lterm Lterm) Boolean)
  (some #'(lambda (uvar) (not (possible-occurrence-in uvar lterm)))
	(permanent-uvars (beta-norm arg) ())))

(defun do-pruning (flex-term lterm indexes type counter really?)
  "Compute and bind the pruning substitution for the given flexible term.
INDEXES should be initialized to (), COUNTER should be initialized to 0,
and REALLY? should be initialized to NIL.  The value returned is the
pruned substitution, if there is one; otherwise NIL is returned.  Notice
that the pruning terms built here are only complete (preserve all solutions)
if the type of the evar head of the flex-term is closed and second order.
A counter-example was pointed out to me by Masami Hagiya:
         ?- pi Y: i\ (X (Z: i\ U) Y = U: i)
has a solution X == V1\V2\ (V1 V2).  Pruning here would miss this solution.
Note: pruning is achieved here by examining the possible/permanent occurrences
of uvars.  Other motivations for pruning are possible.  For example, type
information could be used: is it possible for a variable of type,
say a -> b, to be a non-vacuous abstraction?  This is not attempted here
since the current signature is not so easily accessible in eLP."
  (declare-ftype do-pruning
    (Lterm Lterm (List-of Integer) Ltype Integer Boolean)
    (or Lterm Null))
  (dcase Flex-term
    ((Appl rator rand)
     (if (can-prune-p rand lterm)
	 (do-pruning rator lterm indexes
		     type (1+ counter) t)
	 (do-pruning rator lterm (cons counter indexes)
		     (tfun (ltype-of rand nil) type) (1+ counter) really?)))
    ((:as evar (Evar _ etype _ _ _))
     (cond (really? 
	    (let ((subterm
		   (add-abs counter etype
		    (add-appl (Evar (evar-name evar) type (evar-stamp evar)
				    nil nil)
			      indexes))))
	      (setf (evar-value-slot evar) subterm)
	      subterm))
	   (t nil)))
    (_ (error "Bad flexible term in do-pruning"))))


(defun add-appl (lterm indexes)
  "Add Bvars as applications to LTERM as indicated by the integers in
INDEXES."
  (declare-ftype add-appl (Lterm (List-of Integer)) Lterm)
  (if (null indexes)
      lterm
      (add-appl (Appl lterm (Bvar (car indexes))) (cdr indexes))
      ;; The previous version seemed to reverse things: did not match type.
      ;; (Appl (add-appl lterm (cdr indexes)) (Bvar (car indexes)))
      ))

(defun add-abs (counter type lterm)
  "Add counter number of abstractions to LTERM.  The types of the
abstractions are gotten from the domain types of TYPE."
  (declare-ftype add-abs (Integer Ltype Lterm) Lterm)
  (if (= 0 counter) lterm
      (Abst 'elp::|x| (tfun-domain type)
	    (add-abs (1- counter) (tfun-range type) lterm))))


(defun cycle-thru (in check-left check-right check-gvars out did-sub)
  (declare-ftype cycle-thru
    (Dset (or Dpair Null) (or Dpair Null) (or Dpair Null) Dset Boolean) Dset)
  "The initial call to this tail recursive function should be
      (cycle-thru CONSTRAINTS NIL NIL NIL () NIL).
The arguments are:
  IN          input flexible-flexible unification problem
  OUT         processed pairs are placed here
  DID-SUB     this boolean indicates that during the processing of the
              flexible-flexible pairs, a substitution was or wasn't made
              by either of the pruning phase or the gvar phase
At most one of the following CHECK- arguments will be non-nil.  If is not 
NIL then it is a disagreement pair.
  CHECK-LEFT  names a disagreement pair for which pruning of the left
              term will be attempted
  CHECK-RIGHT names a disagreement pair for which pruning of the right
              term will be attempted
  CHECK-GVAR  names a disagreement pair for which gvars will be checked"
  (if check-left
      (dlet (((Dpair left right lhead rhead) check-left))
	(let ((dpair-type (ltype-of left nil)))
	  (if (and (second-order-closed-type-p (ltype-of lhead nil))
		   (do-pruning left right nil dpair-type 0 nil))
	      (progn 
		(multiple-value-setq (left lhead) (re-normalize left))
		(multiple-value-setq (right rhead) (re-normalize right))
		(cycle-thru in nil (Dpair left right lhead rhead) nil out t))
	      (cycle-thru in nil check-left nil out did-sub))))
      (if check-right
	  (dlet (((Dpair left right lhead rhead) check-right))
	    (let ((dpair-type (ltype-of left nil)))
	      (if (and (second-order-closed-type-p (ltype-of rhead nil))
		       (do-pruning right left nil dpair-type 0 nil))
		  (progn 
		    (multiple-value-setq (left lhead) (re-normalize left))
		    (multiple-value-setq (right rhead) (re-normalize right))
		    (cycle-thru in (Dpair left right lhead rhead)
				nil nil out t))
		  (cycle-thru in nil nil check-right out did-sub))))
	  (if check-gvars
	      (dlet (((:as dp (Dpair left right lhead rhead)) check-gvars))
		(if (is-virtual-Evar left lhead)
		    (if (check-sub-and-bind left right)
			(cycle-thru in nil nil nil out t)
			(if (is-virtual-Evar right rhead)
			    (let ((subterms
				   (do-gvar-gvar left lhead right rhead)))
			      (cycle-thru in nil nil nil out
					  (or subterms did-sub)))
			    (cycle-thru in nil nil nil (cons dp out) did-sub)))
		    (if (is-virtual-Evar right rhead)
			(if (check-sub-and-bind right left)
			    (cycle-thru in nil nil nil out t)
			    (cycle-thru in nil nil nil
					(cons (Dpair left right lhead rhead)
					      out)
					did-sub))
			(cycle-thru in nil nil nil (cons dp out) did-sub))))
	      (dcase in
		((cons (Dpair left right lhead rhead) rest)
		 (unless (headp lhead)
		   (multiple-value-setq (left lhead)
		     (re-normalize left)))
		 (unless (headp rhead)
		   (multiple-value-setq (right rhead)
		     (re-normalize right)))
		 (cycle-thru rest (Dpair left right lhead rhead)
			     nil nil out did-sub))
		(()
		 (if did-sub
		     (cycle-thru out nil nil nil () nil)
		     out)))))))


(defun first-order-closed-type-p (etype)
  "Check if argument is a primtive and closed type expression."
  (declare-ftype first-order-closed-type-p (Ltype) Boolean)
  (dcase etype
    (Tcon t)
    (Tbvar nil)
    (Tevar nil)
    (Tfun nil)
    (Tappl T)))

(defun second-order-closed-type-p (etype)
  "Check if argument is a second-order and closed type expression."
  (declare-ftype second-order-closed-type-p (Ltype) Boolean)
  (dcase etype
    (Tcon t)
    (Tbvar nil)
    (Tevar nil)
    ((Tfun domain range) (and (first-order-closed-type-p domain)
			      (second-order-closed-type-p range)))
    (Tappl T)))
    

(defun check-sub-and-bind (gvar lterm)
  "See if it is ok to simply substitute LTERM for GVAR.   If it is, do 
the substitution.  Return either NIL, meaning that no substitution was done,
or the substitution term used."
  (declare-ftype check-sub-and-bind (Virtual-Evar Lterm) (or Null Lterm))
  (dlet (((values (:as evar (Evar _ _ stamp-a _ nil)) uvars)
	  (strip-appls gvar)))
    (labels ((ok (lterm)
		 (setq lterm (chase-evars lterm))
		 (dcase lterm
		   (Const t)
		   (Bvar t)
		   ((:as evar-b (Evar _ _ stamp-b _ nil))
		    (not (or (eq evar evar-b)
			     (stamp-dominates stamp-a stamp-b))))
		   ((Uvar _ _ stamp-b _)
		    (or (member lterm uvars :test #'EQ)
			(not (stamp-dominates stamp-a stamp-b))))
		   ((Appl rator rand)
		    (and (ok rator) (ok rand)))
		   ((Abst _ _ body) (ok body)))))
      (if (ok lterm) (setf (evar-value-slot evar) (abstract-out uvars lterm))
	  nil))))

(defvar *do-prune-p* t)

(defun prune-constraints (con)
  "Given a flexible-flexible unification problem, do as much pruning as
you can to simplify the problem.  The simplified problem will be the
result.
 The general problem dealt with by the functions in this file is the following:
Consider a rigid-flexible or flexible-flexible disagreement pair of the form
                           t = G s1 ... sn  (G is an Evar)
Let y be a uvar such that 
 (1) y has a permanent (rigid) occurrence in si, and
 (2) y has no possible occurrence in t (that is, does no occur in that
    term and all evars in it dominate y).
If these cases hold, all substitution instances of t will not contain
the uvar y.  Hence, G cannot use it's ith argument, that is,
             G --> w1\...\wn\ (G' w1 ... w{i-1} w{i+1} ... wn)
is an (improper) factor of any unification problem containing this
disagreement pair.  Notice that although this works for t being either
rigid or flexible, the most useful and general time to use this is on
flexible-flexible disagreement pairs since there little else can be 
done to simplify their presentation.  (Type of G must be second-order.)"
  (declare-ftype prune-con (Constraint) Constraint)
  (if (not *do-prune-p*) con
      (cycle-thru con nil nil nil () nil)))


;; A module for testing pruning.
; module prune_test.
; % These declarations are used to provide for closed types.  Pruning
; % cannot work on flexible terms whose heads have polymorphic types
; % since they must be know to be second-order.  Actually, type variables
; % are permitted inside "atomic" types, just not at "prediate" positions.
; kind i type.
; type f   i -> i -> i.
; type g   i -> i.
; 
; type    testa    (i -> i -> i) -> (i -> i) -> o.
; type    testb    (i -> i -> i) -> (i -> i) -> o.
; type    testc    (i -> i) -> (i -> i) -> o.
; type    testd    (i -> i -> i) -> (i -> i -> i) -> o.
; type    teste    (i -> i -> i -> i) -> (i -> i -> i) -> o.
; type    testf    (i -> i -> i -> i) -> (i -> i -> i) -> o.
; type    testg    (i -> i -> i -> i) -> (i -> i -> i) -> o.
; type    testh    (i -> i) -> (i -> i) -> (i -> i -> i -> i) -> o.
; type    testi    (i -> i) -> (i -> i -> i) -> o.
; type    testj    (i -> i) -> (i -> i) -> (i -> i -> i -> i) -> (i -> i) -> o.
; type    testk    (i -> i -> i) -> (i -> i -> i) -> o.
; type    testl    (i -> i -> i) -> (i -> i -> i) -> o.
; type    testm    (i -> i -> i) -> (i -> i) -> (i -> i -> i -> i -> i) -> o.
; 
; testa A B :- pi X\ (pi Y\ (A X (f X Y) = B (g Y))).
; testb A B :- pi X\ (pi Y\ (A X (f X Y) = B (g X))).
; testc A B :- pi X\ (pi Y\ (A (f X Y) = B (g X))).
; testd A B :- pi X\ (pi Y\ (pi Z:i\ (A (f X Y) Z = B Z (g X)))).
; teste A B :- pi X\ (pi Y\ (pi Z\ (A (f X Y) (f Z Y) (f X Z) = B Y Y))).
; testf A B :- pi X\ (pi Y\ (pi Z\ (A (f X Y) (f Z Y) (f X Z) = B Y X))).
; testg A B :- pi X\ (pi Y\ (pi Z\ (A (f X X) (f Z Y) (f Y Z) = B Y X))).
; testh A B C :- pi X\ (pi Y\ (A (B X) = C X Y X)).
; testi A C   :- pi X\ (sigma B\ (pi Y\ (A (B X) = C X Y))).
; testj A B C D :- pi X\ (pi Y\ (A (B X) = C X (D Y) X)).
; testk A B  :- pi X\ (pi Y\ (A X Y = B X Y)).
; testl A B  :- pi X\ (pi Y\ (A X Y = B Y X)).
; testm A B C :- pi X\ (pi Y\ (pi Z\ (A (B X) Z = C X Y Z Z))).
 
; % ?- testa A B, prunecon.
; % Solving . . .
; % A == Pr \ Pr1 \ B1
; % B == Pr \ B1
; % no more solutions
; % ?- testb A B, prunecon.
; % Solving . . .
; % A == Pr \ Pr1 \ (B (g Pr))
; % B == B
; % no more solutions
; % ?- testc A B, prunecon.
; % Solving . . .
; % A == Pr \ B1
; % B == Pr \ B1
; % no more solutions
; % ?- testd A B, prunecon.
; % Solving . . .
; % A == Pr \ Pr1 \ (B1 Pr1)
; % B == Pr \ Pr1 \ (B1 Pr)
; % no more solutions
; % ?- teste A B, prunecon.
; % Solving . . .
; % A == Pr \ Pr1 \ Pr2 \ B1
; % B == Pr \ Pr1 \ B1
; % no more solutions
; % ?- testf A B, prunecon.
; % Solving . . .
; % A == Pr \ Pr1 \ Pr2 \ (A1 Pr)
; % B == Y \ X \ (A1 (f X Y))
; % no more solutions
; % ?- testg A B, prunecon.
; % Solving . . .
; % A == Pr \ Pr1 \ Pr2 \ (A1 Pr)
; % B == Pr \ Pr1 \ (A1 (f Pr1 Pr1))
; % no more solutions
; % ?- testh A B C, prunecon.
; % Solving . . .
; % A == A
; % B == B
; % C == Pr \ Pr1 \ Pr2 \ (C1 Pr2 Pr)
; % Remaining constraint: [
; %     <A (B X) == C1 X X> ]
; % no more solutions
; % ?- testi A C  , prunecon.
; % Solving . . .
; % A == A
; % C == Pr \ Pr1 \ (C1 Pr)
; % Remaining constraint: [
; %     <A (B X) == C1 X> ]
; % no more solutions
; % ?- testj A B C D, prunecon.
; % Solving . . .
; % A == A
; % B == B
; % C == C
; % D == D
; % Remaining constraint: [
; %     <A (B X) == C X (D Y) X> ]
; % no more solutions
; % ?- testk A B , prunecon.
; % Solving . . .
; % A == X \ Y \ (B X Y)
; % B == B
; % no more solutions
; % ?- testl A B , prunecon.
; % Solving . . .
; % A == X \ Y \ (B Y X)
; % B == B
; % no more solutions
; % ?- testm A B C, prunecon.
; % Solving . . .
; % A == A
; % B == B
; % C == Pr \ Pr1 \ Pr2 \ Pr3 \ (C1 Pr3 Pr2 Pr)
; % Remaining constraint: [
; %     <A (B X) Z == C1 Z Z X> ]
; % no more solutions
; % ?- 
