;;; -*- Mode: Lisp; Package: LTERM -*-
;;;
;;; Provide term language dependency info for LTERM
;;;
;;; Author: Conal Elliott.  Last Modified Sat Sep 16 20:47:46 1989
;;;
;;; Sccs Id @(#)lterm-tlang.lisp	1.2 9/16/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :LTERM) (use-package :ergolisp)

(export '(*tlang*))

;;; This progn and setq should be removed later.  It is convenient to have them
;;; here while testing.

(progn

  (defparameter *tlang*
    ;; Using function names rather than function bindings lets us change the
    ;; function definitions without re-evaluating this form.  Also, compilation
    ;; order doesn't matter.
    (elp:make-tlang
      :Kind 'Lkind
      :Typ 'Ltype
      :Term 'Lterm
      :Predicate '(or Lcobj Uvar)
      :Tevar 'Tevar
      :Evar 'Evar
      :Ctyp 'Lctype
      :Cterm 'Lcterm
      :Constraint 'Constraint
      :convert-skind 'from-slkind
      :convert-stype 'ltyfam-from-sltype
      :convert-sdformlist 'convert-sdformlist
      :convert-squery 'convert-squery
      :typ-of 'ltype-of
      :new-tevar 'new-tevar
      :new-evar 'new-evar
      :new-uvar 'new-uvar
      :mk-ctyp 'mk-lctype
      :mk-cterm 'Lcterm
      :copy-term 'copy-lterm
      :copy-typ 'copy-ltype
      :unbind-var 'unbind-var
      :term-to-gform 'lterm-to-gform
      :term-to-dform 'lterm-to-dform
      :gform-to-sterm 'gform-to-slterm
      :dform-to-sterm 'dform-to-slterm
      :unify 'unify
      :clause-head-unify 'clause-head-unify
      :empty-constraint *empty-constraint*
      :read-term 'read-lterm
      :present-solution 'present-solution
      :term-value 'lterm-value
      :term-string-value 'lterm-string-value
      :to-sterm 'to-slterm
      :parse-sterm 'parse-slterm
      :parse-smodule 'parse-smodule
      :unparse-sterm 'unparse-slterm
      :load-sterm-lang 'load-slterm-lang
      :absts-to-gform 'lterm-absts-to-gform
      :term-to-exp 'lterm-to-exp
      :exp-to-sterm 'exp-to-slterm
      :value-term 'value-lterm
      :head-eta-reduce 'head-eta-reduce
      :term-to-pred 'lterm-to-pred
      :typ-anti-unify 'ltype-anti-unify
      :term-nth-arg-head-value 'lterm-nth-arg-head-value
      ))

  (setq elp:*current-tlang* *tlang*))

(defun convert-sdformlist (sltermlist modules)
  "Convert SLTERMLIST to a Dform, using the same environment for constants.
Returns a list of Dforms in their contexts."
  (from-sltermlist sltermlist modules))

(defun convert-squery (slterm)
  "Convert SLTERM to (1) a Gform, (2) a sequence of tevars, and (3) a sequence
of evars.  Uses module main for signatures."
  (declare-ftype convert-squery
    (slterm:Slterm) elp:Gform (List-of Tevar) (List-of Evar))
  (dlet* ( ((values lterm evars uvars tevars)
	    (from-slterm slterm *prop-type*
			 (elp:currently-accessible-modules)) ))
    (values (lterm-to-gform lterm ())
	    (reverse tevars)
	    (reverse evars)
	    (reverse uvars))))

(defun read-lterm (ltype)
  "Read a term of the given LTYPE, doing type inference if appropriate.  Make
bvars and tbvars into evars and tevars.  WARNING: trails for type unification,
but does not do a (elp:trailing ..).  Calling code must take care of this."
  (declare-ftype read-lterm (Ltype) Term)
  (let ((slterm (parse-slterm elp:*elp-standard-input*)))
    (let ((ntevar (new-tevar 'elp::|A|)))
      (dlet ( ((values lterm _ _ _)	; ignoring   evars uvars tevars
	       (from-slterm
		slterm
		ntevar
		(elp:currently-accessible-modules))) )
	;; unify types with trailing on.
	(unless (unify-ltype (ltype-of lterm ()) ltype t)
	  (elp:lp-error "Expecting to read term of type ~S, but ~S has type ~S."
			ltype slterm ntevar))
	lterm))))

(defun lterm-value (lterm)
  "Make a `value' out of LTERM, which had better be a constant, uvar, or
application."
  (declare-ftype lterm-value (Lterm) T)
  (let ((lterm (chase-evars (beta-norm lterm))))
    (dcase lterm
      ((Const cobj _)
       (lcobj-value cobj))
      ;; Uvar remains a Uvar!
      (Uvar lterm)
      ;; Convert an application into a Lisp style list (uncurrying)
      ((Appl rator rand)
       (let ((rator-value (lterm-value rator))
	     (rand-value (lterm-value rand)))
	 (cons rator-value rand-value)))
      (_ (elp:lp-error "Cannot make a value out of the term ~S." lterm)))))

(defun lterm-string-value (lterm)
  (let ((lterm (chase-evars (beta-norm lterm))))
    (dcase lterm
      ((Const cobj _)
       (let ((v (lcobj-value cobj)))
	 (typecase v
	   (symbol (symbol-name v))
	   (string v)
	   (t (error "Expecting string but received other type of constant.
This violates the type system.")))))
      ((Uvar name _ _ ident)
       (if (stringp ident)
	   ident
	   (let ((ident (ident:new-ident (symbol-name name) lterm :lowercase)))
	     (setf (uvar-ident lterm) ident)
	     ident)))
      (_ (elp:lp-error "Cannot make a string out of the term ~S." lterm)))))

(defun lterm-to-pred (lterm)
  "Make a `predicate' out of LTERM, which had better be a constant or uvar."
  (declare-ftype lterm-to-pred (Lterm) Predicate)
  (let ((lterm (chase-evars (beta-norm lterm))))
    (dcase lterm
      ((Const (:as lcobj (named-lcobj name)) ltype)
       (when (elp:special-lookup name)
	   (elp:lp-error "Special predicates are not legal as predicates.
 Here: ~A." lterm))
       (when (member name *logical-primitives*)
	 (elp:lp-error "Logical primitives are not legal as predicates.
 Here: ~A." lterm))
       (dlet (( (values _ range) (strip-domains ltype) ))
	 (when (not (or (is-tevar range)
			(ltype-equal *prop-type* range)))
	   (elp:lp-error "The range type of the constant ~S is ~S and not ~S.
  It can therefore not serve as a predicate."
			 lterm range *prop-type*)))
       lcobj)
      (Uvar lterm)
      (_ (elp:lp-error "Cannot make a predicate out of the term ~S" lterm)))))

(defun lterm-head-value (lterm)
  "Make a `value' out of the head of LTERM, or return T.
This should work since an lcobj should be lower-case."
  (declare-ftype lterm-value (Lterm) T)
  (dlet (( (values _ _ head _) (to-hnf lterm)) )	; will be chased.
    ;; lterm-head would be fine, but it doesn't head-normalize
    ;; and therefore indexing will be helpful less often.
    ;; Unfortunately the work invested in to-hnf is wasted.
    (dcase head
      ((Const cobj _)
       (lcobj-value cobj))
      ;; Uvar remains a Uvar!
      (Uvar head)
      ;; Everything else is T.  Should be an unbound evar or bdvar.
      (_ nil))))

#|

(defun lterm-nth-arg-head-value (index lterm)
  "For now, skip indexing."
  nil)

|#

(defun lterm-nth-arg-head-value (index lterm)
  "Get the head value of the INDEXth argument to LTERM, which had better be in 
an atom applied to zero or more arguments."
  (lterm-head-value (nth index (lterm-rands lterm))))
