;;; -*- Mode: Lisp; Package: LTERM -*-
;;;
;;; Translate from lterms to slterms.
;;;
;;; Author: Conal Elliott.  Last Modified Fri Sep 29 23:33:46 1989
;;;
;;; Sccs Id @(#)lterm-to-slterm.lisp	1.26 9/29/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :LTERM) (use-package :ergolisp)

(export '(*elp-print-depth* *to-slterm-with-types-p* *to-slterm-normalize-p*))
(export '(to-slterm to-slkind))

(defvar *elp-print-depth* -1
  "Controls when subterms are chopped and printed as # (see standard grammar)
A negative number means: do not ellide.")

(defvar *to-slterm-with-types-p* t
  "Controls whether variable types or type parameters are inserted when a lterm
abstraction or constant is translated into an slterm.  Perhaps this should
allow the option of keeping evar and uvar types, using the slterm:Typed
constructor.")

(defvar *to-slterm-normalize-p* t
  "Controls whether lterms are beta-normalized before converting to syntactic
form.")


;;; Probably this should use dynamic variables for vnames and tvnames,
;;; like from-slterm does (especially considering that tvnames never changes).


(defun to-slterm (lterm ltypes tvnames vnames depth)
  "Translate LTERM, given the lists TVNAMES and VNAMES, into a
slterm.  The policy for choosing names of bvars will be to keep as close
as possible to the original name without introducing name shadowing."
  (declare-ftype to-slterm
    (Lterm (List-of Ltype) (List-of Symbol) (List-of Symbol)) slterm:Slterm)
  ;; Normalize if appropriate.
  ;; A lot of this might be unnecessary with ellision!
  (let* ((lterm (if *to-slterm-normalize-p*
		    (beta-norm lterm)
		    lterm)))
    (to-slterm-rec lterm ltypes tvnames vnames depth)))

(defun to-slterm-rec (lterm ltypes tvnames vnames depth)
  "Work-horse for to-slterm."
  (declare-ftype to-slterm-rec
    (Lterm (List-of Ltype) (List-of Symbol) (List-of Symbol)) slterm:Slterm)
  (if (= depth *elp-print-depth*)
      (slterm:EllidedSlterm)
      ;; First skip over evar indirections.
      (let* ((lterm (chase-evars lterm))
	     ;; (ltype (catch :loose-bvar (ltype-of lterm ltypes)))
	     ;; (op (if ltype (target-main-op ltype) nil))
	     ;; (translate-fun (if op (lterm-translator op) nil))
	     )
	;;(if translate-fun
	;; (funcall translate-fun lterm ltypes tvnames vnames (1+ depth))
	(flet ((opt-typed (slterm ltype)
			  ;; Return just SLTERM, or a Typed version of it using LTYPE,
			  ;; depending on the value of *to-slterm-with-types-p*.
			  (if *to-slterm-with-types-p*
			      (slterm:Typed slterm (to-sltype ltype tvnames depth))
			      slterm)))
	  (dcase lterm
	    ((Const lcobj _)
 	     (dcase lcobj
 	       ((String-lcobj string)
 		(slterm:Con (slterm:Cstring string)))
 	       ((Number-lcobj number)
 		(slterm:Con (slterm:Cnumber number)))
 	       ((Named-lcobj name)
 		(slterm:var name))))
	    ((Bvar index)
	     (let ((name (nth index vnames)))
	       (slterm:Var
		(or name
		    (depth-name "v" index)))))
	    ((Evar name type _ ident _)
	     (opt-typed
	      (slterm:Var (intern
			   (or ident
			       ;; Make a new ident and install it.
			       (let ((new-ident 
				      (ident:new-ident-avoiding
				       (symbol-name name) lterm
				       #'(lambda (str)
					   (member str vnames
						   :key #'symbol-name
						   :test #'string=))
				       :uppercase)))
				 (setf (evar-ident lterm) new-ident)
				 new-ident))
			   :ELP))
	      type))
	    ((Uvar name type _ ident)
	     (opt-typed
	      (slterm:Var (intern
			   (or ident
			       ;; Make a new ident and install it.
			       (let ((new-ident 
				      (ident:new-ident-avoiding
				       (symbol-name name) lterm
				       #'(lambda (str)
					   (member str vnames
						   :key #'symbol-name
						   :test #'string=))
				       :lowercase)))
				 (setf (uvar-ident lterm) new-ident)
				 new-ident))
			   :ELP))
	      type))
	    ((Appl lrator lrand)
	     (slterm:Appl (to-slterm-rec lrator ltypes tvnames vnames depth)
			  (to-slterm-rec lrand ltypes tvnames vnames (1+ depth))))
	    ((Abst old-vname ltype lbody)
	     (let ((new-vname (choose-vname old-vname vnames)))
	       (slterm:Abst (if *to-slterm-with-types-p*
				(slterm:Vtyped new-vname
					       (to-sltype ltype tvnames depth))
				(slterm:Vuntyped new-vname))
			    (to-slterm-rec lbody
					   (cons ltype ltypes)
					   tvnames
					   (cons new-vname vnames)
					   (1+ depth)))))))
	;;)
	)))

(defun to-slcobj (lcobj)
  "Convert LCOBJ into a slcobj."
  (declare-ftype to-slcobj (Lcobj) Slcobj)
  (dcase lcobj
    ((String-lcobj string)
     (slterm:Cstring string))
    ((Number-lcobj number)
     (slterm:Cnumber number))
    ((Named-lcobj name)
     (slterm:Cnamed name))))


(defun choose-vname (old-vname used-names)
  "Choose a symbol as close as possible to OLD-VNAME that is not in USED-NAMES
and is not already used as a evar or uvar."
  (declare-ftype choose-vname (Symbol (List-of Symbol)) Symbol)
  (flet ((ok (candidate)
	   (not (or (find candidate used-names)
		    (ident:ident-retrieve (symbol-name candidate))))))
    (if (ok old-vname)
	old-vname
	(dlet ( ((intern prefix-string package) old-vname) )
	  ;; Try numerical suffixes until we find one that has not been taken.
	  (labels ((try-suffix (i)
		     (let ((new-vname
			    (intern (format nil "~a~d" prefix-string i)
				    package)))
		       (if (ok new-vname)
			   new-vname
			   (try-suffix (1+ i))))))
	    (try-suffix 1))))))


(defun to-sltype (ltype tvnames depth)
  "Translate given LTYPE, in the context of TVNAMES, to a sltype"
  (declare-ftype to-sltype (Ltype (List-of Symbol)) slterm:Sltype)
  (if (= depth *elp-print-depth*)
      (slterm:EllidedSltype)
      (let ((ltype (chase-tevars ltype)))
	(dcase ltype
	  ((Tcon name _)
	   (slterm:Tcon name))
	  ((Tbvar index)
	   (let ((name (nth index tvnames)))
	     (slterm:Tvar (or name
			      (depth-name "tv" index)))))
	  ((Tevar name ident _)
	   (slterm:Tvar (intern
			 (or ident
			     ;; Make a new ident and install it.
			     (let ((new-ident 
				    (ident:new-ident-avoiding
				     (symbol-name name)
				     ltype
				     #'(lambda (str)
					 (member str tvnames
						 :key #'symbol-name
						 :test #'string=))
				     :uppercase)))
			       (setf (tevar-ident ltype) new-ident)
			       new-ident))
			 :ELP)))
	  ((Tfun domain range)
	   (slterm:Tfun (to-sltype domain tvnames (1+ depth))
			(to-sltype range tvnames (1+ depth))))
	  ((Tappl lrator lrand)
	   (slterm:Tappl (to-sltype lrator tvnames depth)
			 (to-sltype lrand tvnames (1+ depth))))))))



(defun to-slkind (lkind)
  "Convert LKIND to a slkind."
  (declare-ftype to-slkind (lkind) slterm:Slkind)
  (dcase lkind
    ((Ktype)
     (slterm:Ktype))
    ((Kfun slrange)
     (slterm:Kfun (to-slkind slrange)))))


(ddefun ltyfam-to-sltype ( (Ltyfam vnames body) depth)
  "Translate an ltyfam into a sltyfam."
  (declare-ftype ltyfam-to-sltype (Ltyfam) Sltype)
  (to-sltype body vnames depth))


(defun depth-name (prefix-str depth)
  "Make a name for a bound type or term variable, given the PREFIX-STR and
DEPTH."
  (declare-ftype depth-name (String Nat) Symbol)
  (intern (format nil "~:@(~a~)-~d" prefix-str depth)))
