;;; -*- Mode: Lisp; Package: LTERM -*-
;;;
;;; Ltype unifier.  Destructively instantiates tevars.
;;;
;;; Author: Conal Elliott.  Last Modified Mon Sep 11 06:08:59 1989
;;;
;;; Sccs Id @(#)lterm-tunify.lisp	1.15 9/15/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :LTERM) (use-package :ergolisp) 

;;; Whether to record on the trail.
(defvar *push-trail-p*)


(defun unify-ltype (ltype-a ltype-b trailp)
  "Unify the ltypes LTYPE-A and LTYPE-B (destructively instantiating their
tevars), recording on *trail* if TRAILP.  Returns t or nil,
depending on whether unification succeeded.  The trail is unwind even
on failure."
  (declare-ftype unify-ltype (Ltype-closed Ltype-closed Boolean) Boolean)
  ;; Set up the local trail and call unify-ltype-rec
  (let ((*push-trail-p* trailp))
    (unify-ltype-rec ltype-a ltype-b)))

(defun unify-ltype-rec (ltype-a ltype-b)
  "Workhorse for unify-ltype.  Returns t or nil."
  (declare-ftype unify-ltype-rec (Ltype Ltype) Boolean)
  ;; First chase type indirections.  Then we know that if either type is a
  ;; tevar, it is uninstantiated.
  (let ((ltype-a (chase-tevars ltype-a))
	(ltype-b (chase-tevars ltype-b)))
    ;; Do EQ check.  This should make things go a lot faster.  Take it out
    ;; later and compare.
    (or (eq ltype-a ltype-b)
	(cond ((is-tevar ltype-a)
	       (instantiate-tevar ltype-a ltype-b))
	      ((is-tevar ltype-b)
	       (instantiate-tevar ltype-b ltype-a))
	      (t (dcase ltype-a
		   ;; It might be that tcon's with the same name are guaranteed
		   ;; to be EQ.  If so, the following name eq test will always
		   ;; fail.
		   ((Tcon name-a _)
		    (dcase ltype-b
		      ((Tcon name-b _)
		       (eq name-a name-b))
		      (_ nil)))
		   ((Tfun domain-a range-a)
		    (dcase ltype-b
		      ((Tfun domain-b range-b)
		       (and (unify-ltype-rec domain-a domain-b)
			    (unify-ltype-rec range-a range-b)))
		      (_ nil)))
		   ((Tappl rator-a rand-a)
		    (dcase ltype-b
		      ((Tappl rator-b rand-b)
		       (and (unify-ltype-rec rator-a rator-b)
			    (unify-ltype-rec rand-a rand-b)))
		      (_ nil)))))))))


(defun instantiate-tevar (tevar ltype-b)
  "Instantiate the TEVAR to  LTYPE-B, first doing the occur-check first.
If TEVAR is already eq to LTYPE-B, succeed immediately.   Return t or nil,
depending on whether successful."
  (declare-ftype instantiate-tevar (Ltype Ltype) Boolean)
  ;; First define function that checks for tevar in a given ltype.
  (labels ((occur-check (ltype)
	     (let ((ltype (chase-tevars ltype)))
	       (or (eq tevar ltype)
		   (dcase ltype
		     (Tcon nil)
		     ((Tfun domain range)
		      (or (occur-check domain)
			  (occur-check range)))
		     ((Tappl rator rand)
		      (or (occur-check rator)
			  (occur-check rand)))
		     (_ nil))))))
    (cond ((eq tevar ltype-b)
	   t)
	  ((occur-check ltype-b)
	   nil)
	  (t (setf (tevar-value-slot tevar) ltype-b)
	     (when *push-trail-p*
	       (push tevar elp:*trail*))
	     t))))
