;;; -*- Mode: Lisp; Package: LTERM -*-
;;;
;;; Lterm higher-order unification.  Uses same success continuation control
;;; strategy as the LP interpreter, i.e., for failure, return no values, and
;;; for success, invoke the current success continuation.  Ultimate success
;;; is handled by a throw from the initial success continuation.
;;;
;;; Author: Conal Elliott.  Last Modified Tue Sep 12 22:31:11 1989
;;;
;;; Sccs Id @(#)lterm-unify.lisp	1.29 1/7/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :LTERM) (use-package :ergolisp)

(export '(*proj-first* *warn-typevar* *warn-type-unify*))
(export '(*depth-bound* *bounded-unify* *warn-too-deep*))

(defvar *proj-first* nil
  "Controls whether HOU does projections first.  Defaults to false.")

(defvar *warn-typevar* t
  "Controls whether to warn when the argument to match has uninstantiated
result type.")

(defvar *warn-type-unify* t
  "Controls whether failed type unification results in a warning message.")

(defvar *depth-bound* 20
  "Maximum depth unifier will search.  Negative (default) means no bound.")

(defvar *bounded-unify* nil
  "Controls whether to bound the search for a unifier.")

(defvar *warn-too-deep* t
  "Whether to give warning when the unification depth bound is exceeded.")

(defvar *trace-match* nil
  "Whether to trace the MATCH unification phase.")

;;;;	Disagreement pairs and sets.

(defsconstr "

   ABBREV Rhead == Const or Uvar  	% Rigid head, (bvars eliminated)
   ABBREV Head == Rhead or Bvar or Evar	% Arbitrary head

   datatype Dpair of
     %% Disagreement pair.
     left : Lterm,
     right : Lterm,
     lhead : Lterm,
     rhead : Lterm

  ABBREV Dset == Dpair*
")

(defun headp (lterm)
  "Whether LTERM is a Head."
  (declare-ftype headp (Lterm) Boolean)
  ;; (typep lterm 'Head)
  (or (is-Const lterm)
      (is-Uvar lterm)
      (is-Bvar lterm)
      (and (is-Evar lterm) (not (evar-value-slot lterm)))))

(ddefun print-dpair ( (Dpair left right _ _) stream _)
  (format stream "<~a == ~a>" left right))

#|
(ddefun print-dpair ( (Dpair left right lhead rhead) stream _)
  (format stream "~a <~a> = ~a <~a>" left lhead right rhead))
|#

(deftype Constraint ()
  "Constraint -- flexible-flexible dset."
  `(and Dset (satisfies dset-flex-flex-p))
  )

(defun dset-flex-flex-p (dset)
  "Determine whether DSET is entirely made up of flexible-flexible pairs.
This is expensive, and so should only be done for debug type-checking."
  (declare-ftype dset-flex-flex-p (Dset) Boolean)
  (every (fdlambda ( (Dpair _ _ (iterm lhead) (iterm rhead)) )
	   (and (is-evar lhead)
		(is-evar rhead)))
	 dset))

(defconstant *empty-constraint* ()
  "The empty constraint, as used by the LP term language interface.")


(defvar *unify-sc* 'unify-sc-unbound
  ;; : elp:cont
  "The current scont for the unifier.  Saved as a dynamic variable since it
does not change during a unification problem.")

(defvar *current-depth* 'current-depth-unbound
  "How deep the unifier currently is.")


(defun unify (left right con *unify-sc*)
  "Unify lterms LEFT and RIGHT (which are assumed to have the same type),
extending the constraint CON, and performing the scont UNIFY-SC after each
success."
  (declare-ftype unify (Lterm Lterm Constraint elp:Scont) )
  (let ((*current-depth* 0))
    (elp:trailing
     (SIMPL (list (terms-to-dpair left right)) con)))
  (values))


(defun clause-head-unify (clause-term goal-term new-tevars new-evars con sc)
  "Do clause head unification, unifying the CLAUSE-TERM with the GOAL-TERM,
adding to the constraint CON and invoking the scont SC after each success.
When necessary, copy the CLAUSE-TERM, by substituting the NEW-TEVARS and
NEW-EVARS."
  (declare-ftype clause-head-unify
    (Term Term (List-of Tevar) (List-of Evar)
     Constraint elp:Scont) ) 
  ;; For efficiency, the copying should be postponed until after rigid-rigid
  ;; decompositions.  However, for now, just copy right away, and call SIMPL.
  ;; Also, we know that clause-term and goal-term are rigid terms with the same
  ;; head.
  ;; Changed the order of first two args of the next unify in order to help
  ;; preserve variable names (perhaps).  Could this make the interpreter
  ;; less efficient?  -fp.  Switched back.  Seemed to have no discernible
  ;; effect.
  (unify (copy-lterm clause-term new-tevars new-evars 0)
	 goal-term
	 con sc))

(defsconstr "

    datatype Mpair of
      %% Match pair.  Contains the flexible and rigid heads chosen.
      evar : Evar,
      rhead : Rhead
")

(deftype Pass-stat ()
  "A status value returned by SIMPL-pass.  Can be either (a) :failed (for
rigid-rigid or rigid path check failure), (b) :did-var-term (if one or more
variable-term pairs were handled), (c) :done (only flexible-flexible pairs
remain), (d) an mpair if a flexible-rigid pair was found and no variable-term
cases were handled."
  `(or (member :failed :did-var-term :done)
       Mpair))


(defun SIMPL (dset processed)
  "Given DSET and (already simplified dset) PROCESSED, handle all of the
rigid-rigid, variable-term and higher-order cases.  Choose a pair for match.
Uses scont *unify-sc* when there are only flexible-flexible terms left.
This operation may affect the trail even on failure (because of type
unifications), so be sure to use `trailing'.  As usual, returns (values) on
failure."
  (declare-ftype SIMPL (Dset Dset) )
  (if (and *bounded-unify* (= *current-depth* *depth-bound*))
      (progn
	(elp:lp-warn *warn-too-deep*
		     "Unification depth bound ~d exceeded." *depth-bound*)
	(values))			; Fail
      (dlet ( ((values new-dset status)
	       (catch 'SIMPL (SIMPL-pass dset processed))) )
	(case status
	  (:failed (values))
	  (:did-var-term (SIMPL new-dset ()))
	  (:done (elp:sc-invoke *unify-sc* new-dset))
	  (t 
	   (let ((*current-depth* (1+ *current-depth*)))
	     (MATCH status new-dset)))))))


(defun SIMPL-fail ()
  "Signal SIMPL failure, due to rigid-rigid clash or rigid path check."
  (throw 'SIMPL (values () :failed)))


(defun SIMPL-pass (dset processed)
  "Perform one pass over DSET, extending PROCESSED, handling rigid-rigid,
variable-term and higher-order pairs, and re-establishing head normal form
and looking for a flexible-rigid pair.  Returns two values: the new dset,
and a `status' value (see documentation for the type Pass-stat).  In the
case of rigid-rigid failure, these values are thrown to 'SIMPL, rather
than returned."
  (declare-ftype SIMPL-pass (Dset Dset) Dset Pass-stat)
  (let ((*did-var-term* nil)
	(*flex-rigid-found* nil))
    (declare (special *did-var-term* *flex-rigid-found*))
    (SIMPL-pass-rec dset processed)))

(defun SIMPL-pass-rec (dset now-processed)
  "Make a pass over DSET, where NOW-PROCESSED contains the
already processed dpairs.  Return a fully processed dset,
plus a status.  This is the work-horse for SIMPL-pass."
  (declare (special *did-var-term* *flex-rigid-found*))
  (declare-ftype SIMPL-pass-rec (Dset Dset) Dset Pass-stat)
  ;; (format t "~&PASS ~s ~s~%" dset now-processed)
  (dcase dset
    (() (values now-processed
		(cond (*did-var-term* :did-var-term)
		      (*flex-rigid-found*)
		      (t :done))))
    ((cons (:as dpair (Dpair (iterm left) (iterm right)
			     (iterm lhead) (iterm rhead)))
	   rest-dset)
     (let ((fixed-dpair-p nil))
       (labels ((do-simple-var-term (evar lterm new-dpairs)
		  ;; Just do a variable instantiation and continue.
		  (setf (evar-value-slot evar) lterm)
		  (push evar elp:*trail*)
		  (setq *did-var-term* t)
		  (simpl-pass-rec rest-dset
				  (nconc new-dpairs now-processed)))
		(do-vvar-term (v-evar evar lterm head try-reverse-p)
		  (declare (special *did-var-term*))
		  (if (is-virtual-Evar lterm head) ;; added this check and
		      (progn                       ;; the THEN branch -dmiller
			(cond ((do-gvar-gvar v-evar evar lterm head)
			       (setq *did-var-term* t)))
			(simpl-pass-rec rest-dset now-processed))
		      (dlet ( ((values opt-subst new-dpairs)
			       (vvar-term-case v-evar evar
					       lterm head)) )
			(cond (opt-subst
			       (do-simple-var-term evar opt-subst new-dpairs))
			      ((and try-reverse-p
				    (is-virtual-evar lterm head))
			       (do-vvar-term lterm head v-evar evar nil))
			      (t (pass-shift))))))
		(pass-shift ()
			    ;; Call pass, shifting this pair to
			    ;; now-processed.
			    (simpl-pass-rec rest-dset
				  (cons
				   (if fixed-dpair-p
				       (Dpair left right lhead rhead)
				       dpair)
				   now-processed))))
	 ;; First, check for EQness.  
	 (if (eq left right)
	     (simpl-pass-rec rest-dset now-processed)
	     ;; Then check for terms whose heads have instantiated.
	     (progn
	       ;; Should these tests be instead for instantiated evars, instead
	       ;; of chasing above (via iterm)?
	       (unless (headp lhead)
		 (multiple-value-setq (left lhead)
		   (re-normalize left))
		 (setq fixed-dpair-p t))
	       (unless (headp rhead)
		 (multiple-value-setq (right rhead)
		   (re-normalize right))
		 (setq fixed-dpair-p t))
	       (cond
		;; Check for simplest variable-term case
		((and (is-evar left) (ok-to-subst-p left right))
		 (do-simple-var-term left right ()))
		((and (is-evar right) (ok-to-subst-p right left))
		 (do-simple-var-term right left ()))
		;; Check for higher-order pair.
		;; Should this be done before or after
		;; the generalized variable/term case?
		((or (is-Abst left) (is-Abst right))
		 ;; Make a new uvar and set up applications
		 ;; (usually beta redeces) that will lower
		 ;; the type.
		 (simpl-pass-rec (cons (dpair-add-uvar left right)
			     rest-dset)
		       now-processed))
		;; Check for generalized variable-term case:
		((is-virtual-Evar left lhead)
		 (do-vvar-term left lhead right rhead t))
		((is-virtual-Evar right rhead)
		 (do-vvar-term right rhead left lhead nil))
		;; see if left is flexible
		((is-Evar lhead)
		 (if (is-Evar rhead)
		     ;; Flexible-flexible
		     (pass-shift)
		     ;; Flexible-rigid
		     (progn 
		       (unless *flex-rigid-found*
			 (setq *flex-rigid-found*
			       (Mpair lhead rhead)))
		       (pass-shift))))
		;; Otherwise, left is rigid
		(t
		 (if (is-Evar rhead)
		     ;; Rigid-flexible
		     (progn
		       (unless *flex-rigid-found*
			 (setq *flex-rigid-found*
			       (Mpair rhead lhead)))
		       (pass-shift))
		     ;; Rigid-rigid
		     (simpl-pass-rec (rigid-rigid-add left right
					    lhead rhead
					    rest-dset)
			   now-processed)))))))))))

(defun rigid-rigid-add (left right lhead rhead rest-dset)
  "Do a rigid-rigid decomposition of LEFT and RIGHT, in hnf and having the
given rigid rigid heads LHEAD and RHEAD, adding onto REST-DSET.  Assumes all
indirection-free, and that LEFT and RIGHT are lambda-free head-normal.
Throw the values () and :failed to 'SIMPL if mismatch.  Otherwise, return
extension of REST-DSET."
  (declare-ftype rigid-rigid-add
    ((and Lterm-lfhnf Noind) (and Lterm-lfhnf Noind)
     (and Rhead Noind) (and Rhead Noind) Dset)
    Dset)
  (if (unify-rheads lhead rhead)
      ;; If the heads unify, we have to have the same number of arguments.
      (labels ((add-rands (left right)
		 (dcase left
		   ((Appl (iterm left-rator) (iterm left-rand))
		    (dlet ( ((Appl (iterm right-rator) (iterm right-rand))
			     right) )
		      ;;(format t "~&Adding pair ~s = ~s~%" left-rand right-rand)
		      (push (terms-to-dpair left-rand right-rand)
			    rest-dset)
		      (add-rands left-rator right-rator)))
		   (_ (values)))))
	(add-rands left right)
	rest-dset)
      (SIMPL-fail)))



(defun RESIMPL (evar lterm old-dset)
  "Instantiate EVAR to LTERM and reSIMPLify OLD-DSET."
  (declare-ftype RESIMPL (Evar Lterm Dset) )
  ;; Ideally, this function would reuse old HNF info, as described in our
  ;; design notes, but for now, these HNFs are not saved.
  ;; Do the instantiation.  We probably don't really need to
  ;; maintain the trail here.
  (elp::lp-trace *trace-match*  *current-depth*
	    "MATCH: trying ~a == ~a" evar lterm)
  (elp:trailing
   (setf (evar-value-slot evar) lterm)
   (push evar elp:*trail*)
   (SIMPL old-dset ()))
  (values))


;;; Misc stuff

(defun do-normalize (lterm)
  "Normalize LTERM.  Return new indirection-free term and head."
  (declare-ftype do-normalize
    (Lterm) (and Lterm Noind) (and Head Noind))
  (let ((norm-term (head-beta-norm lterm)))
    (values norm-term (lterm-head norm-term))))

(defun re-normalize (lterm)
  "Re-normalize the LTERM.  Just calls do-normalize, but for now helps us
trace."
  (do-normalize lterm))


#|  Don't call this anymore.
(defun has-fun-type (lterm)
  "Does LTERM, which is in hnf and without indirection, have functional type?"
  (declare-ftype has-fun-type ((and Lterm-hnf Noind)) Boolean)
  ;; Simplest case is when lterm is an Abst.  Otherwise, it may still be
  ;; higher-order, but partially eta-reduced.
  (or (is-Abst lterm)
      ;; This could perhaps be optimized, since we don't need the term, but
      ;; I expect not by much.
      (is-Tfun (ltype-of lterm ()))))
|#

(defun dpair-add-uvar (left right)
  "Make a dpair out of LEFT RIGHT LHEAD and RHEAD (where the first two are 
head-normal, the latter two are atomic, and none are indirected) by applying
to a new uvar."
  (declare-ftype dpair-add-uvar
    ((and Lterm-hnf Noind) (and Lterm-hnf Noind)) Dpair)
  ;; Get a reasonable name, if either term is an abstraction, use a suggested
  ;; name, preferring the left to the right.  Otherwise, use "Etavar"
  (let ((uvar (cond ((is-Abst left)
		     (new-uvar (abst-vname left) (abst-type left)))
		    ((is-Abst right)
		     (new-uvar (abst-vname right) (abst-type right)))
		    (t #| (new-uvar 'elp::|x|
				    (tfun-domain (ltype-of left ()))) |#
		       (error "DPAIR-ADD-UVAR called on two non-Absts.")))))
    ;; The heads will usually be the same.  They can only change in a case like
    ;; "\x. \y. x", whose head becomes the new uvar.
    (terms-to-dpair (Appl left uvar) (Appl right uvar))))

(defun terms-to-dpair (left right)
  "Convert indirection-free lterms LEFT and RIGHT into a dpair."
  (declare-ftype terms-to-dpair ((and Lterm Noind) (and Lterm Noind)) Dpair)
  (multiple-value-bind (new-left new-lhead)
      (do-normalize left)
    (multiple-value-bind (new-right new-rhead)
	(do-normalize right)
      (Dpair new-left new-right new-lhead new-rhead))))
	

(defun unify-rheads (rhead-a rhead-b)
  "Try to unify the rigid heads RHEAD-A and RHEAD-B, i.e., check that they
are the same bvar or uvar, or constants with the same cobj.  In the latter case,
unify the types as well."
  (declare-ftype unify-rheads (Rhead Rhead) Boolean)
  ;; First try EQ test.  Must succeed for matching bvars and uvars.
  (or (eq rhead-a rhead-b)
      (dcase rhead-a
	((Const cobj-a type-a)
	 (dcase rhead-b
	   ((Const cobj-b type-b)
	    (and (eq cobj-a cobj-b)
		 (if (unify-ltype type-a type-b t)
		     t
		     (progn (elp:lp-warn *warn-type-unify*
				     "Unification of types of constants ~S and ~S fails"
				     rhead-a rhead-b)
			    nil))))
	   (_ nil)))
	(_ nil))))


;;;; MATCH phase

(ddefun MATCH ((Mpair (:as evar (Evar ename etype estamp _ _)) rhead) dset)
  "Do MATCH phase, using MPAIR and DSET."
  (declare-ftype MATCH (Mpair Dset) )
  ;; Note: Cannot compute what the substitution terms are until we are ready
  ;; to use them, because checking legal projections involves doing type
  ;; unification.
  (elp::lp-trace *trace-match* *current-depth*
	    "MATCH called with ~a : ~a and ~a : ~a"
	    evar (ltype-of evar nil) rhead (ltype-of rhead nil))
  (dlet ( ((values edomains erange)
	   (strip-domains etype)) )
    (let ((erange (chase-tevars erange))
	  (rev-edomains (reverse edomains)))
      (when (is-Tevar erange)
	(elp:lp-warn *warn-typevar*
		 "MATCH: evar ~s has uninstantiated result type.  May lose completeness."
		 evar))
      ;; Make the bvars, which will be arguments to the evars.
      (let ((bvars (labels ((mk-bvars (edoms-left i)
			      (dcase edoms-left
				(() ())
				((cons _ rest)
				 (cons (bvar i)
				       (mk-bvars rest (1- i)))))))
		     (mk-bvars edomains (1- (length edomains))))))
	(flet ((do-subst-from-head (subst-head)
		 ;; Given a head, build the substitution term.
		 ;; First make the new evars.
		 (let ((new-evars
			(mapcar #'(lambda (shdom)
				    (Evar ename
					  (add-domains edomains shdom)
					  estamp
					  nil nil))
				(strip-domains (ltype-of subst-head
							 rev-edomains)))))
		   (RESIMPL evar
			    (add-absts ()
				       edomains
				       (add-appls
					subst-head
					(mapcar #'(lambda (new-evar)
						    (add-appls new-evar bvars))
						new-evars)))
			    dset))))
	  (flet ((imitate ()
		   ;; Do the imitation if possible, i.e., unless the rhead is
		   ;; a uvar whose stamp dominates estamp.
		   (unless (and (is-uvar rhead)
				(stamp-dominates estamp (uvar-stamp rhead)))
		     (do-subst-from-head rhead)))
		 (project ()
		   ;; Do each projection if legal, i.e., if the final result type
		   ;; of the bvar is unifiable with erange.  The trailing around
		   ;; it is because type unification instantiates variables.
		   ;; The ith projection will have head (Bvar N-1-i), where
		   ;; N is the number domains.  Thus we iterate forwards through
		   ;; the domains and backwards through the bvar indices.
		   ;;; -*- Mode: Lisp; Package: LTERM -*-
		   ;; (format t "Projections for ~a with type ~a~%" evar etype)
		   (do ((proj-head-index (1- (length edomains))
					 (1- proj-head-index))
			(domains-left edomains
				      (rest domains-left)))
		       ((null domains-left))
		       ;; (format t " Trying projection ~a of type ~a~%" proj-head-index (first domains-left))
		     (dlet ( ((values _ proj-range)
			      (strip-domains (first domains-left))) )
		       (when (is-Tevar (chase-tevars proj-range))
			 (elp:lp-warn *warn-typevar*
			       "MATCH: Projection ~s has uninstantiated result type.  May lose completeness."
			       proj-head-index))
		       (elp:trailing
			(when (unify-ltype erange proj-range t)
			  (do-subst-from-head (Bvar proj-head-index))))))))
	    (if *proj-first*
		(progn (project) (imitate))
		(progn (imitate) (project))))))))
  (values))


(defun present-solution (tevars evars uvars solution-con
				&optional (stream elp:*elp-terminal-io*) )
  "Present final solution, given EVARS and SOLUTION-CON.  Optional keyword
argument STREAM defaults to elp:*elp-terminal-io*."
  (declare (ignore uvars))		; ignore for now.
  ;; (declare-ftype present-solution ((List-of Evar) Constraint) )
  ;; If we don't have anything to say, print "Solved"
;;  (unless (null solution-con) (format stream "~&Pruning..."))
  (let ((solution-con (prune-constraints solution-con)))
    (let ((said-something nil))
;;      (unless (null evars)
;;	(format stream "~&Formatting answer..."))
      (dolist (lterm evars)
	(dlet ( ((Evar _ _ _ ident _) lterm) )
	  (format stream "~&~A == ~A" ident lterm)
	  (setq said-something t)))
      (unless (null tevars)
	(format stream "~&Type variable substitutions:")
	(dolist (tev tevars)
	  (dlet ( ((Tevar _ ident _) tev) )
	    (format stream "~&    ~A == ~A" ident tev)
	    (setq said-something t))))
      (unless (null solution-con)
	(setq said-something t)
	(format stream "~&Remaining constraints: [")
	(dolist (dpair solution-con)
	  (format stream "~&    ~A" dpair))
	(format stream " ]")
	(format stream "~&Prefix Fragment:~&    ")
	(print-prefix-fragment evars solution-con stream))
      (unless said-something
	(format stream "~&Query succeeded" stream)))))

(defun collect-ue-vars (lterm acc)
  "Add to acc the uvars and evars that have any occurrence in lterm.
The result will not have duplicates."
  (declare-ftype collect-ue-vars
    (Lterm (list-of (or Evar Uvar))) (list-of (or Evar Uvar)))
  (let ((lterm (chase-evars lterm)))
    (dcase lterm
      (Const acc)
      (Bvar acc)
      (Evar (if (member lterm acc :test #'eq) acc (cons lterm acc)))
      (Uvar (if (member lterm acc :test #'eq) acc (cons lterm acc)))
      ((Appl rator rand)
       (collect-ue-vars rator (collect-ue-vars rand acc)))
      ((Abst _ _ body) (collect-ue-vars body acc)))))

(defun collect-from-con (con acc)
  "Add to acc the uvars and evars that are present in the constraint
set con.  The resulting list of uvars and evars will not have duplicates."
  (declare-ftype collect-from-con
    (Constraint (list-of (or Evar Uvar))) (list-of (or Evar Uvar)))
  (dcase con
    (() acc)
    ((cons (Dpair left right _ _) rest)
     (collect-ue-vars left (collect-ue-vars right
					    (collect-from-con rest acc))))))

(defun collect-all (evars con)
  "Collect the uvars and evars that are in the instances of the evars
in EVARS and in the constraint CON."
  (declare-ftype collect-all ((List-of Evar) Constraint)
		 (List-of (or Evar Uvar)))
  (reduce #'collect-ue-vars evars
	  :from-end t :initial-value (collect-from-con con ())))

(defun ue-stamp (ue-var)
  (declare-ftype ue-stamp ((or Evar Uvar)) Stamp)
  (dcase ue-var
    ((Uvar _ _ stamp _) stamp)
    ((Evar _ _ stamp _ _) stamp)))

(defun sort-ue-vars (ue-vars)
  "Sort the evars and uvars in UE-VARS by stamp, and then within evar-blocks,
by name."
  (declare-ftype sort-ue-vars ((List-of (or Evar Uvar)))
		 (List-of (or Evar Uvar)))
  (sort ue-vars
	#'(lambda (x y)
	    (if (stamp-dominates (ue-stamp x) (ue-stamp y))
		t
		(if (eql (ue-stamp x) (ue-stamp y))
		    (dcase x
		      (Uvar t)
		      (Evar (dcase y
			      (Uvar nil)
			      (Evar (string< (Evar-ident x) (Evar-ident y))))))
		    nil)))))

(defun print-ue-vars (in var mode &optional (stream t))
  "This tail recursive program prints out the partial prefix.
Parentheses () are used to denote universal quantification and 
square brackets [] are used to denote existential quantification.
The value for MODE is either :start, :uvar, or an integer, which is the
time stamp of an Evar.  MODE remembers what was the previous action."
  (dcase in
    (() (case mode
	  (:start (dcase var
		    (Uvar (format stream "(pi ~a)" var))
		    (Evar (format stream "[sigma ~a]" var))))
	  (:uvar (dcase var
		   (Uvar (format stream " ~a)" var))
		   (Evar (format stream ")[sigma ~a]" var))))
	  (t     (dcase var
		   (Evar (if (eql mode (evar-stamp var))
			     (format stream " ~a]" var)
			     (format stream "][sigma ~a]" var)))
		   (Uvar (format stream "](pi ~a)" var))))))
    ((cons next-var rest)
     (case mode
       (:start (dcase var
		 (Uvar (format stream "(pi ~a" var)
		       (print-ue-vars rest next-var :uvar stream))
		 (Evar (format stream "[sigma ~a" var)
		       (print-ue-vars rest next-var (evar-stamp var)
				      stream))))
       (:uvar (dcase var
		(Uvar (format stream " ~a" var)
		      (print-ue-vars rest next-var :uvar stream))
		(Evar (format stream ")[sigma ~a" var)
		      (print-ue-vars rest next-var (evar-stamp var) stream))))
       (T     (dcase var
		(Evar (if (eql mode (evar-stamp var))
			  (format stream " ~a" var)
			  (format stream "][sigma ~a" var))
		      (print-ue-vars rest next-var (evar-stamp var) stream))
		(Uvar (format stream "](pi ~a" var)
		      (print-ue-vars rest next-var :uvar stream))))))))

(defun print-prefix-fragment (evars solution-con &optional (stream t))
  "This function just calls print-ue-vars with the right number of 
arguments."
  (let ((ue-vars (sort-ue-vars (collect-all evars solution-con))))
    (if (null ue-vars) nil
	(print-ue-vars (cdr ue-vars) (car ue-vars) :start stream))))

