;;; -*- Mode: Lisp; Package: LTERM -*-
;;;
;;; Handle the variable-term unification optimization.
;;;
;;; Author: Conal Elliott.  Last Modified Mon Sep 11 06:08:55 1989
;;;
;;; Sccs Id @(#)lterm-vtopt.lisp	1.14 5/10/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :LTERM)


;;;;  The variable/term case.


(defun is-virtual-Evar (lterm head)
  "Whether LTERM (hnf and indirection-free), having the given HEAD, is
`virtually' an evar, i.e., whether it is an evar applied to zero or more
distinct uvars that it dominates."
  (declare-ftype is-virtual-Evar ((and Lterm-hnf Noind) Head) T)
  (and (not (is-Abst (chase-evars lterm)))
       (dcase head
	 ((Evar _ _ estamp _ nil)
	  (let ((args (mapcar #'chase-evars (lterm-rands lterm))))
	    (and (every #'(lambda (rand)
			    (and (is-Uvar rand)
				 (stamp-dominates estamp (uvar-stamp rand))))
			args)
		 (all-diff-uvars args))))
	 (_ nil))))

(deftype Virtual-Evar ()
  `(and Lterm-hnf Noind
	(satisfies is-virtual-Evar)))

(defun all-diff-uvars (uvars)
  "Check that all of the given UVARS are distinct."
  (declare-ftype all-diff-uvars ((List-of Uvar)) Boolean)
  (dcase uvars
    (nil t)
    ((cons first rest)
     (and (not (member first rest :test #'EQ))
	  (all-diff-uvars rest)))))


;;; For accumulating bvars on the way down in the variable/term optimization.
(defvar *vterm-vnames*)
(defvar *vterm-vtypes*)
(defvar *vterm-depth*)

(defmacro vterm-push-abst-in (vname vtype &rest body)
  "Add VNAME and VTYPE to the current variable/term binding context while
processing thee rest parameter BODY."
  `(let ((*vterm-vnames* (cons ,vname *vterm-vnames*))
	 (*vterm-vtypes* (cons ,vtype *vterm-vtypes*))
	 (*vterm-depth* (1+ *vterm-depth*)))
     ;; (format t "Adding abst ~s : ~s.~%" vname vtype)
     ,@body))

;;; Some utility specials.
(defvar *evar*)
(defvar *ename*)
(defvar *estamp*)
(defvar *new-dpairs*)
(defvar *uvar-args* ())			; Initialized for ok-to-subst-p
(defvar *uvar-adjust*)			; leng-1
(defvar *uvar-names*)
(defvar *uvar-types*)
(defvar *descendedp*)

(ddefun vvar-term-case (v-evar *evar* lterm head)
  "Generalized variable/term case, unifying the virtual evar V-EVAR (see
is-virtual-Evar), having head EVAR, with the nonEQ LTERM having the given HEAD."
  (declare-ftype vvar-term-case
    (Virtual-Evar Evar (and Lterm Noind) Head)
    (Optional Lterm) Dset )
  (dlet ( ((Evar *ename* _ *estamp* _ nil) *evar*) )
    (let ((*new-dpairs* ())
	  (*uvar-args* (lterm-rands v-evar))
	  (*descendedp* nil)
	  (*vterm-vtypes* ())
	  (*vterm-vnames* ())
	  (*vterm-depth* 0))
      (let ((*uvar-names* (mapcar #'uvar-name *uvar-args*))
	    (*uvar-types* (mapcar #'uvar-type *uvar-args*))
	    (*uvar-adjust* (1- (length *uvar-args*))))
	(let ((opt-subst-term (catch 'VAR-TERM-NO-PROGRESS
				(do-vt lterm head))))
	  (if opt-subst-term
	      (values (add-uvar-absts opt-subst-term)
		      *new-dpairs*)
	      (values nil ())))))))

(defun add-uvar-absts (lterm)
  "Make an abstraction out of LTERM using the uvar-args."
  (add-absts *uvar-names* *uvar-types* lterm))


(defun do-vt (lterm head)
  "Keep imitating.  LTERM is head normal with HEAD, and both are
indirection-free."
  (declare-ftype do-vt ((and Lterm Noind) Head) Lterm)
  (dcase lterm
    ((Abst vname vtype body)
     (vterm-push-abst-in
      vname vtype
      (re-Abst-body lterm vname vtype body 
		    (do-vt body head))))
    (_
     ;; Check for imitable head
     (let ((imit-head (cond
		       ((or (is-const head)
			    (is-bvar head)
			    (and (is-uvar head)
				 (not (stamp-dominates *estamp*
						       (uvar-stamp head)))))
			head)
		       ((member head *uvar-args* :test #'eq)
			(adjust-uvar head)))))
       (cond (imit-head
	      (setq *descendedp* t)
	      (do-vt-imitate imit-head lterm))
	     ;; Then for non-imitable uvar, or for the same evar,
	     ;; properly within lterm.
	     ((or (is-uvar head)
		  (and (eq *evar* head)
		       *descendedp*))
	      (SIMPL-fail))
	     ;; Otherwise, we have a flexible term.  Treat no-arg case
	     ;; specially.  (Cannot be the same evar, since they
	     ;; are not EQ.)
;; The following case is not correct, a counter example is the following:
;; ?- pi X : a\ (sigma H: d\ (pi Y : b \ (H = G X Y, write H))).
;; Here, there is a term to the left of H that is an argument of the 
;; generalized variable.  For now, let this case drop through. -dmiller
;	     ((is-evar lterm)
;	      (dlet ( ((:as evar-b
;			    (Evar ename-b etype-b estamp-b _ nil))
;		       head) )
;		;; For "exists X. all Y. exists Z. X = c Z", we
;		;; instantiate Z to a variable Z' with X's stamp (which
;		;; says it may not be instantiated to contain Y) and
;		;; add the equation.  Works even if there are lambdas
;		;; before the Z, since Z cannot depend on them.  Actually, more
;		;; complicated (raised) because of uvar-args.
;		(if (stamp-dominates *estamp* estamp-b)
;		    (let* ((new-evar (Evar ename-b
;					   (add-domains *uvar-types* etype-b)
;					   *estamp* nil nil))
;			   (raised-new-evar
;			    (add-reloc-bvar-args
;			     new-evar
;			     (length *uvar-args*)
;			     *vterm-depth*)))
;		      (setf (evar-value-slot evar-b) 
;			    (add-appls new-evar
;				       *uvar-args*))
;		      (push evar-b elp:*trail*)
;		      raised-new-evar)
;		    lterm)))
	     ;; Left with flexible term having some arguments.
	     ;; Two cases: If we can copy the whole flexible term, do
	     ;; so. Otherwise, just make a new evar to put in here, and
	     ;; add a new flexible-flexible equation.  Must "raise"
	     ;; over the variables abstracted over so far.  However, if
	     ;; we haven't yet descended through a rigid head, we
	     ;; aren't really making any progress, so just return
	     ;; nil, saying that we didn't do anything.
	     ;; NOTE: There is another possible optimization: if the
	     ;; arguments to the flexible head are the bound variables,
	     ;; we ought to be able to make a direct substitution.  I
	     ;; think this would be fairly common.
	     ((ok-to-subst-p *evar* lterm)
	      (uvar-args-to-bvars lterm))
	     (*descendedp*
	      (let ((new-evar (Evar *ename*
				    (add-domains
				     *uvar-types*
				     (add-domains
				      *vterm-vtypes*
				      (ltype-of lterm *vterm-vtypes*)))
				    *estamp* nil nil)))
		(push (terms-to-dpair
		       (add-appls new-evar *uvar-args*)
		       (add-absts *vterm-vnames* *vterm-vtypes* lterm))
		      *new-dpairs*)
		(add-bvar-args new-evar
			       (+ (length *uvar-args*)
				  (length *vterm-vtypes*)))))
	     (t (throw 'VAR-TERM-NO-PROGRESS ())))))))

(defun do-vt-imitate (head lterm)
  "Build the variable-term imitation, using HEAD for the head, and the result
of applying do-vt to the rands of LTERM for the rands."
  (declare-ftype do-vt-imitate 
    (Head (and Lterm-hnf Noind)) (and Lterm-hnf Noind))
  ;;; Modified from map-rands
  (labels ((rec-appls-ni (lterm)
	     (dcase lterm
	       ((Appl rator rand)
		(re-Appl lterm rator rand
			 (rec-appls-ni (chase-evars rator))
			 (let ((norm-rand (head-beta-norm rand)))
			   (do-vt norm-rand (lterm-head norm-rand)))))
	       (_ head))))
    (rec-appls-ni lterm)))

(ddefun ok-to-subst-p ((:as evar (Evar _ _ stamp-a _ nil)) lterm)
  "See if it is ok to simply substitute for the given EVAR the given LTERM.
The uvars in *uvar-args* are allowed to show up."
  (declare-ftype ok-to-subst-p (Evar Lterm) Boolean)
  (labels ((ok (lterm)
	     (setq lterm (chase-evars lterm))
	     (dcase lterm
	       (Const t)
	       (Bvar t)
	       ((:as evar-b (Evar _ _ stamp-b _ nil))
		(not (or (eq evar evar-b)
			 (stamp-dominates stamp-a stamp-b))))
	       ((Uvar _ _ stamp-b _)
		(or (member lterm *uvar-args* :test #'EQ)
		    (not (stamp-dominates stamp-a stamp-b))))
	       ((Appl rator rand)
		(and (ok rator)
		     (ok rand)))
	       ((Abst _ _ body)
		(ok body)))))
    (ok lterm)))

(defun uvar-args-to-bvars (lterm)
  "Replace members of *uvar-args* in LTERM by appropriate bvars."
  (declare-ftype uvar-args-to-bvars (Lterm) Lterm)
  (setq lterm (chase-evars lterm))
  (dcase lterm
    (Uvar (adjust-uvar lterm))
    ((Appl rator rand)
     (re-Appl lterm rator rand
	      (uvar-args-to-bvars rator)
	      (uvar-args-to-bvars rand)))
    ((Abst name type body)
     (vterm-push-abst-in
      name type
      (re-Abst-body lterm name type body
		    (uvar-args-to-bvars body))))
    (_ lterm)))


;;; This would be simpler if the uvars arguments were in reverse order.

(defun adjust-uvar (uvar)
  "Make UVAR into a bvar if it is one of *uvar-args*.  Otherwise, leave it."
  (declare-ftype adjust-uvar (Uvar) (or Uvar Bvar))
  (let ((pos (position uvar *uvar-args* :test #'eq)))
    (if pos
	(Bvar (+ (- *uvar-adjust*
		    pos)
		 *vterm-depth*))
	uvar)))

(defun add-bvar-args (head num-args)
  "Make a term by applying HEAD to (Bvar 0), ..., (Bvar (- NUM-ARGS 1) )."
  (declare-ftype add-bvar-args (Lterm Nat) Lterm)
  (dcase num-args
    (0 head)
    ((1+ n)
     (add-bvar-args (Appl head (Bvar n)) n))))

(defun add-reloc-bvar-args (head num-args skip)
  "Make a term by applying HEAD to (Bvar (+ SKIP 0), ...,
(Bvar (+ SKIP (- NUM-ARGS 1)) )."
  (declare-ftype add-bvar-args (Lterm Nat Nat) Lterm)
  (dcase num-args
    (0 head)
    ((1+ n)
     (add-reloc-bvar-args (Appl head (Bvar (+ skip n))) n skip))))
