;;; -*- Mode: Lisp; Package: SLTERM -*-
;;;
;;; Establish the term representation for slterms and smodules.
;;;
;;; Author: Conal Elliott.  Last Modified Mon Sep 11 06:09:28 1989
;;;
;;; Sccs Id @(#)slterm-term-rep.lisp	1.26 9/15/89
;;; ******************************************************************* ;;;
;;;         (c) Copyright 1989 by Carnegie Mellon University.           ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;


(in-package :SLTERM) (use-package :ergolisp)


(defmacro defrep-sl (sort &rest cases)
  "Like gterm:defrep-cases, but assumes sort-sym and rep-type to be the same,
and requires specification of only one direction, inferring the other to be
the obvious inverse.  For convenience, If the two versions are the same, the
second instance may be omitted."
  ;; First go through the list and duplicate single versions.
  (let ((cases (mapcar #'(lambda (case)
			   (dlet ( ((list (cons op-sym args) rep)
				    (if (= (length case) 1)
					(list (first case) (first case))
					case)) )
			     (list (cons op-sym args) rep)))
		       cases)))
    `(progn (gterm:defrep-cases
	     ,sort ,sort
	     ,cases
	     ,(mapcar (fdlambda ( (list (cons op-sym args) rep) )
			(list rep `(values (oper:mk-sim-op ',op-sym)
					   (list ,@args))))
		      cases))
	    ;; Now, make all of the constr (sub)types have the same
	    ;; ds-term-impl as multiconstr.
	    ,@(mapcar (fdlambda ( (list (cons constr _) _) )
			`(setf (gterm:ds-term-impl ',constr)
			       (gterm:ds-term-impl ',sort)))
		      cases))))


;;;; Now the representations. 

;; See below for these constructor definitions.

(declare-constructor slterm-to-vbind (vbind-to-slterm))
(declare-constructor slcobj-to-slterm (slterm-to-slcobj))
(declare-constructor name-to-sltype (sltype-to-name))


(defrep-sl Slterm			; New version.
  ((con slcobj)
   (slcobj-to-slterm slcobj))
  ;; The next one is not used, but is needed for auto-unparsing.
  ((var name))
  ((appl rator rand))
  ((abst left right)
   (abst (slterm-to-vbind left) right))
  ((typed term type))
  ((ellidedslterm)) )

;;; To unparse declared (infix) operators as keywords.

(defun cnamed-key (name)
  (cnamed name))

(ddefun cnamed-key-name ( (cnamed name) )
  name)

;;; Next could and should be made more efficient.

(defun is-cnamed-key (obj)
  (and (is-cnamed obj)
       (member (symbol-name (cnamed-first obj))
	       sltermg::slterm-keyword-list
	       :test #'string= :key #'symbol-name)
       ))

(declare-constructor cnamed-key (cnamed-key-name) is-cnamed-key)

(defrep-sl Slcobj
  ;; The keyword case has to come first for unparsing purposes.
  ((ckeyword (elp:lp-key name))		; keywords parse into strings.
   (cnamed-key name))
  ((cnamed (elp:lp-id name :slterm))
   (cnamed name))
  ((cstring (term:mk-string str))
   (cstring str))
  ((cnumber (term:mk-number num))
   (cnumber num))
  ;; We only use the next case for parsing, not for unparsing of slcobj's.
  ;; Commented out for now: not really needed.
  ;;((cliteral (elp:lp-lit name :slterm))	
  ;;(cnamed name))
  )


(defrep-sl Sltype
  ((tcon (elp:lp-id name :slterm))
   (name-to-sltype name))
  ;; The next one is not used, but is needed for auto-unparsing.
  ((tvar name))
  ((tfun domain range))
  ((tappl rator rand))
  ((ellidedsltype)) )

(defrep-sl Slkind
  ((ktype))
  ((kfun slkind)))

(defun slcobj-to-slterm (slcobj)
  "Convert SLCOBJ into an slterm, taking the case into account for building
variables."
  (declare-ftype slcobj-to-slterm (Slcobj) Slterm)
  (dcase slcobj
    ((Cnamed name)
     (Var name))
    (_ (Con slcobj))))

(defun slcobj-to-slterm-p (slterm)
  "Is SLTERM something that was converted from a slcobj?  In other words, is
it a variable or constant?"
  (declare-ftype slcobj-to-slterm-p (Slterm) Boolean)
  (typep slterm '(or Var Con)))

(defun slterm-to-slcobj (slterm)
  "Convert SLTERM to a slcobj to reverse the affect of slcobj-to-slterm.
Doesn't make sure the case is right."
  (declare-ftype slterm-to-slcobj (Slterm) Slcobj)
  (dcase slterm
    ((Var name)
     (Cnamed name))
    ((Con slcobj)
     slcobj)))

(ddefun name-to-sltype ( (:as name (intern name-string _)) )
  "Convert NAME into an sltype, taking the case into account for building
variables."
  (declare-ftype name-to-sltype (Symbol) Sltype)
  (if (upper-case-p (schar name-string 0))
      (Tvar name)
      (Tcon name)))

(defun name-to-sltype-p (sltype)
  "Is SLTYPE something that was converted from a name?  In other words, is
it a type variable or constant?"
  (declare-ftype name-to-sltype-p (Sltype) Boolean)
  (typep sltype '(or Tvar Tcon)))

(defun sltype-to-name (sltype)
  "Convert SLTYPE to a symbol to reverse the affect of name-to-sltype.
Doesn't make sure the case is right."
  (declare-ftype sltype-to-name (Sltype) Symbol)
  (dcase sltype
    ((Tvar name)
     name)
    ((Tcon name)
     name)))


(defun slterm-to-vbind (slterm)
  "Translate SLTERM to a vbind.  Signal an error if anything other than a
variable or a typed variable."
  (declare-ftype slterm-to-vbind (Slterm) Vbind)
  (flet ((complain ()
	   (elp:lp-error "The slterm ~s, appearing on the left of a \\ should be a variable or typed variable."
		  slterm)))
    (dcase slterm
      ((Var name)
       (Vuntyped name))
      ((Typed bind-slterm sltype)
       (dcase bind-slterm
	 ((Var name)
	  (Vtyped name sltype))
	 (_ (complain))))
      (_ (complain)))))

(defun vbind-to-slterm (vbind)
  "Translate VBIND to a slterm"
  (declare-ftype slterm-to-vbind (Vbind) Slterm)
  (dcase vbind
    ((Vuntyped name)
     (Var name))
    ((Vtyped name sltype)
     (Typed (Var name) sltype))))



;;;; Smodules

;;; These are fairly simple, because we only give parsing, not unparsing
;;; support.


(ddefun mk-smodule ( _ (list (term:mk-sim-term _ (list modname))
			     (term:mk-sim-term _ simports)
			     (term:mk-sim-term _ entries)) )
  (elp:smodule (dcase modname
		 ((term:mk-id (intern name _)) name)
		 ((term:mk-string str) str))
	       (apply #'append simports)
	       entries))

(ddefun mk-simport
    (_ (list (term:mk-sim-term _ modnamelist)))
  ;; Represent as list of names
  (mapcar (fdlambda ( (term:mk-sim-term _ (list modname)) )
	      (dcase modname
		((term:mk-id (intern name _)) name)
		((term:mk-string str) str)))
	  modnamelist))

(ddefun mk-smodentry ( (oper:mk-sim-op op-name) args )
  (ecase op-name
    (mkinding
     (dlet ( ((list (elp::lp-id kind-name :slterm) slkind) args) )
       (elp:Mkinding kind-name slkind)))
    (mtyping
     (dlet ( ((list (elp::lp-id type-name :slterm) sltype) args) )
       (elp:Mtyping type-name sltype)))
    (mkeytyping
     (dlet ( ((list (elp::lp-key type-name) sltype) args) )
       (elp:Mtyping type-name sltype)))
    (mindexing
     (dlet ( ((list (elp::lp-id const-name :slterm) (term:mk-number n)) args) )
       (elp:Mindexing const-name n)))
    (mkeyindexing
     (dlet ( ((list (elp::lp-key const-name) (term:mk-number n)) args) )
       (elp:Mindexing const-name n)))
    (msdform
     (dlet ( ((list slterm) args) )
       (elp:Msdform slterm)))))


(ddefun mk-slterm-dot ( (oper:mk-sim-op op-name) args )
  ;; Represent as just the slterm, or "signal" end-of-input.
  (ecase op-name
    (slterm-dot (first args))
    (end-of-input :eof)))

(setf (gterm:mk-term-impl 'smodule) 'mk-smodule)
(setf (gterm:mk-term-impl 'simport) 'mk-simport)
(setf (gterm:mk-term-impl 'smodentry) 'mk-smodentry)
(setf (gterm:mk-term-impl 'slterm-dot) 'mk-slterm-dot)
